// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*



////////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cstdio>
#include <cmath>

#include "vx_math/vx_math.h"
#include "vx_met_util/vsdb_columns.h"
#include "vx_analysis_util/analysis_utils.h"
#include "vx_analysis_util/vsdb_job.h"
#include "vx_analysis_util/vsdbjobtype_to_string.h"
#include "vx_analysis_util/vsdblinetype_to_string.h"

////////////////////////////////////////////////////////////////////////

static void timestring(const unixtime t, char * out);

////////////////////////////////////////////////////////////////////////
//
// Code for class VSDBAnalysisJob
//
////////////////////////////////////////////////////////////////////////


VSDBAnalysisJob::VSDBAnalysisJob() {

   init_from_scratch();

}

////////////////////////////////////////////////////////////////////////


VSDBAnalysisJob::~VSDBAnalysisJob() {

   clear();
}


////////////////////////////////////////////////////////////////////////


VSDBAnalysisJob::VSDBAnalysisJob(const VSDBAnalysisJob & aj) {

   init_from_scratch();

   assign(aj);
}

////////////////////////////////////////////////////////////////////////


VSDBAnalysisJob & VSDBAnalysisJob::operator=(const VSDBAnalysisJob & aj) {

   if(this == &aj ) return ( * this );

   assign(aj);

   return (* this);
}

////////////////////////////////////////////////////////////////////////


void VSDBAnalysisJob::init_from_scratch() {

   col_name  = (char *) 0;
   boot_rng  = (char *) 0;
   boot_seed = (char *) 0;

   clear();

   return;
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::clear() {

   job_type = no_vsdb_job_type;

   model.clear();
   obtype.clear();
   mask.clear();
   line_type.clear();
   var.clear();
   level.clear();
   interp_mthd.clear();

   init_beg  = init_end  = (unixtime) 0;
   valid_beg = valid_end = (unixtime) 0;

   fcst_lead.clear();
   init_hour.clear();
   interp_pnts.clear();

   dump_row.clear();

   if(col_name) { delete [] col_name; col_name = (char *) 0; }

   // Set to default values
   alpha         = default_alpha;
   boot_interval = default_boot_interval;
   boot_rep_prop = default_boot_rep_prop;
   n_boot_rep    = default_n_boot_rep;
   set_boot_rng(default_boot_rng);
   set_boot_seed(default_boot_seed);

   out_line_type = no_vsdb_line_type;

   threshold.clear();
   thresh.clear();

   mask_grid.clear();

   mask_poly.clear();
   poly.clear();

   return;
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::assign(const VSDBAnalysisJob & aj) {

   clear();

   job_type = aj.job_type;

   model = aj.model;
   obtype = aj.obtype;
   mask = aj.mask;
   line_type = aj.line_type;
   var = aj.var;
   level = aj.level;
   interp_mthd = aj.interp_mthd;

   init_beg = aj.init_beg;
   init_end = aj.init_end;

   valid_beg = aj.valid_beg;
   valid_end = aj.valid_end;

   fcst_lead = aj.fcst_lead;
   init_hour = aj.init_hour;
   interp_pnts = aj.interp_pnts;

   dump_row = aj.dump_row;
   set_col_name(aj.col_name);
   alpha    = aj.alpha;

   boot_interval   = aj.boot_interval;
   boot_rep_prop   = aj.boot_rep_prop;
   n_boot_rep      = aj.n_boot_rep;
   set_boot_rng(aj.boot_rng);
   set_boot_seed(aj.boot_seed);

   out_line_type = aj.out_line_type;

   threshold = aj.threshold;
   thresh = aj.thresh;

   mask_grid = aj.mask_grid;
   grid = aj.grid;

   mask_poly = aj.mask_poly;
   poly = aj.poly;

   return;
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::dump(ostream & out, int depth) const {
   Indent prefix(depth);
   char junk[512];

   vsdbjobtype_to_string(job_type, junk);
   out << prefix << "job = " << junk << "\n";

   out << prefix << "model ...\n";

   model.dump(out, depth + 1);

   out << prefix << "obtype ...\n";

   obtype.dump(out, depth + 1);

   out << prefix << "mask ...\n";

   mask.dump(out, depth + 1);

   out << prefix << "line_type ...\n";

   line_type.dump(out, depth + 1);

   out << prefix << "var ...\n";

   var.dump(out, depth + 1);

   out << prefix << "level ...\n";

   level.dump(out, depth + 1);

   out << prefix << "fcst_lead ...\n";

   fcst_lead.dump(out, depth + 1);

   out << prefix << "init_hour ...\n";

   init_hour.dump(out, depth + 1);

   out << prefix << "interp_mthd ...\n";

   interp_mthd.dump(out, depth + 1);

   out << prefix << "interp_pnts ...\n";

   interp_pnts.dump(out, depth + 1);

   out << prefix << "dump_row ...\n";

   dump_row.dump(out, depth + 1);

   out << prefix << "col_name = " << prefix << col_name << "\n";

   out << prefix << "alpha = " << prefix << alpha << "\n";

   out << prefix << "boot_interval = " << prefix << boot_interval << "\n";

   out << prefix << "boot_rep_prop = " << prefix << boot_rep_prop << "\n";

   out << prefix << "n_boot_rep = " << prefix << n_boot_rep << "\n";

   out << prefix << "boot_rng = " << prefix << boot_rng << "\n";

   out << prefix << "boot_seed = " << prefix << boot_seed << "\n";

   vsdblinetype_to_string(out_line_type, junk);

   out << prefix << "out_line_type = " << prefix << junk << "\n";

   out << prefix << "threshold ...\n";

   threshold.dump(out, depth + 1);

   out << prefix << "mask_grid ...\n";

   mask_grid.dump(out, depth + 1);

   out << prefix << "mask_poly ...\n";

   mask_poly.dump(out, depth + 1);

   if(init_beg > 0) {
      timestring(init_beg, junk);
      out << prefix << "Init Begin  = " << init_beg << "   ("
          << junk << ")\n";
   }

   if(init_end > 0) {
      timestring(init_end, junk);
      out << prefix << "Init End    = " << init_end << "   ("
          << junk << ")\n";
   }

   if(valid_beg > 0) {
      timestring(valid_beg, junk);
      out << prefix << "Valid Begin = " << valid_beg << "   ("
          << junk << ")\n";
   }

   if(valid_end > 0) {
      timestring(valid_end, junk);
      out << prefix << "Valid End   = " << valid_end << "   ("
          << junk << ")\n";
   }

   out.flush();

   return;
}

////////////////////////////////////////////////////////////////////////

int VSDBAnalysisJob::is_keeper(const VSDBLine & L) const {
   int j, h, found, n;

   //
   // model
   //
   if(model.n_elements() > 0) {
      if(!(model.has(L.model()))) return(0);
   }

   //
   // fcst hms
   //
   if(fcst_lead.n_elements() > 0) {

      n = fcst_lead.n_elements();
      h = L.fcst_lead();

      found = 0;
      for(j=0; j<n; ++j) {
         if(h == (int) fcst_lead[j]) {
            found = 1;
            break;
         }
      }
      if(!found) return(0);
   }

   //
   // initialization hour
   //
   if(init_hour.n_elements() > 0) {

      n = init_hour.n_elements();
      h = L.init_hour();

      found = 0;
      for(j=0; j<n; ++j) {
         if(h == (int) init_hour[j]) {
            found = 1;
            break;
         }
      }
      if(!found) return(0);
   }

   //
   // begin valid time
   //
   if((valid_beg > 0) && (L.valid_time() < valid_beg)) return(0);

   //
   // end valid time
   //
   if((valid_end > 0) && (L.valid_time() > valid_end)) return(0);

   //
   // begin init time
   //
   if((init_beg > 0) && (L.init_time() < init_beg)) return(0);

   //
   // end init time
   //
   if((init_end > 0) && (L.init_time() > init_end)) return(0);

   //
   // obtype
   //
   if(obtype.n_elements() > 0) {
      if(!(obtype.has(L.obtype()))) return(0);
   }

   //
   // vx_mask
   //
   if(mask.n_elements() > 0) {
      if(!(mask.has(L.vx_mask()))) return(0);
   }

   //
   // line_type
   //
   if(line_type.n_elements() > 0) {
      if(!(line_type.has(L.line_type()))) return(0);
   }

   //
   // var
   //
   if(var.n_elements() > 0) {
      if(!(var.has(L.var()))) return(0);
   }

   //
   // level
   //
   if(level.n_elements() > 0) {
      if(!(level.has(L.level()))) return(0);
   }

   //
   // interp_mthd
   //
   if(interp_mthd.n_elements() > 0) {
      if(!(interp_mthd.has(L.interp_mthd()))) return(0);
   }

   //
   // interp_pnts
   //
   if(interp_pnts.n_elements() > 0) {

      n = interp_pnts.n_elements();
      h = L.interp_pnts();

      found = 0;
      for(j=0; j<n; ++j) {
         if(h == (int) interp_pnts[j]) {
            found = 1;
            break;
         }
      }

      if(!found) return(0);

   }

   return(1);
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::parse_job_command(const char *jobstring) {
   char tmp_str[1024];
   char *ptr;
   int i, k;

   // Job Command Line Array
   StringArray jc_array;

   //
   // Create a temporary copy of the jobstring for use in parsing
   //
   strcpy(tmp_str, jobstring);

   //
   // Parse the command line entries into a StringArray object
   //
   ptr = strtok(tmp_str, " ");
   while(ptr != NULL) {
      jc_array.add(ptr);
      ptr = strtok(0, " ");
   }

   //
   // If command line switches are present, clear out the values
   // already specified in the job for that option.
   //
   for(i=0; i<jc_array.n_elements(); i++) {

      if(     strcmp(jc_array[i], "-model") == 0)         model.clear();
      else if(strcmp(jc_array[i], "-fcst_lead") == 0)     fcst_lead.clear();
      else if(strcmp(jc_array[i], "-init_hour") == 0)     init_hour.clear();
      else if(strcmp(jc_array[i], "-obtype") == 0)        obtype.clear();
      else if(strcmp(jc_array[i], "-vx_mask") == 0)       mask.clear();
      else if(strcmp(jc_array[i], "-line_type") == 0)     line_type.clear();
      else if(strcmp(jc_array[i], "-var") == 0)           var.clear();
      else if(strcmp(jc_array[i], "-level") == 0)         level.clear();
      else if(strcmp(jc_array[i], "-interp_mthd") == 0)   interp_mthd.clear();
      else if(strcmp(jc_array[i], "-interp_pnts") == 0)   interp_pnts.clear();
      else if(strcmp(jc_array[i], "-dump_row") == 0)      dump_row.clear();
      else if(strcmp(jc_array[i], "-threshold") == 0)     threshold.clear();
      else if(strcmp(jc_array[i], "-mask_grid") == 0)     mask_grid.clear();
      else if(strcmp(jc_array[i], "-mask_poly") == 0)     mask_poly.clear();
   }

   //
   // Parse the command line and set the options
   //
   for(i=0; i<jc_array.n_elements(); i++) {

      //
      // Parse the job command line switches requiring one argument.
      // Check that the switch is followed by an argument and
      // that the argument is not a switch itself.
      //
      if(i+1 >= jc_array.n_elements() ||
         (jc_array[i+1])[0] == '-') {

         cerr << "\n\nERROR: parse_job_command() -> "
              << "error processing the job command line switch \""
              << jc_array[i] << "\" in the job command line: "
              << jobstring << "\n\n" << flush;

         throw(1);
      }

      if(strcmp(jc_array[i], "-job") == 0) {

         if(set_job_type(jc_array[i+1]) != 0) {
            cerr << "\n\nERROR: parse_job_command() -> "
                 << "unrecognized job type specified \"" << jc_array[i]
                 << "\" in job command line: " << jobstring << "\n\n"
                 << flush;

            throw(1);
         }
         i++;
      }
      else if(strcmp(jc_array[i], "-model") == 0) {
         model.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-fcst_lead") == 0) {
         k = timestring_to_sec(jc_array[i+1]);
         fcst_lead.add(k);
         i++;
      }
      else if(strcmp(jc_array[i], "-valid_beg") == 0) {
         valid_beg = timestring_to_unix(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-valid_end") == 0) {
         valid_end = timestring_to_unix(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-init_beg") == 0) {
         init_beg = timestring_to_unix(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-init_end") == 0) {
         init_end = timestring_to_unix(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-init_hour") == 0) {
         k = timestring_to_sec(jc_array[i+1]);
         init_hour.add(k);
         i++;
      }
      else if(strcmp(jc_array[i], "-obtype") == 0) {
         obtype.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-vx_mask") == 0) {
         mask.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-line_type") == 0) {
         line_type.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-var") == 0) {
         var.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-level") == 0) {
         level.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-interp_mthd") == 0) {
         interp_mthd.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-interp_pnts") == 0) {
         interp_pnts.add(atoi(jc_array[i+1]));
         i++;
      }
      else if(strcmp(jc_array[i], "-dump_row") == 0) {
         dump_row.add(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-column") == 0) {
         set_col_name(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-alpha") == 0) {
         alpha = atof(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-boot_interval") == 0) {
         boot_interval = atoi(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-boot_rep_prop") == 0) {
         boot_rep_prop = atof(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-n_boot_rep") == 0) {
         n_boot_rep = atoi(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-boot_rng") == 0) {
         set_boot_rng(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-boot_seed") == 0) {
         set_boot_seed(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-out_line_type") == 0) {
         out_line_type = determine_line_type(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-threshold") == 0) {

         if(threshold.n_elements() != 0) {
            cerr << "\n\nERROR: parse_job_command() -> "
                 << "-threshold may only be specified once "
                 << " in job command line: " << jobstring << "\n\n"
                 << flush;

            throw(1);
         }

         threshold.add(jc_array[i+1]);
         thresh.clear();
         thresh.set(jc_array[i+1]);
         i++;
      }
      else if(strcmp(jc_array[i], "-mask_grid") == 0) {

         if(mask_grid.n_elements() != 0) {
            cerr << "\n\nERROR: parse_job_command() -> "
                 << "-mask_grid may only be specified once "
                 << " in job command line: " << jobstring << "\n\n"
                 << flush;

            throw(1);
         }

         mask_grid.add(jc_array[i+1]);
         find_grid(jc_array[i+1], grid);
         i++;
      }
      else if(strcmp(jc_array[i], "-mask_poly") == 0) {

         if(mask_poly.n_elements() != 0) {
            cerr << "\n\nERROR: parse_job_command() -> "
                 << "-mask_poly may only be specified once "
                 << " in job command line: " << jobstring << "\n\n"
                 << flush;

            throw(1);
         }

         mask_poly.add(jc_array[i+1]);
         poly.clear();
         parse_latlon_poly_file(jc_array[i+1], poly);
         i++;
      }
      else {
         cerr << "\n\nERROR: parse_job_command() -> "
              << "unrecognized switch \"" << jc_array[i]
              << "\" in job command line: "
              << jobstring << "\n\n" << flush;

         throw(1);
      } // end if

   } // end for

   return;
}

////////////////////////////////////////////////////////////////////////

int VSDBAnalysisJob::set_job_type(const char *c) {
   char junk[512];
   int i, found;

   found = 0;
   for(i=0; i<n_vsdbjobtypes; i++) {

      vsdbjobtype_to_string((VSDBJobType) i, junk);

      if(strcmp(junk, c) == 0) {
         job_type = (VSDBJobType) i;
         found = 1;
         break;
      }
   }

   if(found) return(0);
   else      return(1);
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::set_col_name(const char *c) {

   if(col_name) { delete [] col_name; col_name = (char *) 0; }

   if(!c) return;

   col_name = new char [strlen(c) + 1];

   strcpy(col_name, c);

   return;
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::set_boot_rng(const char *c) {

   if(boot_rng) { delete [] boot_rng; boot_rng = (char *) 0; }

   if(!c) return;

   boot_rng = new char [strlen(c) + 1];

   strcpy(boot_rng, c);

   return;
}

////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::set_boot_seed(const char *c) {

   if(boot_seed) { delete [] boot_seed; boot_seed = (char *) 0; }

   if(!c) return;

   boot_seed = new char [strlen(c) + 1];

   strcpy(boot_seed, c);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The pointer passed in must contain enough space to hold the contents
// of the jobstring to be computed.
//
////////////////////////////////////////////////////////////////////////

void VSDBAnalysisJob::get_jobstring(char *js) {
   int i;
   char junk[512];

   // Initialize the jobstring
   strcpy(js, "");

   // Job type
   vsdbjobtype_to_string(job_type, junk);
   sprintf(js, "-job %s", junk);

   // Model Name
   if(model.n_elements() > 0) {
      for(i=0; i<model.n_elements(); i++)
         sprintf(js, "%s -model %s", js, model[i]);
   }

   // Forecast Lead Time
   if(fcst_lead.n_elements() > 0) {
      for(i=0; i<fcst_lead.n_elements(); i++) {
         sec_to_hhmmss(nint(fcst_lead[i]), junk);
         sprintf(js, "%s -fcst_lead %s", js, junk);
      }
   }

   // Forecast Begin and End Valid Time
   if(valid_beg > 0) {
      unix_to_yyyymmdd_hhmmss(valid_beg, junk);
      sprintf(js, "%s -valid_beg %s", js, junk);
   }
   if(valid_end > 0) {
      unix_to_yyyymmdd_hhmmss(valid_end, junk);
      sprintf(js, "%s -valid_end %s", js, junk);
   }

   // Forecast Begin and End Init Time
   if(init_beg > 0) {
      unix_to_yyyymmdd_hhmmss(init_beg, junk);
      sprintf(js, "%s -init_beg %s", js, junk);
   }
   if(init_end > 0) {
      unix_to_yyyymmdd_hhmmss(init_end, junk);
      sprintf(js, "%s -init_end %s", js, junk);
   }

   // Forecast Init Hour
   if(init_hour.n_elements() > 0) {
      for(i=0; i<init_hour.n_elements(); i++) {
         sec_to_hhmmss(nint(init_hour[i]), junk);
         sprintf(js, "%s -init_hour %s", js, junk);
      }
   }

   // Obtype
   if(obtype.n_elements() > 0) {
      for(i=0; i<obtype.n_elements(); i++)
         sprintf(js, "%s -obtype %s", js, obtype[i]);
   }

   // Verification Masking Region
   if(mask.n_elements() > 0) {
      for(i=0; i<mask.n_elements(); i++)
         sprintf(js, "%s -vx_mask %s", js, mask[i]);
   }

   // VSDB Line Type
   if(line_type.n_elements() > 0) {
      for(i=0; i<line_type.n_elements(); i++)
         sprintf(js, "%s -line_type %s", js, line_type[i]);
   }

   // Variable Type
   if(var.n_elements() > 0) {
      for(i=0; i<var.n_elements(); i++)
         sprintf(js, "%s -var %s", js, var[i]);
   }

   // Level Info
   if(level.n_elements() > 0) {
      for(i=0; i<level.n_elements(); i++)
         sprintf(js, "%s -level %s", js, level[i]);
   }

   // Interpolation Method
   if(interp_mthd.n_elements() > 0) {
      for(i=0; i<interp_mthd.n_elements(); i++)
         sprintf(js, "%s -interp_mthd %s", js, interp_mthd[i]);
   }

   // Interpolation Points
   if(interp_pnts.n_elements() > 0) {
      for(i=0; i<interp_pnts.n_elements(); i++)
         sprintf(js, "%s -interp_pnts %i", js, nint(interp_pnts[i]));
   }

   // Column Name
   if(col_name) sprintf(js, "%s -column %s", js, col_name);

   // Dump Row Option
   if(dump_row.n_elements() > 0) {
      for(i=0; i<dump_row.n_elements(); i++)
         sprintf(js, "%s -dump_row %s", js, dump_row[i]);
   }

   if(job_type == vsdb_job_summary ||
      (job_type == vsdb_job_aggr_mpr &&
       (out_line_type == vsdb_cnt || out_line_type == vsdb_cts))
     ) {
      // Alpha value
      sprintf(js, "%s -alpha %.2f", js, alpha);

      // Bootstrap Information
      sprintf(js, "%s -boot_interval %i", js, boot_interval);
      sprintf(js, "%s -boot_rep_prop %.2f", js, boot_rep_prop);
      sprintf(js, "%s -n_boot_rep %i", js, n_boot_rep);
      sprintf(js, "%s -boot_rng %s", js, boot_rng);
      sprintf(js, "%s -boot_seed %s", js, boot_seed);
   }

   // Out Line Type Option
   if(out_line_type != no_vsdb_line_type) {
      vsdblinetype_to_string(out_line_type, junk);
      sprintf(js, "%s -out_line_type %s", js, junk);
   }

   // Threshold Option
   if(threshold.n_elements() > 0) {
      for(i=0; i<threshold.n_elements(); i++)
         sprintf(js, "%s -threshold %s", js, threshold[i]);
   }

   // Mask Grid Option
   if(mask_grid.n_elements() > 0) {
      for(i=0; i<mask_grid.n_elements(); i++)
         sprintf(js, "%s -mask_grid %s", js, mask_grid[i]);
   }

   // Mask Poly Option
   if(mask_poly.n_elements() > 0) {
      for(i=0; i<mask_poly.n_elements(); i++)
         sprintf(js, "%s -mask_poly %s", js, mask_poly[i]);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

int VSDBAnalysisJob::is_in_grid_mask(double lat, double lon) const {
   int r = 1;
   double x, y;

   //
   // Only check if a masking grid has been specified
   //
   if(mask_grid.n_elements() > 0) {

      grid.latlon_to_xy(lat, lon, x, y);

      //
      // Convert lat/lon to grid x/y and check to see if it's on the
      // grid.
      //
      if(x < 0 || x > grid.nx() ||
         y < 0 || y > grid.ny()) r = 0;
   }

   return(r);
}

////////////////////////////////////////////////////////////////////////

int VSDBAnalysisJob::is_in_poly_mask(double lat, double lon) const {
   int r = 1;

   //
   // Only check if a masking poly has been specified
   //
   if(mask_poly.n_elements() > 0) {
      r = poly.is_inside(lat, lon);
   }

   return(r);
}

////////////////////////////////////////////////////////////////////////
//
// Code for misc functions
//
////////////////////////////////////////////////////////////////////////

void timestring(const unixtime t, char * out) {
   int month, day, year, hour, minute, second;

   unix_to_mdyhms(t, month, day, year, hour, minute, second);

   sprintf(out, "%s %d, %d   %02d:%02d:%02d",
           short_month_name[month], day, year,
           hour, minute, second);

   return;
}

////////////////////////////////////////////////////////////////////////
