// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*


///////////////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "vx_met_util/read_pcp_combine_netcdf.h"

///////////////////////////////////////////////////////////////////////////////

void read_pcp_combine_netcdf(NcFile *f_in, const char *var_name, char *lvl_name,
                             WrfData &wd, Grid &gr, int verbosity,
                             PlateCarreeData &pc_data, LambertData &lc_data,
                             StereographicData &st_data) {
   double v, v_min, v_max;
   int x, y, accum;
   unixtime init_ut, valid_ut;
   
   float *pcp_data = (float *) 0;

   NcAtt *proj_att = (NcAtt *) 0;
   NcVar *pcp_var  = (NcVar *) 0;
   NcAtt *lvl_att  = (NcAtt *) 0;
   NcAtt *time_att = (NcAtt *) 0;
   
   //
   // Parse the grid specification out of the global attributes
   //
   proj_att = f_in->get_att("Projection");
   if(!proj_att) {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "\"Projection\" attribute not found.\n\n" 
           << flush;
      exit(1);
   }

   //
   // Parse out the grid specification depending on the projection type
   // The following 4 Projection types are supported:
   //    - Lat/Lon
   //    - Mercator
   //    - Lambert Conformal
   //    - Polar Stereographic
   //
   
   //
   // Latitude/Longitude Projections Grid 
   // Also called Equidistant Cylindrical or Plate Carree Projection Grid
   //
   if(strcmp(proj_att->as_string(0), proj_type[0]) == 0) {
      
      if(verbosity > 3) {
         cout << "It's a Lat/Lon (PlateCarree or Equidistant Cylindrical) grid...\n" 
              << flush;
      }

      // Store the grid name
      pc_data.name = proj_type[0];
      
      // Latitude of the bottom left corner
      proj_att = f_in->get_att("lat_ll_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"lat_ll_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      pc_data.lat_ll_deg = atof(proj_att->as_string(0));
      
      // Longitude of the bottom left corner
      proj_att = f_in->get_att("lon_ll_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"lon_ll_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      pc_data.lon_ll_deg = -1.0*atof(proj_att->as_string(0));

      // Latitudinal increment
      proj_att = f_in->get_att("delta_lat_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"delta_lat_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      pc_data.delta_lat_deg = atof(proj_att->as_string(0));
      
      // Longitudinal increment
      proj_att = f_in->get_att("delta_lon_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"delta_lon_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      pc_data.delta_lon_deg = atof(proj_att->as_string(0));      

      // Number of points in the Latitudinal (y) direction
      proj_att = f_in->get_att("Nlat");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"Nlat\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      pc_data.Nlat = atoi(proj_att->as_string(0));
      
      // Number of points in the Longitudinal (x) direction
      proj_att = f_in->get_att("Nlon");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"Nlon\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      pc_data.Nlon = atoi(proj_att->as_string(0));      

      Grid pcp_grid(pc_data);
      gr = pcp_grid;

      if(verbosity > 3) {
         cout << "Latitude/Longitude Data:\n"
              << "lat_ll_deg = " << pc_data.lat_ll_deg << "\n"
              << "lon_ll_deg = " << pc_data.lon_ll_deg << "\n"
              << "delta_lat_deg = " << pc_data.delta_lat_deg << "\n"
              << "delta_lon_deg = " << pc_data.delta_lon_deg << "\n"
              << "Nlat = " << pc_data.Nlat << "\n"
              << "Nlon = " << pc_data.Nlon << "\n" << flush;
      }
   }
   //
   // Mercator Projection
   //
   else if(strcmp(proj_att->as_string(0), proj_type[1]) == 0) {
      
      if(verbosity > 3) {
         cout << "It's a Mercator grid...\n" << flush;
      }
      
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "Mercator projection type not currently supported.\n\n"
           << flush;
      exit(1);
   }       
   //
   // Lambert Conformal Projection
   //
   else if(strcmp(proj_att->as_string(0), proj_type[2]) == 0) {
   
      if(verbosity > 3) {
         cout << "It's a Lambert Conformal grid...\n" << flush;
      }

      // Copy the grid name
      lc_data.name = proj_type[2];
      
      // First latitude
      proj_att = f_in->get_att("p1_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"p1_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.p1_deg = atof(proj_att->as_string(0));
      
      // Second latitude
      proj_att = f_in->get_att("p2_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"p2_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.p2_deg = atof(proj_att->as_string(0));

      // Latitude of first point
      proj_att = f_in->get_att("p0_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"p0_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.p0_deg = atof(proj_att->as_string(0));

      // Longitude of first point
      proj_att = f_in->get_att("l0_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"l0_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.l0_deg = -1.0*atof(proj_att->as_string(0));

      // Center longitude
      proj_att = f_in->get_att("lcen_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"lcen_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.lcen_deg = -1.0*atof(proj_att->as_string(0));

      // Grid spacing in km
      proj_att = f_in->get_att("d_km");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"d_km\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.d_km = atof(proj_att->as_string(0));

      // Radius of the earth
      proj_att = f_in->get_att("r_km");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"r_km\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.r_km = atof(proj_att->as_string(0));

      // Number of points in the x-direction
      proj_att = f_in->get_att("nx");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"nx\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.nx = atoi(proj_att->as_string(0));

      // Number of points in the y-direction
      proj_att = f_in->get_att("ny");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"ny\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      lc_data.ny = atoi(proj_att->as_string(0));

      Grid pcp_grid(lc_data);
      gr = pcp_grid;

      if(verbosity > 3) {
         cout << "Lambert Conformal Data:\n"
              << "p1_deg = " << lc_data.p1_deg << "\n"
              << "p2_deg = " << lc_data.p2_deg << "\n"
              << "p0_deg = " << lc_data.p0_deg << "\n"
              << "l0_deg = " << lc_data.l0_deg << "\n"
              << "lcen_deg = " << lc_data.lcen_deg << "\n"
              << "d_km = " << lc_data.d_km << "\n"
              << "r_km = " << lc_data.r_km << "\n"
              << "nx = " << lc_data.nx << "\n"
              << "ny = " << lc_data.ny << "\n" << flush;
      }
   }   
   //
   // Polar Stereographic Projection
   //
   else if(strcmp(proj_att->as_string(0), proj_type[3]) == 0) {
     if(verbosity > 3) {
         cout << "It's a Polar Stereographic grid...\n" << flush;
      }
      
      // Copy the grid name
      st_data.name = proj_type[3];

      // First latitude
      proj_att = f_in->get_att("p1_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"p1_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.p1_deg = atof(proj_att->as_string(0));

      // Latitude of first point
      proj_att = f_in->get_att("p0_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"p0_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.p0_deg = atof(proj_att->as_string(0));

      // Longitude of first point
      proj_att = f_in->get_att("l0_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"l0_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.l0_deg = -1.0*atof(proj_att->as_string(0));

      // Center longitude
      proj_att = f_in->get_att("lcen_deg");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"lcen_deg\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.lcen_deg = -1.0*atof(proj_att->as_string(0));

      // Grid spacing in km
      proj_att = f_in->get_att("d_km");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"d_km\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.d_km = atof(proj_att->as_string(0));

      // Radius of the earth
      proj_att = f_in->get_att("r_km");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"r_km\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.r_km = atof(proj_att->as_string(0));

      // Number of points in the x-direction
      proj_att = f_in->get_att("nx");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"nx\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.nx = atoi(proj_att->as_string(0));

      // Number of points in the y-direction
      proj_att = f_in->get_att("ny");
      if(!proj_att) {
         cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
              << "\"ny\" attribute not found.\n\n" 
              << flush;
         exit(1);
      }
      st_data.ny = atoi(proj_att->as_string(0));

      Grid pcp_grid(st_data);
      gr = pcp_grid;

      if(verbosity > 3) {
         cout << "Stereographic Data:\n"
              << "p1_deg = " << st_data.p1_deg << "\n"
              << "p0_deg = " << st_data.p0_deg << "\n"
              << "l0_deg = " << st_data.l0_deg << "\n"
              << "lcen_deg = " << st_data.lcen_deg << "\n"
              << "d_km = " << st_data.d_km << "\n"
              << "r_km = " << st_data.r_km << "\n"
              << "nx = " << st_data.nx << "\n"
              << "ny = " << st_data.ny << "\n" << flush;
      }      
   }  
   //
   // Unsupported projection type
   //
   else {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "Projection type " << proj_att->as_string(0)
           << " not currently supported.\n\n"
           << flush;
      exit(1);   
   }
    
   //
   // Access the precip variable in the NetCDF file
   // 
   pcp_var = f_in->get_var(var_name);
   if(!pcp_var || !pcp_var->is_valid()) {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "\"" << var_name << "\" variable not found.\n\n"
           << flush;
      exit(1);
   }

   //
   // Extract the level name from the precip variable
   //
   lvl_att = pcp_var->get_att("level");
   if(!lvl_att) {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "\"level\" attribute not found for the \""
           << var_name << "\" variable.\n\n" << flush;
      exit(1);
   }
   strcpy(lvl_name, lvl_att->as_string(0));
   
   //
   // Extract the valid, lead, and accumulation time attributes from
   // the precip variable
   //
   time_att = pcp_var->get_att("init_time_ut");
   if(!time_att) {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "\"init_time_ut\" attribute not found for the \""
           << var_name << "\" variable.\n\n" << flush;
      exit(1);
   }
   init_ut = atoi(time_att->as_string(0));

   time_att = pcp_var->get_att("valid_time_ut");
   if(!time_att) {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "\"valid_time_ut\" attribute not found for the \""
           << var_name << "\" variable.\n\n" << flush;
      exit(1);
   }
   valid_ut = atoi(time_att->as_string(0));

   time_att = pcp_var->get_att("accum_time");
   if(!time_att) {
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "\"accum_time\" attribute not found for the \""
           << var_name << "\" variable.\n\n" << flush;
      exit(1);
   }
   accum = sec_per_hour*atoi(time_att->as_string(0));
   
   //
   // Allocate space for the precip data
   //
   pcp_data = new float [gr.nx()*gr.ny()];
   
   if( !pcp_var->get(&pcp_data[0], gr.ny(), gr.nx()) ) {  
      cerr << "\n\nERROR: read_pcp_combine_netcdf() -> "
           << "error with the pcp_var->get\n\n" << flush;
      exit(1);
   }   

   //
   // Set up the size of the wrfdata object
   //      
   wd.set_size(gr.nx(), gr.ny());
   
   //
   // Read through the precip data to find the min/max values
   //
   v_min = 1.0e30;
   v_max = -1.0e30;
   
   for(x=0; x<gr.nx(); x++) {
      for(y=0; y<gr.ny(); y++) {
         v = (double) pcp_data[wd.two_to_one(x, y)];
         if(v != bad_data_double && v > v_max) v_max = v;
         if(v != bad_data_double && v < v_min) v_min = v;
      }
   }
   
   if(verbosity > 1) {
      cout << "NetCDF " << var_name << " Variable (min, max) = ("
           << v_min << ", " << v_max << ")\n" << flush;
   }   
   
   //
   // Set up the wrfdata object
   //
   wd.set_valid_time(valid_ut);
   wd.set_lead_time( (int) (valid_ut - init_ut));
   wd.set_accum_time(accum);
   wd.set_b(v_min);
   wd.set_m( (double) (v_max-v_min)/wrfdata_int_data_max);
   
   //
   // Parse the precip data into the wrfdata object
   //
   for(x=0; x<gr.nx(); x++) {
      for(y=0; y<gr.ny(); y++) {
         v = (double) pcp_data[wd.two_to_one(x, y)];
         wd.put_xy_int(wd.double_to_int(v), x, y);
      }
   }

   if(pcp_data) { delete pcp_data; pcp_data = (float *) 0; }

   return;
}

///////////////////////////////////////////////////////////////////////////////
