// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
////////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cstdio>
#include <cmath>

#include "vx_met_util/constants.h"
#include "vx_met_util/vsdb_columns.h"

////////////////////////////////////////////////////////////////////////

int get_column_offset(const char **arr, int n_cols, const char *col_name) {
   int i, offset, found;

   found = 0;

   //
   // Search the VSDB header columns first
   //
   for(i=0; i<n_header_columns; i++) {
      if(strcmp(hdr_columns[i], col_name) == 0) {
         found  = 1;
         offset = i;
         break;
      }
   }

   //
   // If not found, search the columns provided
   //
   if(!found) {
      for(i=0; i<n_cols; i++) {

         if(strcmp(arr[i], col_name) == 0) {
            found  = 1;
            offset = i+n_header_columns;
            break;
         }
      }
   }

   if(!found) {
      cerr << "\n\nERROR: get_column_offset() -> "
           << "no match found in the indicated array for the column name "
           << "specified: \"" << col_name << "\"\n\n" 
           << flush;

      throw(1);
   }

   return(offset);
}

////////////////////////////////////////////////////////////////////////
//
// For the following format and names routines, the character pointer
// must contain enough allocated memory to store the resulting string.
//
////////////////////////////////////////////////////////////////////////

void get_header_format(char *fmt_str) {

   // Setup the header format string
   sprintf(fmt_str, "%s",
      "%-7s %-5s %-9s %-15s %-6s %-12s %-30s %-9s %-12s %-1s");

   return;
}

////////////////////////////////////////////////////////////////////////

void get_header_names(char *hdr_str) {
   char fmt_str[max_str_len];

   // Retrieve the VSDB format string
   get_header_format(fmt_str);

   // Setup the VSDB header columns
   sprintf(hdr_str, fmt_str,
      hdr_columns[0],
      hdr_columns[1],
      hdr_columns[2],
      hdr_columns[3],
      hdr_columns[4],
      hdr_columns[5],
      hdr_columns[6],
      hdr_columns[7],
      hdr_columns[8],
      hdr_columns[9]);

   return;
}

////////////////////////////////////////////////////////////////////////

void get_column_format(const char **arr, int n_cols, char *fmt_str) {
   int i;

   //
   // Initialize the fmt_str
   //
   strcpy(fmt_str, "%-15s");

   //
   // Append an additional format string for each column
   //
   for(i=1; i<n_cols; i++) {
      sprintf(fmt_str, "%s %s", fmt_str, "%-15s");
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void get_column_names(const char **arr, int n_cols, char *col_str) {
   int i;

   //
   // Initialize the col_str
   //
   strcpy(col_str, "");
   sprintf(col_str, "%-15s", arr[0]);

   //
   // Append the additional column names
   //
   for(i=1; i<n_cols; i++) {
      sprintf(col_str, "%s %-15s", col_str, arr[i]);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void get_line_format(const char **arr, int n_cols, char *fmt_str) {
   char hdr_fmt[max_line_len], col_fmt[max_line_len];

   //
   // Get the header format string
   //
   get_header_format(hdr_fmt);

   //
   // Get the column format string
   //
   get_column_format(arr, n_cols, col_fmt);

   //
   // Append the two together
   //
   sprintf(fmt_str, "%s %s", hdr_fmt, col_fmt);

   return;
}

////////////////////////////////////////////////////////////////////////

void get_line_names(const char **arr, int n_cols, char *col_str) {
   char hdr_names[max_line_len], col_names[max_line_len];

   //
   // Get the header column names
   //
   get_header_names(hdr_names);

   //
   // Get the column format string
   //
   get_column_names(arr, n_cols, col_names);

   //
   // Append the two together
   //
   sprintf(col_str, "%s %s", hdr_names, col_names);

   return;
}

////////////////////////////////////////////////////////////////////////

void get_mthd_str(char *interp_mthd_str, int i_mthd) {

   if(i_mthd == bad_data_int) strcpy(interp_mthd_str, bad_data_str);
   else                       strcpy(interp_mthd_str, mthd_str[i_mthd]);

   return;
}

////////////////////////////////////////////////////////////////////////

void get_pnts_str(char *interp_pnts_str, int i_wdth) {

   if(i_wdth == bad_data_int) strcpy(interp_pnts_str, bad_data_str);
   else                       sprintf(interp_pnts_str, "%i",
                                      i_wdth*i_wdth);

   return;
}

////////////////////////////////////////////////////////////////////////

void build_header_str(char *str,
                      const char *model_str,   const int  lead_sec,
                      const unixtime valid_ut, const char *obtype_str,
                      const char *vx_mask_str, const char *type_str,
                      const char *var_str,     const char *level_str) {
   char fmt_str[max_str_len];
   char fcst_lead_str[max_str_len], fcst_valid_str[max_str_len];
   int yr, mon, day, hr, min, sec;

   //
   // VSDB Header columns:
   //    VRS,         MODEL,       FCST_LEAD,
   //    FCST_VALID,  OBTYPE,      VX_MASK,
   //    TYPE,        VAR,         LEVEL,
   //    =
   //

   // Convert lead time to a string
   sec_to_hms(lead_sec, hr, min, sec);
   sprintf(fcst_lead_str, "%.2i%.2i%.2i", hr, min, sec);

   // Convert valid time to a string
   unix_to_mdyhms(valid_ut, mon, day, yr, hr, min, sec);
   sprintf(fcst_valid_str, "%.4i%.2i%.2i_%.2i%.2i%.2i",
      yr, mon, day, hr, min, sec);

   //
   // Setup the format string
   //
   get_header_format(fmt_str);

   //
   // Build the VSDB columns
   //
   sprintf(str, fmt_str,
      met_version,       // MET Version
      model_str,         // Model Name
      fcst_lead_str,     // Lead Time HHMMSS
      fcst_valid_str,    // Valid Time YYYYMMDD_HHMMSS
      obtype_str,        // Verifying Observation Type
      vx_mask_str,       // Verification masking region
      type_str,          // Threshold applied
      var_str,           // Grib field name
      level_str,         // Grib level name
      hdr_columns[9]     // Equal sign
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_sl1l2_str(char *str, SL1L2Info &sl1l2_info) {
   char var_str[n_sl1l2_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, sl1l2_info.interp_mthd);
   get_pnts_str(interp_pnts_str, sl1l2_info.interp_wdth);

   //
   // Scalar L1L2 Line Type (SL1L2)
   // Dump out the SL1L2 VSDB line:
   //    TOTAL,       FBAR,        OBAR,
   //    FOBAR,       FFBAR,       OOBAR,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",     sl1l2_info.scount);
   dbl_to_str(sl1l2_info.fbar,   var_str[1]);
   dbl_to_str(sl1l2_info.obar,   var_str[2]);
   dbl_to_str(sl1l2_info.fobar,  var_str[3]);
   dbl_to_str(sl1l2_info.ffbar,  var_str[4]);
   dbl_to_str(sl1l2_info.oobar,  var_str[5]);

   //
   // Setup the format string
   //
   get_column_format(sl1l2_columns, n_sl1l2_columns, fmt_str);

   //
   // Build the SL1L2 line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // fbar
      var_str[2],         // obar
      var_str[3],         // fobar
      var_str[4],         // ffbar
      var_str[5],         // oobar
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_sl1l2_str(char *str, CNTInfo &cnt_info) {
   char var_str[n_sl1l2_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cnt_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cnt_info.interp_wdth);

   //
   // Scalar L1L2 Line Type (SL1L2)
   // Dump out the SL1L2 VSDB line:
   //    TOTAL,       FBAR,        OBAR,
   //    FOBAR,       FFBAR,       OOBAR,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",   cnt_info.n);
   dbl_to_str(cnt_info.fbar.v, var_str[1]);
   dbl_to_str(cnt_info.obar.v, var_str[2]);
   dbl_to_str(cnt_info.fobar,  var_str[3]);
   dbl_to_str(cnt_info.ffbar,  var_str[4]);
   dbl_to_str(cnt_info.oobar,  var_str[5]);

   //
   // Setup the format string
   //
   get_column_format(sl1l2_columns, n_sl1l2_columns, fmt_str);

   //
   // Build the SL1L2 line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // fbar
      var_str[2],         // obar
      var_str[3],         // fobar
      var_str[4],         // ffbar
      var_str[5],         // oobar
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_sal1l2_str(char *str, SL1L2Info &sl1l2_info) {
   char var_str[n_sal1l2_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, sl1l2_info.interp_mthd);
   get_pnts_str(interp_pnts_str, sl1l2_info.interp_wdth);

   //
   // Scalar Anomaly L1L2 Line Type (SAL1L2)
   // Dump out the SL1L2 VSDB line:
   //    TOTAL,       FABAR,       OABAR,
   //    FOABAR,      FFABAR,      OOABAR,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",      sl1l2_info.sacount);
   dbl_to_str(sl1l2_info.fabar,   var_str[1]);
   dbl_to_str(sl1l2_info.oabar,   var_str[2]);
   dbl_to_str(sl1l2_info.foabar,  var_str[3]);
   dbl_to_str(sl1l2_info.ffabar,  var_str[4]);
   dbl_to_str(sl1l2_info.ooabar,  var_str[5]);

   //
   // Setup the format string
   //
   get_column_format(sal1l2_columns, n_sal1l2_columns, fmt_str);

   //
   // Build the SAL1L2 line
   //
   sprintf(str, fmt_str,
         var_str[0],      // Total anomaly count
         var_str[1],      // fabar
         var_str[2],      // oabar
         var_str[3],      // foabar
         var_str[4],      // ffabar
         var_str[5],      // ooabar
         interp_mthd_str, // Interpolation method
         interp_pnts_str  // Interpolation points
      );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_vl1l2_str(char *str, VL1L2Info &vl1l2_info) {
   char var_str[n_vl1l2_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, vl1l2_info.interp_mthd);
   get_pnts_str(interp_pnts_str, vl1l2_info.interp_wdth);

   //
   // Vector L1L2 Line Type (VL1L2)
   // Dump out the VL1L2 VSDB line:
   //    TOTAL,       UFBAR,       VFBAR,
   //    UOBAR,       VOBAR,       UVFOBAR,
   //    UVFFBAR,     UVOOBAR,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",      vl1l2_info.vcount);
   dbl_to_str(vl1l2_info.ufbar,   var_str[1]);
   dbl_to_str(vl1l2_info.vfbar,   var_str[2]);
   dbl_to_str(vl1l2_info.uobar,   var_str[3]);
   dbl_to_str(vl1l2_info.vobar,   var_str[4]);
   dbl_to_str(vl1l2_info.uvfobar, var_str[5]);
   dbl_to_str(vl1l2_info.uvffbar, var_str[6]);
   dbl_to_str(vl1l2_info.uvoobar, var_str[7]);

   //
   // Setup the format string
   //
   get_column_format(vl1l2_columns, n_vl1l2_columns, fmt_str);

   //
   // Build the VL1L2 line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // ufbar
      var_str[2],         // vfbar
      var_str[3],         // uobar
      var_str[4],         // vobar
      var_str[5],         // uvfobar
      var_str[6],         // uvffbar
      var_str[7],         // uvoobar
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_val1l2_str(char *str, VL1L2Info &vl1l2_info) {
   char var_str[n_val1l2_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, vl1l2_info.interp_mthd);
   get_pnts_str(interp_pnts_str, vl1l2_info.interp_wdth);

   //
   // Vector Anomaly L1L2 Line Type (VAL1L2)
   // Dump out the VAL1L2 VSDB line:
   //    TOTAL,       UFABAR,      VFABAR,
   //    UOABAR,      VOABAR,      UVFOABAR,
   //    UVFFABAR,    UVOOABAR,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",       vl1l2_info.vacount);
   dbl_to_str(vl1l2_info.ufabar,   var_str[1]);
   dbl_to_str(vl1l2_info.vfabar,   var_str[2]);
   dbl_to_str(vl1l2_info.uoabar,   var_str[3]);
   dbl_to_str(vl1l2_info.voabar,   var_str[4]);
   dbl_to_str(vl1l2_info.uvfoabar, var_str[5]);
   dbl_to_str(vl1l2_info.uvffabar, var_str[6]);
   dbl_to_str(vl1l2_info.uvooabar, var_str[7]);

   //
   // Setup the format string
   //
   get_column_format(val1l2_columns, n_val1l2_columns, fmt_str);

   //
   // Build the VAL1L2 line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total anomaly count
      var_str[1],         // ufabar
      var_str[2],         // vfabar
      var_str[3],         // uoabar
      var_str[4],         // voabar
      var_str[5],         // uvfoabar
      var_str[6],         // uvffabar
      var_str[7],         // uvooabar
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_fho_str(char *str, CTSInfo &cts_info) {
   char var_str[n_fho_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];
   int  status;

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cts_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cts_info.interp_wdth);

   //
   // FHO
   // Dump out the FHO VSDB line:
   //    TOTAL,       F_RATE,      H_RATE,
   //    O_RATE,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",               cts_info.cts.n());
   dbl_to_str(cts_info.cts.f_rate(status), var_str[1]);
   dbl_to_str(cts_info.cts.h_rate(status), var_str[2]);
   dbl_to_str(cts_info.cts.o_rate(status), var_str[3]);

   //
   // Setup the format string
   //
   get_column_format(fho_columns, n_fho_columns, fmt_str);

   //
   // Build the FHO line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total number of points
      var_str[1],         // Forecast Rate = FY/N
      var_str[2],         // Hit Rate = FY_OY/N
      var_str[3],         // Observation = OY/N
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_ctc_str(char *str, CTSInfo &cts_info) {
   char var_str[n_ctc_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cts_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cts_info.interp_wdth);

   //
   // Contingency Table Counts
   // Dump out the CTC VSDB line:
   //    TOTAL,       FY_OY,       FY_ON,
   //    FN_OY,       FN_ON,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i", cts_info.cts.n());
   sprintf(var_str[1], "%i", cts_info.cts.fy_oy());
   sprintf(var_str[2], "%i", cts_info.cts.fy_on());
   sprintf(var_str[3], "%i", cts_info.cts.fn_oy());
   sprintf(var_str[4], "%i", cts_info.cts.fn_on());

   //
   // Setup the format string
   //
   get_column_format(ctc_columns, n_ctc_columns, fmt_str);

   //
   // Build the CTC line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // fy_oy count
      var_str[2],         // fy_on count
      var_str[3],         // fn_oy count
      var_str[4],         // fn_on count
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_ctp_str(char *str, CTSInfo &cts_info) {
   char var_str[n_ctp_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];
   int  status;

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cts_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cts_info.interp_wdth);

   //
   // Contingency Table Proportions
   // Dump out the CTP VSDB line:
   //    TOTAL,       FY_OY_TP,    FY_ON_TP,
   //    FN_OY_TP,    FN_ON_TP,    FY_TP,
   //    FN_TP,       OY_TP,       ON_TP,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",                 cts_info.cts.n());
   dbl_to_str(cts_info.cts.fy_oy_tp(status), var_str[1]);
   dbl_to_str(cts_info.cts.fy_on_tp(status), var_str[2]);
   dbl_to_str(cts_info.cts.fn_oy_tp(status), var_str[3]);
   dbl_to_str(cts_info.cts.fn_on_tp(status), var_str[4]);
   dbl_to_str(cts_info.cts.fy_tp(status),    var_str[5]);
   dbl_to_str(cts_info.cts.fn_tp(status),    var_str[6]);
   dbl_to_str(cts_info.cts.oy_tp(status),    var_str[7]);
   dbl_to_str(cts_info.cts.on_tp(status),    var_str[8]);

   //
   // Setup the format string
   //
   get_column_format(ctp_columns, n_ctp_columns, fmt_str);

   //
   // Build the CTP line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // fy_oy_tp
      var_str[2],         // fy_on_tp
      var_str[3],         // fn_oy_tp
      var_str[4],         // fn_on_tp
      var_str[5],         // fy_tp
      var_str[6],         // fn_tp
      var_str[7],         // oy_tp
      var_str[8],         // on_tp
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_cfp_str(char *str, CTSInfo &cts_info) {
   char var_str[n_cfp_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];
   int  status;

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cts_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cts_info.interp_wdth);

   //
   // Contingency Table Forecast Proportions
   // Dump out the CFP VSDB line:
   //    TOTAL,       FY_OY_FP,    FY_ON_FP,
   //    FN_OY_FP,    FN_ON_FP,    FY,
   //    FN,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",                 cts_info.cts.n());
   dbl_to_str(cts_info.cts.fy_oy_fp(status), var_str[1]);
   dbl_to_str(cts_info.cts.fy_on_fp(status), var_str[2]);
   dbl_to_str(cts_info.cts.fn_oy_fp(status), var_str[3]);
   dbl_to_str(cts_info.cts.fn_on_fp(status), var_str[4]);
   sprintf(var_str[5], "%i",                 cts_info.cts.fy());
   sprintf(var_str[6], "%i",                 cts_info.cts.fn());

   //
   // Setup the format string
   //
   get_column_format(cfp_columns, n_cfp_columns, fmt_str);

   //
   // Build the CFP line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // fy_oy_fp
      var_str[2],         // fy_on_fp
      var_str[3],         // fn_oy_fp
      var_str[4],         // fn_on_fp
      var_str[5],         // fy count
      var_str[6],         // fn count
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_cop_str(char *str, CTSInfo &cts_info) {
   char var_str[n_cop_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];
   int  status;

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cts_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cts_info.interp_wdth);

   //
   // Contingency Table Observation Proportions
   // Dump out the COP VSDB line:
   //    TOTAL,       FY_OY_OP,    FY_ON_OP,
   //    FN_OY_OP,    FN_ON_OP,    OY,
   //    ON,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",                 cts_info.cts.n());
   dbl_to_str(cts_info.cts.fy_oy_op(status), var_str[1]);
   dbl_to_str(cts_info.cts.fy_on_op(status), var_str[2]);
   dbl_to_str(cts_info.cts.fn_oy_op(status), var_str[3]);
   dbl_to_str(cts_info.cts.fn_on_op(status), var_str[4]);
   sprintf(var_str[5], "%i",                 cts_info.cts.oy());
   sprintf(var_str[6], "%i",                 cts_info.cts.on());

   //
   // Setup the format string
   //
   get_column_format(cop_columns, n_cop_columns, fmt_str);

   //
   // Build the COP line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total count
      var_str[1],         // fy_oy_op
      var_str[2],         // fy_on_op
      var_str[3],         // fn_oy_op
      var_str[4],         // fn_on_op
      var_str[5],         // oy count
      var_str[6],         // on count
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_cts_str(char *str, CTSInfo &cts_info, int i) {
   char var_str[n_cts_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cts_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cts_info.interp_wdth);

   //
   // Contingency Table Stats
   // Dump out the CTS VSDB line:
   //    TOTAL,
   //    BASER,       BASER_NCL,   BASER_NCU,   BASER_BCL,   BASER_BCU,
   //    FMEAN,       FMEAN_NCL,   FMEAN_NCU,   FMEAN_BCL,   FMEAN_BCU,
   //    ACC,         ACC_NCL,     ACC_NCU,     ACC_BCL,     ACC_BCU,
   //    FBIAS,       FBIAS_BCL,   FBIAS_BCU,
   //    PODY,        PODY_NCL,    PODY_NCU,    PODY_BCL,    PODY_BCU,
   //    PODN,        PODN_NCL,    PODN_NCU,    PODN_BCL,    PODN_BCU,
   //    POFD,        POFD_NCL,    POFD_NCU,    POFD_BCL,    POFD_BCU,
   //    FAR,         FAR_NCL,     FAR_NCU,     FAR_BCL,     FAR_BCU,
   //    CSI,         CSI_NCL,     CSI_NCU,     CSI_BCL,     CSI_BCU,
   //    GSS,         GSS_BCL,     GSS_BCU,
   //    HK,          HK_NCL,      HK_NCU,      HK_BCL,      HK_BCU,
   //    HSS,         HSS_BCL,     HSS_BCU,
   //    ODDS,        ODDS_NCL,    ODDS_NCU,    ODDS_BCL,    ODDS_BCU,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i", cts_info.cts.n());
   dbl_to_str(cts_info.baser.v,        var_str[1]);
   dbl_to_str(cts_info.baser.v_ncl[i], var_str[2]);
   dbl_to_str(cts_info.baser.v_ncu[i], var_str[3]);
   dbl_to_str(cts_info.baser.v_bcl[i], var_str[4]);
   dbl_to_str(cts_info.baser.v_bcu[i], var_str[5]);
   dbl_to_str(cts_info.fmean.v,        var_str[6]);
   dbl_to_str(cts_info.fmean.v_ncl[i], var_str[7]);
   dbl_to_str(cts_info.fmean.v_ncu[i], var_str[8]);
   dbl_to_str(cts_info.fmean.v_bcl[i], var_str[9]);
   dbl_to_str(cts_info.fmean.v_bcu[i], var_str[10]);
   dbl_to_str(cts_info.acc.v,          var_str[11]);
   dbl_to_str(cts_info.acc.v_ncl[i],   var_str[12]);
   dbl_to_str(cts_info.acc.v_ncu[i],   var_str[13]);
   dbl_to_str(cts_info.acc.v_bcl[i],   var_str[14]);
   dbl_to_str(cts_info.acc.v_bcu[i],   var_str[15]);
   dbl_to_str(cts_info.fbias.v,        var_str[16]);
   dbl_to_str(cts_info.fbias.v_bcl[i], var_str[17]);
   dbl_to_str(cts_info.fbias.v_bcu[i], var_str[18]);
   dbl_to_str(cts_info.pody.v,         var_str[19]);
   dbl_to_str(cts_info.pody.v_ncl[i],  var_str[20]);
   dbl_to_str(cts_info.pody.v_ncu[i],  var_str[21]);
   dbl_to_str(cts_info.pody.v_bcl[i],  var_str[22]);
   dbl_to_str(cts_info.pody.v_bcu[i],  var_str[23]);
   dbl_to_str(cts_info.podn.v,         var_str[24]);
   dbl_to_str(cts_info.podn.v_ncl[i],  var_str[25]);
   dbl_to_str(cts_info.podn.v_ncu[i],  var_str[26]);
   dbl_to_str(cts_info.podn.v_bcl[i],  var_str[27]);
   dbl_to_str(cts_info.podn.v_bcu[i],  var_str[28]);
   dbl_to_str(cts_info.pofd.v,         var_str[29]);
   dbl_to_str(cts_info.pofd.v_ncl[i],  var_str[30]);
   dbl_to_str(cts_info.pofd.v_ncu[i],  var_str[31]);
   dbl_to_str(cts_info.pofd.v_bcl[i],  var_str[32]);
   dbl_to_str(cts_info.pofd.v_bcu[i],  var_str[33]);
   dbl_to_str(cts_info.far.v,          var_str[34]);
   dbl_to_str(cts_info.far.v_ncl[i],   var_str[35]);
   dbl_to_str(cts_info.far.v_ncu[i],   var_str[36]);
   dbl_to_str(cts_info.far.v_bcl[i],   var_str[37]);
   dbl_to_str(cts_info.far.v_bcu[i],   var_str[38]);
   dbl_to_str(cts_info.csi.v,          var_str[39]);
   dbl_to_str(cts_info.csi.v_ncl[i],   var_str[40]);
   dbl_to_str(cts_info.csi.v_ncu[i],   var_str[41]);
   dbl_to_str(cts_info.csi.v_bcl[i],   var_str[42]);
   dbl_to_str(cts_info.csi.v_bcu[i],   var_str[43]);
   dbl_to_str(cts_info.gss.v,          var_str[44]);
   dbl_to_str(cts_info.gss.v_bcl[i],   var_str[45]);
   dbl_to_str(cts_info.gss.v_bcu[i],   var_str[46]);
   dbl_to_str(cts_info.hk.v,           var_str[47]);
   dbl_to_str(cts_info.hk.v_ncl[i],    var_str[48]);
   dbl_to_str(cts_info.hk.v_ncu[i],    var_str[49]);
   dbl_to_str(cts_info.hk.v_bcl[i],    var_str[50]);
   dbl_to_str(cts_info.hk.v_bcu[i],    var_str[51]);
   dbl_to_str(cts_info.hss.v,          var_str[52]);
   dbl_to_str(cts_info.hss.v_bcl[i],   var_str[53]);
   dbl_to_str(cts_info.hss.v_bcu[i],   var_str[54]);
   dbl_to_str(cts_info.odds.v,         var_str[55]);
   dbl_to_str(cts_info.odds.v_ncl[i],  var_str[56]);
   dbl_to_str(cts_info.odds.v_ncu[i],  var_str[57]);
   dbl_to_str(cts_info.odds.v_bcl[i],  var_str[58]);
   dbl_to_str(cts_info.odds.v_bcu[i],  var_str[59]);

   //
   // Setup the format string
   //
   get_column_format(cts_columns, n_cts_columns, fmt_str);

   //
   // Build the CTS line
   //
   sprintf(str, fmt_str,
      var_str[0],               // Total count
      var_str[1],               // Base Rate (oy_tp)
      var_str[2],  var_str[3],  // Base Rate (oy_tp) NCI
      var_str[4],  var_str[5],  // Base Rate (oy_tp) BCI
      var_str[6],               // Forecast Mean (fy_tp)
      var_str[7],  var_str[8],  // Forecast Mean (fy_tp) NCI
      var_str[9],  var_str[10], // Forecast Mean (fy_tp) BCI
      var_str[11],              // Accuracy
      var_str[12], var_str[13], // Accuracy NCI
      var_str[14], var_str[15], // Accuracy BCI
      var_str[16],              // Frequency Bias
      var_str[17], var_str[18], // Frequency Bias BCI
      var_str[19],              // POD Yes
      var_str[20], var_str[21], // POD Yes NCI
      var_str[22], var_str[23], // POD Yes BCI
      var_str[24],              // POD No
      var_str[25], var_str[26], // POD No NCI
      var_str[27], var_str[28], // POD No BCI
      var_str[29],              // POFD
      var_str[30], var_str[31], // POFD NCI
      var_str[32], var_str[33], // POFD BCI
      var_str[34],              // FAR
      var_str[35], var_str[36], // FAR NCI
      var_str[37], var_str[37], // FAR BCI
      var_str[39],              // CSI (Threat Score)
      var_str[40], var_str[41], // CSI (Threat Score) NCI
      var_str[42], var_str[43], // CSI (Threat Score) BCI
      var_str[44],              // Gilbert Skill Score (ETS)
      var_str[45], var_str[46], // Gilbert Skill Score (ETS) BCI
      var_str[47],              // Hanssen-Kuipers Discriminant (TSS)
      var_str[48], var_str[49], // Hanssen-Kuipers Discriminant (TSS) NCI
      var_str[50], var_str[51], // Hanssen-Kuipers Discriminant (TSS) BCI
      var_str[52],              // Heidke Skill Score
      var_str[53], var_str[54], // Heidke Skill Score BCI
      var_str[55],              // Odds Ratio
      var_str[56], var_str[57], // Odds Ratio NCI
      var_str[58], var_str[59], // Odds Ratio BCI
      interp_mthd_str,          // Interpolation method
      interp_pnts_str           // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_cnt_str(char *str, CNTInfo &cnt_info, int i) {
   char var_str[n_cnt_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, cnt_info.interp_mthd);
   get_pnts_str(interp_pnts_str, cnt_info.interp_wdth);

   //
   // Continuous Variable Stats
   // Dump out the CNT VSDB line:
   //    TOTAL,
   //    FBAR,        FBAR_NCL,    FBAR_NCU,    FBAR_BCL,    FBAR_BCU,
   //    FSTDEV,      FSTDEV_NCL,  FSTDEV_NCU,  FSTDEV_BCL,  FSTDEV_BCU,
   //    OBAR,        OBAR_NCL,    OBAR_NCU,    OBAR_BCL,    OBAR_BCU,
   //    OSTDEV,      OSTDEV_NCL,  OSTDEV_NCU,  OSTDEV_BCL,  OSTDEV_BCU,
   //    PR_CORR,     PR_CORR_NCL, PR_CORR_NCU, PR_CORR_BCL, PR_CORR_BCU,
   //    SP_CORR,
   //    KT_CORR,
   //    RANKS,
   //    FRANK_TIES,
   //    ORANK_TIES,
   //    ME,          ME_NCL,      ME_NCU,      ME_BCL,      ME_BCU,
   //    ESTDEV,      ESTDEV_NCL,  ESTDEV_NCU,  ESTDEV_BCL,  ESTDEV_BCU,
   //    MBIAS,       MBIAS_BCL,   MBIAS_BCU,
   //    MAE,         MAE_BCL,     MAE_BCU,
   //    MSE,         MSE_BCL,     MSE_BCU,
   //    BCMSE,       BCMSE_BCL,   BCMSE_BCU,
   //    RMSE,        RMSE_BCL,    RMSE_BCU,
   //    E10,         E10_BCL,     E10_BCU,
   //    E25,         E25_BCL,     E25_BCU,
   //    E50,         E50_BCL,     E50_BCU,
   //    E75,         E75_BCL,     E75_BCU,
   //    E90,         E90_BCL,     E90_BCU,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",             cnt_info.n);
   dbl_to_str(cnt_info.fbar.v,           var_str[1]);
   dbl_to_str(cnt_info.fbar.v_ncl[i],    var_str[2]);
   dbl_to_str(cnt_info.fbar.v_ncu[i],    var_str[3]);
   dbl_to_str(cnt_info.fbar.v_bcl[i],    var_str[4]);
   dbl_to_str(cnt_info.fbar.v_bcu[i],    var_str[5]);
   dbl_to_str(cnt_info.fstdev.v,         var_str[6]);
   dbl_to_str(cnt_info.fstdev.v_ncl[i],  var_str[7]);
   dbl_to_str(cnt_info.fstdev.v_ncu[i],  var_str[8]);
   dbl_to_str(cnt_info.fstdev.v_bcl[i],  var_str[9]);
   dbl_to_str(cnt_info.fstdev.v_bcu[i],  var_str[10]);
   dbl_to_str(cnt_info.obar.v,           var_str[11]);
   dbl_to_str(cnt_info.obar.v_ncl[i],    var_str[12]);
   dbl_to_str(cnt_info.obar.v_ncu[i],    var_str[13]);
   dbl_to_str(cnt_info.obar.v_bcl[i],    var_str[14]);
   dbl_to_str(cnt_info.obar.v_bcu[i],    var_str[15]);
   dbl_to_str(cnt_info.ostdev.v,         var_str[16]);
   dbl_to_str(cnt_info.ostdev.v_ncl[i],  var_str[17]);
   dbl_to_str(cnt_info.ostdev.v_ncu[i],  var_str[18]);
   dbl_to_str(cnt_info.ostdev.v_bcl[i],  var_str[19]);
   dbl_to_str(cnt_info.ostdev.v_bcu[i],  var_str[20]);
   dbl_to_str(cnt_info.pr_corr.v,        var_str[21]);
   dbl_to_str(cnt_info.pr_corr.v_ncl[i], var_str[22]);
   dbl_to_str(cnt_info.pr_corr.v_ncu[i], var_str[23]);
   dbl_to_str(cnt_info.pr_corr.v_bcl[i], var_str[24]);
   dbl_to_str(cnt_info.pr_corr.v_bcu[i], var_str[25]);
   dbl_to_str(cnt_info.sp_corr.v,        var_str[26]);
   dbl_to_str(cnt_info.kt_corr.v,        var_str[27]);
   sprintf(var_str[28], "%i",            cnt_info.n_ranks);
   sprintf(var_str[29], "%i",            cnt_info.frank_ties);
   sprintf(var_str[30], "%i",            cnt_info.orank_ties);
   dbl_to_str(cnt_info.me.v,             var_str[31]);
   dbl_to_str(cnt_info.me.v_ncl[i],      var_str[32]);
   dbl_to_str(cnt_info.me.v_ncu[i],      var_str[33]);
   dbl_to_str(cnt_info.me.v_bcl[i],      var_str[34]);
   dbl_to_str(cnt_info.me.v_bcu[i],      var_str[35]);
   dbl_to_str(cnt_info.estdev.v,         var_str[36]);
   dbl_to_str(cnt_info.estdev.v_ncl[i],  var_str[37]);
   dbl_to_str(cnt_info.estdev.v_ncu[i],  var_str[38]);
   dbl_to_str(cnt_info.estdev.v_bcl[i],  var_str[39]);
   dbl_to_str(cnt_info.estdev.v_bcu[i],  var_str[40]);
   dbl_to_str(cnt_info.mbias.v,          var_str[41]);
   dbl_to_str(cnt_info.mbias.v_bcl[i],   var_str[42]);
   dbl_to_str(cnt_info.mbias.v_bcu[i],   var_str[43]);
   dbl_to_str(cnt_info.mae.v,            var_str[44]);
   dbl_to_str(cnt_info.mae.v_bcl[i],     var_str[45]);
   dbl_to_str(cnt_info.mae.v_bcu[i],     var_str[46]);
   dbl_to_str(cnt_info.mse.v,            var_str[47]);
   dbl_to_str(cnt_info.mse.v_bcl[i],     var_str[48]);
   dbl_to_str(cnt_info.mse.v_bcu[i],     var_str[49]);
   dbl_to_str(cnt_info.bcmse.v,          var_str[50]);
   dbl_to_str(cnt_info.bcmse.v_bcl[i],   var_str[51]);
   dbl_to_str(cnt_info.bcmse.v_bcu[i],   var_str[52]);
   dbl_to_str(cnt_info.rmse.v,           var_str[53]);
   dbl_to_str(cnt_info.rmse.v_bcl[i],    var_str[54]);
   dbl_to_str(cnt_info.rmse.v_bcu[i],    var_str[55]);
   dbl_to_str(cnt_info.e10.v,            var_str[56]);
   dbl_to_str(cnt_info.e10.v_bcl[i],     var_str[57]);
   dbl_to_str(cnt_info.e10.v_bcu[i],     var_str[58]);
   dbl_to_str(cnt_info.e25.v,            var_str[59]);
   dbl_to_str(cnt_info.e25.v_bcl[i],     var_str[60]);
   dbl_to_str(cnt_info.e25.v_bcu[i],     var_str[61]);
   dbl_to_str(cnt_info.e50.v,            var_str[62]);
   dbl_to_str(cnt_info.e50.v_bcl[i],     var_str[63]);
   dbl_to_str(cnt_info.e50.v_bcu[i],     var_str[64]);
   dbl_to_str(cnt_info.e75.v,            var_str[65]);
   dbl_to_str(cnt_info.e75.v_bcl[i],     var_str[66]);
   dbl_to_str(cnt_info.e75.v_bcu[i],     var_str[67]);
   dbl_to_str(cnt_info.e90.v,            var_str[68]);
   dbl_to_str(cnt_info.e90.v_bcl[i],     var_str[69]);
   dbl_to_str(cnt_info.e90.v_bcu[i],     var_str[70]);

   //
   // Setup the format string
   //
   get_column_format(cnt_columns, n_cnt_columns, fmt_str);

   //
   // Build the CNT line
   //
   sprintf(str, fmt_str,
      var_str[0],               // Total Number of Grid Points
      var_str[1],               // Forecast Mean
      var_str[2],  var_str[3],  // Forecast Mean NCI
      var_str[4],  var_str[5],  // Forecast Mean BCI
      var_str[6],               // Forecast Standard Deviation
      var_str[7],  var_str[8],  // Forecast Standard Deviation NCI
      var_str[9],  var_str[10], // Forecast Standard Deviation BCI
      var_str[11],              // Observation Mean
      var_str[12], var_str[13], // Observation Mean NCI
      var_str[14], var_str[15], // Observation Mean BCI
      var_str[16],              // Obsevation Standard Deviation
      var_str[17], var_str[18], // Obsevation Standard Deviation NCI
      var_str[19], var_str[20], // Obsevation Standard Deviation BCI
      var_str[21],              // Pearson's Correlation Coefficient
      var_str[22], var_str[23], // Pearson's Correlation Coefficient NCI
      var_str[24], var_str[25], // Pearson's Correlation Coefficient BCI
      var_str[26],              // Spearman's Rank Correlation Coefficient
      var_str[27],              // Kendall Tau Rank Correlation Coefficient
      var_str[28],              // Number of ranks compared
      var_str[29],              // Number of tied forecast ranks
      var_str[30],              // Number of tied observation ranks
      var_str[31],              // Mean Error
      var_str[32], var_str[33], // Mean Error NCI
      var_str[34], var_str[35], // Mean Error BCI
      var_str[36],              // Error Standard Deviation
      var_str[37], var_str[38], // Error Standard Deviation NCI
      var_str[39], var_str[40], // Error Standard Deviation BCI
      var_str[41],              // Multiplicative Bias
      var_str[42], var_str[43], // Multiplicative Bias BCI
      var_str[44],              // Mean Absolute Error
      var_str[45], var_str[46], // Mean Absolute Error BCI
      var_str[47],              // Mean Squared Error
      var_str[48], var_str[49], // Mean Squared Error BCI
      var_str[50],              // Bias-Corrected Mean Squared Error
      var_str[51], var_str[52], // Bias-Corrected Mean Squared Error BCI
      var_str[53],              // Root Mean Squared Error
      var_str[54], var_str[55], // Root Mean Squared Error BCI
      var_str[56],              // 10th Percentile of the Error
      var_str[57], var_str[58], // 10th Percentile of the Error BCI
      var_str[59],              // 25th Percentile of the Error
      var_str[60], var_str[61], // 25th Percentile of the Error BCI
      var_str[62],              // 50th Percentile of the Error
      var_str[63], var_str[64], // 50th Percentile of the Error BCI
      var_str[65],              // 75th Percentile of the Error
      var_str[66], var_str[67], // 75th Percentile of the Error BCI
      var_str[68],              // 90th Percentile of the Error
      var_str[69], var_str[70], // 90th Percentile of the Error BCI
      interp_mthd_str,          // Interpolation method
      interp_pnts_str           // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_mpr_str(char *str, PairData *pd_ptr, int i) {
   char var_str[n_mpr_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_mthd_str[max_str_len], interp_pnts_str[max_str_len];

   //
   // Interpolation method and number of points
   //
   get_mthd_str(interp_mthd_str, pd_ptr->interp_mthd);
   get_pnts_str(interp_pnts_str, pd_ptr->interp_wdth);

   //
   // Matched Pair Line Type (MPR)
   // Dump out the MPR VSDB line:
   //    TOTAL,       INDEX,       OBS_LAT,
   //    OBS_LON,     OBS_LVL,     OBS_ELV,
   //    FCST,        OBS,         CLIMO,
   //    INTERP_MTHD, INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i", pd_ptr->n_pair);
   sprintf(var_str[1], "%i", i+1);
   dbl_to_str(pd_ptr->obs_lat[i], var_str[2]);
   dbl_to_str(pd_ptr->obs_lon[i], var_str[3]);
   dbl_to_str(pd_ptr->obs_lvl[i], var_str[4]);
   dbl_to_str(pd_ptr->obs_elv[i], var_str[5]);
   dbl_to_str(pd_ptr->fdata[i],   var_str[6]);
   dbl_to_str(pd_ptr->odata[i],   var_str[7]);
   dbl_to_str(pd_ptr->cdata[i],   var_str[8]);

   //
   // Setup the format string
   //
   get_column_format(mpr_columns, n_mpr_columns, fmt_str);

   //
   // Build the MPR line
   //
   sprintf(str, fmt_str,
      var_str[0],         // Total Number of matched pairs
      var_str[1],         // Index of current matched pair
      var_str[2],         // Observation Latitude (degrees north)
      var_str[3],         // Observation Longitude (degrees east)
      var_str[4],         // Observation Level information
      var_str[5],         // Observation Elevation
      var_str[6],         // Forecast value
      var_str[7],         // Observation value
      var_str[8],         // Climatological value
      interp_mthd_str,    // Interpolation method
      interp_pnts_str     // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_nbrctc_str(char *str, NBRCTSInfo &nbrcts_info) {
   char var_str[n_nbrcts_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_pnts_str[max_str_len];

   //
   // Number of points in the neighborhood
   //
   get_pnts_str(interp_pnts_str, nbrcts_info.nbr_wdth);

   //
   // Neighborhood Method Contingency Table Counts Line Type (NBRCTC)
   // Dump out the NBRCTC VSDB line:
   //    TOTAL,       FY_OY,       FY_ON,
   //    FN_OY,       FN_ON,
   //    INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i", nbrcts_info.cts_info.cts.n());
   sprintf(var_str[1], "%i", nbrcts_info.cts_info.cts.fy_oy());
   sprintf(var_str[2], "%i", nbrcts_info.cts_info.cts.fy_on());
   sprintf(var_str[3], "%i", nbrcts_info.cts_info.cts.fn_oy());
   sprintf(var_str[4], "%i", nbrcts_info.cts_info.cts.fn_on());

   //
   // Setup the format string
   //
   get_column_format(nbrctc_columns, n_nbrctc_columns, fmt_str);

   //
   // Build the NBRCTC line
   //
   sprintf(str, fmt_str,
      var_str[0],        // Total count
      var_str[1],        // fy_oy count
      var_str[2],        // fy_on count
      var_str[3],        // fn_oy count
      var_str[4],        // fn_on count
      interp_pnts_str    // Neighborhood points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_nbrcts_str(char *str, NBRCTSInfo &nbrcts_info, int i) {
   char var_str[n_nbrcts_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_pnts_str[max_str_len];

   //
   // Number of points in the neighborhood
   //
   get_pnts_str(interp_pnts_str, nbrcts_info.nbr_wdth);

   //
   // Neighborhood Method Contingency Table Statistics Line Type (NBRCTS)
   // Dump out the NBRCTS VSDB line:
   //    TOTAL,
   //    BASER,       BASER_NCL,   BASER_NCU,   BASER_BCL,   BASER_BCU,
   //    FMEAN,       FMEAN_NCL,   FMEAN_NCU,   FMEAN_BCL,   FMEAN_BCU,
   //    ACC,         ACC_NCL,     ACC_NCU,     ACC_BCL,     ACC_BCU,
   //    FBIAS,       FBIAS_BCL,   FBIAS_BCU,
   //    PODY,        PODY_NCL,    PODY_NCU,    PODY_BCL,    PODY_BCU,
   //    PODN,        PODN_NCL,    PODN_NCU,    PODN_BCL,    PODN_BCU,
   //    POFD,        POFD_NCL,    POFD_NCU,    POFD_BCL,    POFD_BCU,
   //    FAR,         FAR_NCL,     FAR_NCU,     FAR_BCL,     FAR_BCU,
   //    CSI,         CSI_NCL,     CSI_NCU,     CSI_BCL,     CSI_BCU,
   //    GSS,         GSS_BCL,     GSS_BCU,
   //    HK,          HK_NCL,      HK_NCU,      HK_BCL,      HK_BCU,
   //    HSS,         HSS_BCL,     HSS_BCU,
   //    ODDS,        ODDS_NCL,    ODDS_NCU,    ODDS_BCL,    ODDS_BCU,
   //    INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i", nbrcts_info.cts_info.cts.n());
   dbl_to_str(nbrcts_info.cts_info.baser.v,        var_str[1]);
   dbl_to_str(nbrcts_info.cts_info.baser.v_ncl[i], var_str[2]);
   dbl_to_str(nbrcts_info.cts_info.baser.v_ncu[i], var_str[3]);
   dbl_to_str(nbrcts_info.cts_info.baser.v_bcl[i], var_str[4]);
   dbl_to_str(nbrcts_info.cts_info.baser.v_bcu[i], var_str[5]);
   dbl_to_str(nbrcts_info.cts_info.fmean.v,        var_str[6]);
   dbl_to_str(nbrcts_info.cts_info.fmean.v_ncl[i], var_str[7]);
   dbl_to_str(nbrcts_info.cts_info.fmean.v_ncu[i], var_str[8]);
   dbl_to_str(nbrcts_info.cts_info.fmean.v_bcl[i], var_str[9]);
   dbl_to_str(nbrcts_info.cts_info.fmean.v_bcu[i], var_str[10]);
   dbl_to_str(nbrcts_info.cts_info.acc.v,          var_str[11]);
   dbl_to_str(nbrcts_info.cts_info.acc.v_ncl[i],   var_str[12]);
   dbl_to_str(nbrcts_info.cts_info.acc.v_ncu[i],   var_str[13]);
   dbl_to_str(nbrcts_info.cts_info.acc.v_bcl[i],   var_str[14]);
   dbl_to_str(nbrcts_info.cts_info.acc.v_bcu[i],   var_str[15]);
   dbl_to_str(nbrcts_info.cts_info.fbias.v,        var_str[16]);
   dbl_to_str(nbrcts_info.cts_info.fbias.v_bcl[i], var_str[17]);
   dbl_to_str(nbrcts_info.cts_info.fbias.v_bcu[i], var_str[18]);
   dbl_to_str(nbrcts_info.cts_info.pody.v,         var_str[19]);
   dbl_to_str(nbrcts_info.cts_info.pody.v_ncl[i],  var_str[20]);
   dbl_to_str(nbrcts_info.cts_info.pody.v_ncu[i],  var_str[21]);
   dbl_to_str(nbrcts_info.cts_info.pody.v_bcl[i],  var_str[22]);
   dbl_to_str(nbrcts_info.cts_info.pody.v_bcu[i],  var_str[23]);
   dbl_to_str(nbrcts_info.cts_info.podn.v,         var_str[24]);
   dbl_to_str(nbrcts_info.cts_info.podn.v_ncl[i],  var_str[25]);
   dbl_to_str(nbrcts_info.cts_info.podn.v_ncu[i],  var_str[26]);
   dbl_to_str(nbrcts_info.cts_info.podn.v_bcl[i],  var_str[27]);
   dbl_to_str(nbrcts_info.cts_info.podn.v_bcu[i],  var_str[28]);
   dbl_to_str(nbrcts_info.cts_info.pofd.v,         var_str[29]);
   dbl_to_str(nbrcts_info.cts_info.pofd.v_ncl[i],  var_str[30]);
   dbl_to_str(nbrcts_info.cts_info.pofd.v_ncu[i],  var_str[31]);
   dbl_to_str(nbrcts_info.cts_info.pofd.v_bcl[i],  var_str[32]);
   dbl_to_str(nbrcts_info.cts_info.pofd.v_bcu[i],  var_str[33]);
   dbl_to_str(nbrcts_info.cts_info.far.v,          var_str[34]);
   dbl_to_str(nbrcts_info.cts_info.far.v_ncl[i],   var_str[35]);
   dbl_to_str(nbrcts_info.cts_info.far.v_ncu[i],   var_str[36]);
   dbl_to_str(nbrcts_info.cts_info.far.v_bcl[i],   var_str[37]);
   dbl_to_str(nbrcts_info.cts_info.far.v_bcu[i],   var_str[38]);
   dbl_to_str(nbrcts_info.cts_info.csi.v,          var_str[39]);
   dbl_to_str(nbrcts_info.cts_info.csi.v_ncl[i],   var_str[40]);
   dbl_to_str(nbrcts_info.cts_info.csi.v_ncu[i],   var_str[41]);
   dbl_to_str(nbrcts_info.cts_info.csi.v_bcl[i],   var_str[42]);
   dbl_to_str(nbrcts_info.cts_info.csi.v_bcu[i],   var_str[43]);
   dbl_to_str(nbrcts_info.cts_info.gss.v,          var_str[44]);
   dbl_to_str(nbrcts_info.cts_info.gss.v_bcl[i],   var_str[45]);
   dbl_to_str(nbrcts_info.cts_info.gss.v_bcu[i],   var_str[46]);
   dbl_to_str(nbrcts_info.cts_info.hk.v,           var_str[47]);
   dbl_to_str(nbrcts_info.cts_info.hk.v_ncl[i],    var_str[48]);
   dbl_to_str(nbrcts_info.cts_info.hk.v_ncu[i],    var_str[49]);
   dbl_to_str(nbrcts_info.cts_info.hk.v_bcl[i],    var_str[50]);
   dbl_to_str(nbrcts_info.cts_info.hk.v_bcu[i],    var_str[51]);
   dbl_to_str(nbrcts_info.cts_info.hss.v,          var_str[52]);
   dbl_to_str(nbrcts_info.cts_info.hss.v_bcl[i],   var_str[53]);
   dbl_to_str(nbrcts_info.cts_info.hss.v_bcu[i],   var_str[54]);
   dbl_to_str(nbrcts_info.cts_info.odds.v,         var_str[55]);
   dbl_to_str(nbrcts_info.cts_info.odds.v_ncl[i],  var_str[56]);
   dbl_to_str(nbrcts_info.cts_info.odds.v_ncu[i],  var_str[57]);
   dbl_to_str(nbrcts_info.cts_info.odds.v_bcl[i],  var_str[58]);
   dbl_to_str(nbrcts_info.cts_info.odds.v_bcu[i],  var_str[59]);

   //
   // Setup the format string
   //
   get_column_format(nbrcts_columns, n_nbrcts_columns, fmt_str);

   //
   // Build the NBRCTS line
   //
   sprintf(str, fmt_str,
      var_str[0],               // Total count
      var_str[1],               // Base Rate (oy_tp)
      var_str[2],  var_str[3],  // Base Rate (oy_tp) NCI
      var_str[4],  var_str[5],  // Base Rate (oy_tp) BCI
      var_str[6],               // Forecast Mean (fy_tp)
      var_str[7],  var_str[8],  // Forecast Mean (fy_tp) NCI
      var_str[9],  var_str[10], // Forecast Mean (fy_tp) BCI
      var_str[11],              // Accuracy
      var_str[12], var_str[13], // Accuracy NCI
      var_str[14], var_str[15], // Accuracy BCI
      var_str[16],              // Frequency Bias
      var_str[17], var_str[18], // Frequency Bias BCI
      var_str[19],              // POD Yes
      var_str[20], var_str[21], // POD Yes NCI
      var_str[22], var_str[23], // POD Yes BCI
      var_str[24],              // POD No
      var_str[25], var_str[26], // POD No NCI
      var_str[27], var_str[28], // POD No BCI
      var_str[29],              // POFD
      var_str[30], var_str[31], // POFD NCI
      var_str[32], var_str[33], // POFD BCI
      var_str[34],              // FAR
      var_str[35], var_str[36], // FAR NCI
      var_str[37], var_str[37], // FAR BCI
      var_str[39],              // CSI (Threat Score)
      var_str[40], var_str[41], // CSI (Threat Score) NCI
      var_str[42], var_str[43], // CSI (Threat Score) BCI
      var_str[44],              // Gilbert Skill Score (ETS)
      var_str[45], var_str[46], // Gilbert Skill Score (ETS) BCI
      var_str[47],              // Hanssen-Kuipers Discriminant (TSS)
      var_str[48], var_str[49], // Hanssen-Kuipers Discriminant (TSS) NCI
      var_str[50], var_str[51], // Hanssen-Kuipers Discriminant (TSS) BCI
      var_str[52],              // Heidke Skill Score
      var_str[53], var_str[54], // Heidke Skill Score BCI
      var_str[55],              // Odds Ratio
      var_str[56], var_str[57], // Odds Ratio NCI
      var_str[58], var_str[59], // Odds Ratio BCI
      interp_pnts_str           // Interpolation points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_nbrcnt_str(char *str, NBRCNTInfo &nbrcnt_info, int i) {
   char var_str[n_nbrcnt_columns][vsdb_col_len];
   char fmt_str[max_str_len];
   char interp_pnts_str[max_str_len];

   //
   // Number of points in the neighborhood
   //
   get_pnts_str(interp_pnts_str, nbrcnt_info.nbr_wdth);

   //
   // Neighborhood Method Continuous Statistics (NBRCNT)
   // Dump out the NBRCNT VSDB line:
   //    TOTAL,
   //    FBS,         FBS_BCL,     FBS_BCU,
   //    FSS,         FSS_BCL,     FSS_BCU,
   //    INTERP_PNTS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i",  nbrcnt_info.cnt_info.n);
   dbl_to_str(nbrcnt_info.fbs.v, var_str[1]);
   dbl_to_str(nbrcnt_info.fbs.v_bcl[i], var_str[2]);
   dbl_to_str(nbrcnt_info.fbs.v_bcu[i], var_str[3]);
   dbl_to_str(nbrcnt_info.fss.v, var_str[4]);
   dbl_to_str(nbrcnt_info.fss.v_bcl[i], var_str[5]);
   dbl_to_str(nbrcnt_info.fss.v_bcu[i], var_str[6]);

   //
   // Setup the format string
   //
   get_column_format(nbrcnt_columns, n_nbrcnt_columns, fmt_str);

   //
   // Build the NBRCNT line
   //
   sprintf(str, fmt_str,
      var_str[0],             // Total count
      var_str[1],             // Fractions Brier score
      var_str[2], var_str[3], // Fractions Brier score BCI
      var_str[4],             // Fractions Skill score
      var_str[5], var_str[6], // Fractions Skill score BCI
      interp_pnts_str         // Neighborhood points
   );

   return;
}

////////////////////////////////////////////////////////////////////////

void build_isc_str(char *str, ISCInfo &isc_info, int i_scale) {
   char var_str[n_isc_columns][vsdb_col_len];
   char fmt_str[max_str_len];

   //
   // Neighborhood Method Continuous Statistics (isc)
   // Dump out the isc VSDB line:
   //    TOTAL,
   //    NSCALE,      MSE,         ISC,
   //    FENERGY2,    OENERGY2,    BASER,
   //    FBIAS
   //

   // Convert double values to strings
   sprintf(var_str[0], "%i", isc_info.total);
   sprintf(var_str[1], "%i", isc_info.nscale);

   if(i_scale < 0) {
      dbl_to_str(isc_info.mse, var_str[2]);
      dbl_to_str(isc_info.isc, var_str[3]);
      dbl_to_str(isc_info.fen, var_str[4]);
      dbl_to_str(isc_info.oen, var_str[5]);
   }
   else {
      dbl_to_str(isc_info.mse_scale[i_scale], var_str[2]);
      dbl_to_str(isc_info.isc_scale[i_scale], var_str[3]);
      dbl_to_str(isc_info.fen_scale[i_scale], var_str[4]);
      dbl_to_str(isc_info.oen_scale[i_scale], var_str[5]);
   }
   dbl_to_str(isc_info.baser, var_str[6]);
   dbl_to_str(isc_info.fbias, var_str[7]);

   //
   // Setup the format string
   //
   get_column_format(isc_columns, n_isc_columns, fmt_str);

   //
   // Build the isc line
   //
   sprintf(str, fmt_str,
      var_str[0], // Total count
      var_str[1], // Number of scales
      var_str[2], // Mean Squared Error at this scale
      var_str[3], // Intensity-Scale score at this scale
      var_str[4], // Forecast engery
      var_str[5], // Observation Energy
      var_str[6], // Base Rate for this threshold
      var_str[7]  // Frequency Bias for this threshold
   );

   return;
}

////////////////////////////////////////////////////////////////////////
