// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*


///////////////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "vx_wrfmode/engine.h"
#include "vx_util/vx_util.h"

///////////////////////////////////////////////////////////////////////////////

static int model_width = default_model_width;

static inline int max(int a, int b) { return((a > b) ? a : b); }

///////////////////////////////////////////////////////////////////////////////

static DistributionPercentiles get_percentiles(const WrfData &, const WrfData &, const double);

///////////////////////////////////////////////////////////////////////////////
//
// Code for class Engine
//
///////////////////////////////////////////////////////////////////////////////

Engine::Engine() {

   init_from_scratch();
}

///////////////////////////////////////////////////////////////////////////////

Engine::~Engine() {

   clear_features();

   //
   // Clear the grid mask
   //
   if(grid_mask) {
      delete grid_mask;
      grid_mask = (Grid *) 0;
   }

   //
   // Clear the polyline mask
   //
   if(poly_mask) {
      delete poly_mask;
      poly_mask = (Polyline *) 0;
   }

   //
   // Clear single features
   //
   if(fcst_single) {
      delete [] fcst_single;
      fcst_single = (SingleFeature *) 0;
   }
   if(obs_single) {
      delete [] obs_single;
      obs_single = (SingleFeature *) 0;
   }

   //
   // Clear pair features
   //
   if(pair) {
      delete [] pair;
      pair = (PairFeature *) 0;
   }

   //
   // Clear fcst WrfData objects
   //
   if(fcst_raw) {
      delete fcst_raw;
      fcst_raw = (WrfData *) 0;
   }
   if(fcst_filter) {
      delete fcst_filter;
      fcst_filter = (WrfData *) 0;
   }
   if(fcst_conv) {
      delete fcst_conv;
      fcst_conv = (WrfData *) 0;
   }
   if(fcst_mask) {
      delete fcst_mask;
      fcst_mask = (WrfData *) 0;
   }
   if(fcst_split) {
      delete fcst_split;
      fcst_split = (WrfData *) 0;
   }
   if(fcst_comp_split) {
      delete fcst_comp_split;
      fcst_comp_split = (WrfData *) 0;
   }

   //
   // Clear obs WrfData objects
   //
   if(obs_raw) {
      delete obs_raw;
      obs_raw = (WrfData *) 0;
   }
   if(obs_filter) {
      delete obs_filter;
      obs_filter = (WrfData *) 0;
   }
   if(obs_conv) {
      delete obs_conv;
      obs_conv = (WrfData *) 0;
   }
   if(obs_mask) {
      delete obs_mask;
      obs_mask = (WrfData *) 0;
   }
   if(obs_split) {
      delete obs_split;
      obs_split = (WrfData *) 0;
   }
   if(obs_comp_split) {
      delete obs_comp_split;
      obs_comp_split = (WrfData *) 0;
   }

   //
   // Clear fcst and obs engines
   //
   if(fcst_engine) {
      delete fcst_engine;
      fcst_engine = (Engine *) 0;
   }
   if(obs_engine) {
      delete obs_engine;
      obs_engine = (Engine *) 0;
   }
}

///////////////////////////////////////////////////////////////////////////////

Engine::Engine(const Engine &eng) {

   cerr << "\n\nERROR: Engine::Engine(const Engine &) -> "
        << "should never be called!\n\n" << flush;
   exit(1);
}

///////////////////////////////////////////////////////////////////////////////

Engine & Engine::operator=(const Engine & eng) {

   cerr << "\n\nERROR: Engine::operator=(const Engine &) -> "
        << "should never be called\n\n" << flush;
   exit(1);
}

///////////////////////////////////////////////////////////////////////////////

void Engine::init_from_scratch() {

   grid_mask = new Grid;
   poly_mask = new Polyline;

   //
   // Reset all fcst and obs processing flags to initial state
   //
   need_fcst_filter     = 1;
   need_fcst_conv       = 1;
   need_fcst_thresh     = 1;
   need_fcst_split      = 1;
   need_fcst_merge      = 1;
   need_fcst_comp_split = 1;

   need_obs_filter      = 1;
   need_obs_conv        = 1;
   need_obs_thresh      = 1;
   need_obs_split       = 1;
   need_obs_merge       = 1;
   need_obs_comp_split  = 1;

   need_match           = 1;

   fcst_raw         = new WrfData;
   fcst_filter      = new WrfData;
   fcst_conv        = new WrfData;
   fcst_mask        = new WrfData;
   fcst_split       = new WrfData;
   fcst_comp_split  = new WrfData;

   obs_raw          = new WrfData;
   obs_filter       = new WrfData;
   obs_conv         = new WrfData;
   obs_mask         = new WrfData;
   obs_split        = new WrfData;
   obs_comp_split   = new WrfData;

   fcst_engine      = (Engine *) 0;
   obs_engine       = (Engine *) 0;

   n_fcst           = 0;
   n_obs            = 0;

   fcst_single      = (SingleFeature *) 0;
   obs_single       = (SingleFeature *) 0;

   pair             = (PairFeature *) 0;

   fcst_single      = new SingleFeature [max_singles];
   obs_single       = new SingleFeature [max_singles];

   pair             = new PairFeature [max_singles*max_singles];

   if(!fcst_single || !obs_single || !pair) {
      cerr << "\n\nERROR: Engine::init_from_scratch() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   collection.clear();

   clear_features();

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::clear_features() {
   int j;

   for(j=0; j<max_singles; j++) {
      fcst_single[j].clear();
      obs_single[j].clear();
   }

   for(j=0; j<(max_singles*max_singles); j++) {
      pair[j].clear();
   }

   n_fcst = 0;
   n_obs = 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::clear_colors() {
   int j;

   for(j=0; j<max_singles; j++) {
      fcst_color[j].clear();
      obs_color[j].clear();
   }

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::set(const char *fcst_filename, const char *obs_filename) {

   if(!(fcst_raw->read(fcst_filename))) {
      cerr << "\n\nERROR: Engine::set() -> "
           << "unable to open fcst file \"" << fcst_filename << "\"\n\n"
           << flush;
      exit(1);
   }

   if(!(obs_raw->read(obs_filename))) {
      cerr << "\n\nERROR: Engine::set() -> "
           << "unable to open obs file \"" << obs_filename << "\"\n\n"
           << flush;
      exit(1);
   }

   set(*fcst_raw, *obs_raw);

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::set(const WrfData &fcst_wd, const WrfData &obs_wd) {
   char tmp_str[max_str_len];

   clear_features();
   clear_colors();
   ctable.clear();

   set_fcst(fcst_wd);
   set_obs(obs_wd);

   replace_string(met_base_str, MET_BASE, wconf.mode_color_table().sval(),
                  tmp_str);
   ctable.read(tmp_str);

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::set_fcst(const char *fcst_filename) {

   if(!(fcst_raw->read(fcst_filename))) {
      cerr << "\n\nERROR: Engine::set_fcst() -> "
           << "unable to open fcst file \"" << fcst_filename << "\"\n\n"
           << flush;
      exit(1);
   }

   set_fcst(*fcst_raw);

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::set_fcst(const WrfData &fcst_wd) {

   *fcst_raw = fcst_wd;

   need_fcst_filter     = 1;
   need_fcst_conv       = 1;
   need_fcst_thresh     = 1;
   need_fcst_split      = 1;
   need_fcst_merge      = 1;
   need_fcst_comp_split = 1;
   need_match           = 1;

   int k = wconf.zero_border_size().ival();
   fcst_raw->zero_border(k, bad_data_flag);

   do_fcst_filter();
   do_fcst_convolution();
   do_fcst_thresholding();
   do_fcst_splitting();

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::set_obs(const char *obs_filename) {

   if(!(obs_raw->read(obs_filename))) {
      cerr << "\n\nERROR: Engine::set_obs() -> "
           << "unable to open obs file \"" << obs_filename << "\"\n\n"
           << flush;
      exit(1);
   }

   set_obs(*obs_raw);

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::set_obs(const WrfData &obs_wd) {

   *obs_raw = obs_wd;

   need_obs_filter     = 1;
   need_obs_conv       = 1;
   need_obs_thresh     = 1;
   need_obs_split      = 1;
   need_obs_merge      = 1;
   need_obs_comp_split = 1;
   need_match          = 1;

   int k = wconf.zero_border_size().ival();
   obs_raw->zero_border(k, bad_data_flag);

   do_obs_filter();
   do_obs_convolution();
   do_obs_thresholding();
   do_obs_splitting();

   return;
}


///////////////////////////////////////////////////////////////////////////////


int Engine::two_to_one(int n_f, int n_o) const {
   int n;

   n = n_o*n_fcst + n_f;

   return(n);
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_filter() {
   SingleThresh st;

   if(!need_fcst_filter) return;

   *fcst_filter = *fcst_raw;

   //
   // filter out the data which doesn't meet the fcst_raw_threshold
   //
   st.thresh = fcst_filter->double_to_int(fcst_raw_threshold.thresh);
   st.type   = fcst_raw_threshold.type;
   fcst_filter->filter(st);

   need_fcst_filter     = 0;
   need_fcst_conv       = 1;
   need_fcst_thresh     = 1;
   need_fcst_split      = 1;
   need_fcst_merge      = 1;
   need_fcst_comp_split = 1;
   need_match           = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_filter() {
   SingleThresh st;

   if(!need_obs_filter) return;

   *obs_filter = *obs_raw;

   //
   // filter out the data which doesn't meet the obs_raw_threshold
   //
   st.thresh = obs_filter->double_to_int(obs_raw_threshold.thresh);
   st.type   = obs_raw_threshold.type;
   obs_filter->filter(st);

   need_obs_filter     = 0;
   need_obs_conv       = 1;
   need_obs_thresh     = 1;
   need_obs_split      = 1;
   need_obs_merge      = 1;
   need_obs_comp_split = 1;
   need_match          = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_convolution() {
   int r;

   if(need_fcst_filter) do_fcst_filter();

   if(!need_fcst_conv) return;

   r = wconf.fcst_conv_radius().ival();

   *fcst_conv = *fcst_filter;

   //
   // Apply a circular convolution to the filtered field
   //
   if(r > 0) fcst_conv->conv_filter_circ(2*r + 1, wconf.bad_data_threshold().dval());

   fcst_conv->zero_border(wconf.zero_border_size().ival(), bad_data_flag);

   need_fcst_conv       = 0;
   need_fcst_thresh     = 1;
   need_fcst_split      = 1;
   need_fcst_merge      = 1;
   need_fcst_comp_split = 1;
   need_match           = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_convolution() {
   int r;

   if(need_obs_filter) do_obs_filter();

   if(!need_obs_conv) return;

   r = wconf.obs_conv_radius().ival();

   *obs_conv = *obs_filter;

   //
   // Apply a circular convolution to the filtered field
   //
   if(r > 0) obs_conv->conv_filter_circ(2*r + 1, wconf.bad_data_threshold().dval());

   obs_conv->zero_border(wconf.zero_border_size().ival(), bad_data_flag);

   need_obs_conv       = 0;
   need_obs_thresh     = 1;
   need_obs_split      = 1;
   need_obs_merge      = 1;
   need_obs_comp_split = 1;
   need_match          = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_thresholding() {

   if(need_fcst_conv) do_fcst_convolution();

   if(!need_fcst_thresh) return;

   *fcst_mask = *fcst_conv;

   //
   // Threshold the convolved field
   //
   fcst_mask->threshold_double(fcst_conv_threshold);

   //
   // Apply the area threshold
   //
   fcst_mask->threshold_area(fcst_area_threshold);

   //
   // Apply the intensity threshold
   //
   fcst_mask->threshold_intensity(fcst_filter,
                                  wconf.fcst_inten_perc().ival(),
                                  fcst_inten_perc_threshold);

   need_fcst_thresh     = 0;
   need_fcst_split      = 1;
   need_fcst_merge      = 1;
   need_fcst_comp_split = 1;
   need_match           = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_thresholding() {

   if(need_obs_conv) do_obs_convolution();

   if(!need_obs_thresh) return;

   *obs_mask = *obs_conv;

   //
   // Threshold the convolved field
   //
   obs_mask->threshold_double(obs_conv_threshold);

   //
   // Apply the area threshold
   //
   obs_mask->threshold_area(obs_area_threshold);

   //
   // Apply the intensity threshold
   //
   obs_mask->threshold_intensity(obs_filter,
                                 wconf.obs_inten_perc().ival(),
                                 obs_inten_perc_threshold);

   need_obs_thresh     = 0;
   need_obs_split      = 1;
   need_obs_merge      = 1;
   need_obs_comp_split = 1;
   need_match          = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_splitting() {

   if(need_fcst_thresh) do_fcst_thresholding();

   if(!need_fcst_split) return;

   *fcst_split = split(*fcst_mask, n_fcst);

   need_fcst_split      = 0;
   need_fcst_merge      = 1;
   need_fcst_comp_split = 1;
   need_match           = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_splitting() {

   if(need_obs_thresh) do_obs_thresholding();

   if(!need_obs_split) return;

   *obs_split = split(*obs_mask, n_obs);

   need_obs_split      = 0;
   need_obs_merge      = 1;
   need_obs_comp_split = 1;
   need_match          = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_merging() {

   do_fcst_merging("");

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_merging() {

   do_obs_merging("");

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_merging(const char *config_merge) {

   if(need_fcst_thresh) do_fcst_thresholding();

   if(!need_fcst_merge) return;

   if(wconf.fcst_merge_flag().ival() == 1 ||
      wconf.fcst_merge_flag().ival() == 3)   do_fcst_merge_thresh();

   if(wconf.fcst_merge_flag().ival() == 2 ||
      wconf.fcst_merge_flag().ival() == 3)   do_fcst_merge_engine(config_merge);

   //
   // Done
   //

   need_fcst_merge      = 0;
   need_fcst_comp_split = 1;
   need_match           = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_merging(const char *config_merge) {

   if(need_obs_thresh) do_obs_thresholding();

   if(!need_obs_merge) return;

   if(wconf.obs_merge_flag().ival() == 1 ||
      wconf.obs_merge_flag().ival() == 3)   do_obs_merge_thresh();

   if(wconf.obs_merge_flag().ival() == 2 ||
      wconf.obs_merge_flag().ival() == 3)   do_obs_merge_engine(config_merge);

   //
   // Done
   //

   need_obs_merge      = 0;
   need_obs_comp_split = 1;
   need_match          = 1;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_matching() {

   if(!need_match) return;

   if(wconf.match_flag().ival() == 0) {
      cout << "***WARNING*** void Engine::do_matching() -> "
           << "no matching requested in configuration file\n"
           << flush;
                                           do_no_match();
   }
   else if(wconf.match_flag().ival() == 1)       do_match_merge();

   else if(wconf.match_flag().ival() == 2)  do_match_fcst_merge();

   else if(wconf.match_flag().ival() == 3)        do_match_only();

   else {
      cerr << "\n\nERROR: void Engine::do_matching() -> "
           << "invalid match_flag value specified.  match_flag must be "
           << "between 0 and 3.\n\n" << flush;
      exit(1);
   }

   //
   // Generate the fcst and obs composite split fields
   //
   do_fcst_comp_splitting();
   do_obs_comp_splitting();

   //
   // Done
   //

   need_match           = 0;
   need_fcst_comp_split = 0;
   need_obs_comp_split  = 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// perform no matching, but still define the single features for each simple
// object.
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_no_match() {
   int j;
   WrfData * fcst_shape = (WrfData *) 0;
   WrfData * obs_shape = (WrfData *) 0;

   do_fcst_splitting();
   do_obs_splitting();

   if((n_fcst >= max_singles) || (n_obs >= max_singles)) {

      cerr << "\n\nERROR: Engine::do_no_match() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << max(n_fcst, n_obs) << "\n\n" << flush;
      exit(1);
   }

   clear_colors();

   fcst_shape = new WrfData [n_fcst];
   obs_shape = new WrfData [n_obs];

   if(!fcst_shape || !obs_shape) {

      cerr << "\n\nERROR: Engine::do_no_match() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Do the single features
   //

   for(j=0; j<n_fcst; j++) {
      fcst_shape[j] = select(*fcst_split, j+1);
      fcst_single[j].set(*fcst_filter, fcst_shape[j],
                         wconf.intensity_percentile().ival());
      fcst_single[j].object_number = j+1;
   }

   for(j=0; j<n_obs; j++) {
      obs_shape[j] = select(*obs_split, j+1);
      obs_single[j].set(*obs_filter, obs_shape[j],
                        wconf.intensity_percentile().ival());
      obs_single[j].object_number = j+1;
   }

   //
   // Clear our any empty sets
   //
   collection.clear_empty_sets();

   //
   // Done
   //
   delete [] fcst_shape; fcst_shape = (WrfData *) 0;
   delete [] obs_shape;  obs_shape = (WrfData *) 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void Engine::do_match_merge() {
   int j, k, n;
   InterestInfo junkinfo;
   WrfData * fcst_shape = (WrfData *) 0;
   WrfData * obs_shape = (WrfData *) 0;

   do_fcst_splitting();
   do_obs_splitting();

   if((n_fcst >= max_singles) || (n_obs >= max_singles)) {

      cerr << "\n\nERROR: Engine::do_match_merge() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << max(n_fcst, n_obs) << "\n\n" << flush;
      exit(1);
   }

   clear_colors();

   fcst_shape = new WrfData [n_fcst];
   obs_shape = new WrfData [n_obs];

   if(!fcst_shape || !obs_shape) {

      cerr << "\n\nERROR: Engine::do_match_merge() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Do the single features
   //

   for(j=0; j<n_fcst; j++) {
      fcst_shape[j] = select(*fcst_split, j+1);
      fcst_single[j].set(*fcst_filter, fcst_shape[j],
                         wconf.intensity_percentile().ival());
      fcst_single[j].object_number = j+1;
   }

   for(j=0; j<n_obs; j++) {
      obs_shape[j] = select(*obs_split, j+1);
      obs_single[j].set(*obs_filter, obs_shape[j],
                        wconf.intensity_percentile().ival());
      obs_single[j].object_number = j+1;
   }

   //
   // Do the pair features
   //
   for(j=0; j<n_fcst; j++) {
      for(k=0; k<n_obs; k++) {

         n = two_to_one(j, k);

         pair[n].set(fcst_single[j], obs_single[k],
                     wconf.max_centroid_dist().ival());
         pair[n].pair_number = n;
      }
   }

   //
   // Calculate the interest values
   //
   for(j=0; j<n_fcst; j++) {
      for(k=0; k<n_obs; k++) {

         n = two_to_one(j, k);

         info[n].fcst_number = (j+1);
         info[n].obs_number = (k + 1);
         info[n].pair_number = n;
         info[n].interest_value = total_interest(wconf, pair[n]);
      }
   }

   //
   // Sort the interest values in decreasing order
   //
   n = n_fcst*n_obs;

   for(j=0; j<(n-1); j++) {
      for(k=(j+1); k<n; k++) {

         if(info[j].interest_value < info[k].interest_value) {

            junkinfo = info[j];
            info[j] = info[k];
            info[k] = junkinfo;
         }
      }
   }

   //
   // Form the sets
   //
   n = n_fcst*n_obs;

   for(j=0; j<n; j++) {

      if(info[j].interest_value < (wconf.total_interest_threshold().dval()))
         continue;

      collection.add_pair(info[j].fcst_number, info[j].obs_number);
   }

   //
   // Clear our any empty sets
   //
   collection.clear_empty_sets();

   //
   // Assign the colors
   //
   if(collection.n_sets > ctable.n_entries()) {

      cerr << "\n\nERROR: Engine::do_match_merge() -> "
           << "not enough colors ... need at least " << (collection.n_sets)
           << "\n\n" << flush;
      exit(1);
   }

   for(j=0; j<n_fcst; j++) {
      fcst_color[j] = unmatched_color;

      for(k=0; k<(collection.n_sets); k++) {

         if(collection.set[k].has_fcst(j+1)) fcst_color[j] = ctable.nearest(k+1);
      }
   }

   for(j=0; j<n_obs; j++) {
      obs_color[j] = unmatched_color;

      for(k=0; k<(collection.n_sets); k++) {

         if(collection.set[k].has_obs(j+1)) obs_color[j] = ctable.nearest(k+1);
      }
  }

   //
   // Done
   //
   delete [] fcst_shape; fcst_shape = (WrfData *) 0;
   delete [] obs_shape;  obs_shape = (WrfData *) 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// perform merging of the forecast field based on a different convolution
// threshold.  It is the user's responsibility to choose a threshold type and
// value to define objects which are suitable for merging.
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_merge_thresh() {
   int j, k, x, y;
   int n_fcst_merge, intersection;
   int count, first_k;
   WrfData fcst_merge_mask, fcst_merge_split;
   WrfData * fcst_shape = (WrfData *) 0;
   WrfData * fcst_merge_shape = (WrfData *) 0;

   do_fcst_splitting();

   if(n_fcst >= max_singles) {

      cerr << "\n\nERROR: Engine::do_fcst_merge_thresh() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << n_fcst << "\n\n" << flush;
      exit(1);
   }

   //
   // Define the forecast merge field by applying the specified threshold
   // to the convolved field
   //
   fcst_merge_mask = *fcst_conv;

   //
   // Threshold the forecast merge field
   //
   fcst_merge_mask.threshold_double(fcst_merge_threshold);

   //
   // Split up the forecast merge field
   //
   fcst_merge_split = split(fcst_merge_mask, n_fcst_merge);

   //
   // Allocate space for all of the simple forecast shapes and
   // forecast merge shapes
   //
   fcst_shape       = new WrfData [n_fcst];
   fcst_merge_shape = new WrfData [n_fcst_merge];

   if(!fcst_shape || !fcst_merge_shape) {

      cerr << "\n\nERROR: Engine::do_fcst_merge_thresh() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Select all of the simple forecast shapes and
   // forecast merge shapes
   //
   for(j=0; j<n_fcst; j++) {
      fcst_shape[j] = select(*fcst_split, j+1);
   }

   for(j=0; j<n_fcst_merge; j++) {
      fcst_merge_shape[j] = select(fcst_merge_split, j+1);
   }

   //
   // Calculate the composite object sets
   //
   for(j=0; j<n_fcst_merge; j++) {

      count = first_k = 0;

      for(k=0; k<n_fcst; k++) {

         //
         // Calculate intersection area
         //
         intersection = 0;

         for(x=0; x<fcst_merge_mask.get_nx(); x++) {
            for(y=0; y<fcst_merge_mask.get_ny(); y++) {

               if(fcst_merge_shape[j].f_is_on(x, y) &&
                  fcst_shape[k].f_is_on(x, y)) intersection++;
            }
         }

         //
         // Add to set collection only if the fcst object is
         // completely contained in the merge object.  Meaning,
         // intersection area >= fcst area
         //

         if(intersection >= fcst_shape[k].area()) {

            //
            // Keep track of the first embedded shape.  You only want to
            // create a composite if there are more than 1 shapes in it.
            //
            if(count == 0) first_k = k+1;

            else if(count == 1) {
               collection.set[collection.n_sets].add_pair(first_k, -1);
               collection.set[collection.n_sets].add_pair(k+1, -1);
            }
            else {
               collection.set[collection.n_sets].add_pair(k+1, -1);
            }

            count++;
         }
      }  // end for k

      if(count > 0) collection.n_sets++;
   }  // end for j

   //
   // Done
   //

   delete [] fcst_shape;       fcst_shape = (WrfData *) 0;
   delete [] fcst_merge_shape; fcst_merge_shape = (WrfData *) 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// perform merging of the observation field based on a different convolution
// threshold.  It is the user's responsibility to choose a threshold type and
// value to define objects which are suitable for merging.
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_merge_thresh() {
   int j, k, x, y;
   int n_obs_merge, intersection;
   int count, first_k;
   WrfData obs_merge_mask, obs_merge_split;
   WrfData * obs_shape = (WrfData *) 0;
   WrfData * obs_merge_shape = (WrfData *) 0;

   do_obs_splitting();

   if(n_obs >= max_singles) {

      cerr << "\n\nERROR: Engine::do_obs_merge_thresh() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << n_obs << "\n\n" << flush;
      exit(1);
   }

   //
   // Define the forecast merge field by applying the specified threshold
   // to the convolved field
   //
   obs_merge_mask = *obs_conv;

   //
   // Threshold the forecast merge field
   //
   obs_merge_mask.threshold_double(obs_merge_threshold);

   //
   // Split up the forecast merge field
   //
   obs_merge_split = split(obs_merge_mask, n_obs_merge);

   //
   // Allocate space for all of the simple observation shapes and
   // observation merge shapes
   //
   obs_shape       = new WrfData [n_obs];
   obs_merge_shape = new WrfData [n_obs_merge];

   if(!obs_shape || !obs_merge_shape) {

      cerr << "\n\nERROR: Engine::do_obs_merge_thresh() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Select all of the simple observation shapes and
   // simple merge shapes
   //
   for(j=0; j<n_obs; j++) {
      obs_shape[j] = select(*obs_split, j+1);
   }

   for(j=0; j<n_obs_merge; j++) {
      obs_merge_shape[j] = select(obs_merge_split, j+1);
   }

   //
   // Calculate the composite object sets
   //
   for(j=0; j<n_obs_merge; j++) {

      count = first_k = 0;

      for(k=0; k<n_obs; k++) {

         //
         // Calculate intersection area
         //
         intersection = 0;

         for(x=0; x<obs_merge_mask.get_nx(); x++) {
            for(y=0; y<obs_merge_mask.get_ny(); y++) {

               if(obs_merge_shape[j].f_is_on(x, y) &&
                  obs_shape[k].f_is_on(x, y)) intersection++;
            }
         }

         //
         // Add to set collection only if the obs object is
         // completely contained in the merge object.  Meaning,
         // intersection area >= obs area
         //

         if(intersection >= obs_shape[k].area()) {

            //
            // Keep track of the first embedded shape.  You only want to
            // create a composite if there are more than 1 shapes in it.
            //
            if(count == 0) first_k = k+1;

            else if(count == 1) {
               collection.set[collection.n_sets].add_pair(-1, first_k);
               collection.set[collection.n_sets].add_pair(-1, k+1);
            }
            else {
               collection.set[collection.n_sets].add_pair(-1, k+1);
            }

            count++;
         }
      }  // end for k

      if(count > 0) collection.n_sets++;
   }  // end for j

   //
   // Done
   //

   delete [] obs_shape;       obs_shape = (WrfData *) 0;
   delete [] obs_merge_shape; obs_merge_shape = (WrfData *) 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// Perform merging of the forecast field using a fuzzy engine matching
// approach
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_merge_engine(const char *config_file) {
   int i, j;
   WrfData fcst_merge_split;
   char tmp_str[max_str_len];

   do_fcst_splitting();

   if(n_fcst >= max_singles) {
      cerr << "\n\nERROR: Engine::do_fcst_merge_engine() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << n_fcst << "\n\n" << flush;
      exit(1);
   }

   //
   // Will be deleted by destructor if allocated
   //
   fcst_engine = new Engine;

   if(!fcst_engine) {
      cerr << "\n\nERROR: Engine::do_fcst_merge_engine() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Specify the configuration for the forecast merging fuzzy engine
   //
   fcst_engine->ctable = ctable;
   if(config_file) {
      fcst_engine->wconf.read(config_file);
      fcst_engine->process_engine_config();
      replace_string(met_base_str, MET_BASE,
                     fcst_engine->wconf.mode_color_table().sval(), tmp_str);
      fcst_engine->ctable.read(tmp_str);
   }

   //
   // Copy the previously defined fuzzy engine fields
   // setting the current forecast field to both the
   // forecast and forecast fields in the fcst_engine.
   // this will capture any previous merging performed
   // in the current forecast field.
   //
   *(fcst_engine->fcst_raw)    = *(fcst_raw);
   *(fcst_engine->obs_raw)     = *(fcst_raw);

   *(fcst_engine->fcst_filter) = *(fcst_filter);
   *(fcst_engine->obs_filter)  = *(fcst_filter);

   *(fcst_engine->fcst_conv)   = *(fcst_conv);
   *(fcst_engine->obs_conv)    = *(fcst_conv);

   *(fcst_engine->fcst_mask)   = *(fcst_mask);
   *(fcst_engine->obs_mask)    = *(fcst_mask);

   *(fcst_engine->fcst_split)  = *(fcst_split);
   *(fcst_engine->obs_split)   = *(fcst_split);

   fcst_engine->n_fcst = n_fcst;
   fcst_engine->n_obs  = n_fcst;

   fcst_engine->need_fcst_filter     = 0;
   fcst_engine->need_fcst_conv       = 0;
   fcst_engine->need_fcst_thresh     = 0;
   fcst_engine->need_fcst_split      = 0;
   fcst_engine->need_fcst_merge      = 1;
   fcst_engine->need_fcst_comp_split = 1;

   fcst_engine->need_obs_filter      = 0;
   fcst_engine->need_obs_conv        = 0;
   fcst_engine->need_obs_thresh      = 0;
   fcst_engine->need_obs_split       = 0;
   fcst_engine->need_obs_merge       = 1;
   fcst_engine->need_obs_comp_split  = 1;

   fcst_engine->need_match           = 1;

   //
   // Copy the top level set collection down to the fcst engine
   // to capture any merging that has already occurred, via
   // the double threshold method
   //
   for(i=0; i<collection.n_sets; i++) {
      if(collection.set[i].n_fcst > 0) {
         for(j=0; j<collection.set[i].n_fcst; j++) {

            fcst_engine->collection.set[fcst_engine->collection.n_sets].add_pair(
               -1, collection.set[i].fcst_number[j]);
         }
         fcst_engine->collection.n_sets++;
      }
   }

   //
   // Perform merging and matching for the fcst_engine using the
   // simple merging and matching approach
   //
   fcst_engine->do_match_merge();

   //
   // Generate the fcst and obs composite split fields
   //
   fcst_engine->do_fcst_comp_splitting();
   fcst_engine->do_obs_comp_splitting();

   //
   // Based on the results of the merging and matching
   // of the fcst_engine, update the set collection
   // to indicate merging
   //
   for(i=0; i<fcst_engine->collection.n_sets; i++) {

      //
      // All objects will at least match themselves.
      // only define sets when they are composed of two or
      // more simple objects.
      //
      if(fcst_engine->collection.set[i].n_fcst >= 2) {
         for(j=0; j<fcst_engine->collection.set[i].n_fcst; j++) {

            collection.set[collection.n_sets].add_pair(
               fcst_engine->collection.set[i].fcst_number[j], -1);
         }
         collection.n_sets++;
      }
   } // end for i

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// Perform merging of the observation field using a fuzzy engine matching
// approach
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_merge_engine(const char *config_file) {
   int i, j;
   WrfData obs_merge_split;
   char tmp_str[max_str_len];

   do_obs_splitting();

   if(n_obs >= max_singles) {

      cerr << "\n\nERROR: Engine::do_obs_merge_engine() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << n_obs << "\n\n" << flush;
      exit(1);
   }

   //
   // Will be deleted by destructor if allocated
   //
   obs_engine = new Engine;

   if(!obs_engine) {
      cerr << "\n\nERROR: Engine::do_obs_merge_engine() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Specify the configuration for the observation merging fuzzy engine
   //
   obs_engine->ctable = ctable;
   if(config_file) {
      obs_engine->wconf.read(config_file);
      obs_engine->process_engine_config();
      replace_string(met_base_str, MET_BASE,
                     obs_engine->wconf.mode_color_table().sval(), tmp_str);
      obs_engine->ctable.read(tmp_str);
   }

   //
   // Copy the previously defined fuzzy engine fields
   // setting the current observation field to both the
   // forecast and observation fields in the obs_engine.
   // this will capture any previous merging performed
   // in the current observation field.
   //
   *(obs_engine->fcst_raw)    = *(obs_raw);
   *(obs_engine->obs_raw)     = *(obs_raw);

   *(obs_engine->fcst_filter) = *(obs_filter);
   *(obs_engine->obs_filter)  = *(obs_filter);

   *(obs_engine->fcst_conv)   = *(obs_conv);
   *(obs_engine->obs_conv)    = *(obs_conv);

   *(obs_engine->fcst_mask)   = *(obs_mask);
   *(obs_engine->obs_mask)    = *(obs_mask);

   *(obs_engine->fcst_split)  = *(obs_split);
   *(obs_engine->obs_split)   = *(obs_split);

   obs_engine->n_fcst = n_obs;
   obs_engine->n_obs  = n_obs;

   obs_engine->need_fcst_filter     = 0;
   obs_engine->need_fcst_conv       = 0;
   obs_engine->need_fcst_thresh     = 0;
   obs_engine->need_fcst_split      = 0;
   obs_engine->need_fcst_merge      = 1;
   obs_engine->need_fcst_comp_split = 1;

   obs_engine->need_obs_filter      = 0;
   obs_engine->need_obs_conv        = 0;
   obs_engine->need_obs_thresh      = 0;
   obs_engine->need_obs_split       = 0;
   obs_engine->need_obs_merge       = 1;
   obs_engine->need_obs_comp_split  = 1;

   obs_engine->need_match           = 1;

   //
   // Copy the top level set collection down to the obs engine
   // to capture any merging that has already occurred, via
   // the double threshold method.  But only copy down those
   // sets containing merged observation objects.
   //
   for(i=0; i<collection.n_sets; i++) {
      if(collection.set[i].n_obs > 0) {
         for(j=0; j<collection.set[i].n_obs; j++) {

            obs_engine->collection.set[obs_engine->collection.n_sets].add_pair(
               -1, collection.set[i].obs_number[j]);
         }
         obs_engine->collection.n_sets++;
      }
   }

   //
   // Perform merging and matching for the obs_engine using the
   // simple merging and matching approach
   //
   obs_engine->do_match_merge();

   //
   // Generate the fcst and obs composite split fields
   //
   obs_engine->do_fcst_comp_splitting();
   obs_engine->do_obs_comp_splitting();

   //
   // Based on the results of the merging and matching
   // of the obs_engine, update the set collection
   // to indicate merging
   //
   for(i=0; i<obs_engine->collection.n_sets; i++) {

      //
      // All objects will at least match themselves.
      // only define sets when they are composed of two or
      // more simple objects.
      //
      if(obs_engine->collection.set[i].n_obs >= 2) {
         for(j=0; j<obs_engine->collection.set[i].n_obs; j++) {

            collection.set[collection.n_sets].add_pair(
               -1, obs_engine->collection.set[i].obs_number[j]);
         }
         collection.n_sets++;
      }
   }

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// Perform matching between the already merged forecast and observation
// fields allowing no additional merging of objects in the observation field
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_match_fcst_merge() {
   int j, k, n;
   InterestInfo junkinfo;
   WrfData * fcst_shape = (WrfData *) 0;
   WrfData * obs_shape = (WrfData *) 0;

   do_fcst_splitting();
   do_obs_splitting();

   if((n_fcst >= max_singles) || (n_obs >= max_singles)) {
      cerr << "\n\nERROR: Engine::do_match_fcst_merge() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << max(n_fcst, n_obs) << "\n\n" << flush;
      exit(1);
   }

   clear_colors();

   fcst_shape = new WrfData [n_fcst];
   obs_shape = new WrfData [n_obs];

   if(!fcst_shape || !obs_shape) {
      cerr << "\n\nERROR: Engine::do_match_merge() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Do the single features
   //
   for(j=0; j<n_fcst; j++) {
      fcst_shape[j] = select(*fcst_split, j+1);
      fcst_single[j].set(*fcst_filter, fcst_shape[j],
                         wconf.intensity_percentile().ival());
      fcst_single[j].object_number = j+1;
   }

   for(j=0; j<n_obs; j++) {
      obs_shape[j] = select(*obs_split, j+1);
      obs_single[j].set(*obs_filter, obs_shape[j],
                        wconf.intensity_percentile().ival());
      obs_single[j].object_number = j+1;
   }

   //
   // Do the pair features
   //
   for(j=0; j<n_fcst; j++) {
      for(k=0; k<n_obs; k++) {

         n = two_to_one(j, k);

         pair[n].set(fcst_single[j], obs_single[k],
                     wconf.max_centroid_dist().ival());
         pair[n].pair_number = n;
      }
   }

   //
   // Calculate the interest values
   //
   for(j=0; j<n_fcst; j++) {
      for(k=0; k<n_obs; k++) {

         n = two_to_one(j, k);

         info[n].fcst_number = (j+1);
         info[n].obs_number = (k + 1);
         info[n].pair_number = n;
         info[n].interest_value = total_interest(wconf, pair[n]);
      }
   }

   //
   // Sort the interest values in decreasing order
   //
   n = n_fcst*n_obs;

   for(j=0; j<(n-1); j++) {
      for(k=(j+1); k<n; k++) {

         if(info[j].interest_value < info[k].interest_value) {

            junkinfo = info[j];
            info[j] = info[k];
            info[k] = junkinfo;
         }
      }
   }

   //
   // Form the sets
   //
   n = n_fcst*n_obs;

   for(j=0; j<n; j++) {

      if(info[j].interest_value < (wconf.total_interest_threshold().dval())) {
         continue;
      }

      //
      // Only add this pair if the forecast object is not already matched
      //
      if(!collection.is_fcst_matched(info[j].fcst_number)) {
         collection.add_pair(info[j].fcst_number, info[j].obs_number);
      }
   }

   //
   // Clear our any empty sets
   //
   collection.clear_empty_sets();

   //
   // Assign the colors
   //
   if(collection.n_sets > ctable.n_entries()) {
      cerr << "\n\nERROR: Engine::do_match_fcst_merge() -> "
           << "not enough colors ... need at least " << (collection.n_sets)
           << "\n\n" << flush;
      exit(1);
   }

   for(j=0; j<n_fcst; j++) {
      fcst_color[j] = unmatched_color;

      for(k=0; k<(collection.n_sets); k++) {

         if(collection.set[k].has_fcst(j+1)) fcst_color[j] = ctable.nearest(k+1);
      }

   }

   for(j=0; j<n_obs; j++) {
      obs_color[j] = unmatched_color;

      for(k=0; k<(collection.n_sets); k++) {

         if(collection.set[k].has_obs(j+1)) obs_color[j] = ctable.nearest(k+1);
      }
   }

   //
   // Done
   //

   delete [] fcst_shape;  fcst_shape = (WrfData *) 0;
   delete [] obs_shape;   obs_shape = (WrfData *) 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// Perform matching between the already merged forecast and observation
// fields allowing no additional merging of objects in the either field
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_match_only() {
   int j, k, n;
   InterestInfo junkinfo;
   WrfData * fcst_shape = (WrfData *) 0;
   WrfData * obs_shape = (WrfData *) 0;

   do_fcst_splitting();
   do_obs_splitting();

   if((n_fcst >= max_singles) || (n_obs >= max_singles)) {
      cerr << "\n\nERROR: Engine::do_match_only() -> "
           << "too many shapes ... increase \"max_singles\" to at least "
           << max(n_fcst, n_obs) << "\n\n" << flush;
      exit(1);
   }

   clear_colors();

   fcst_shape = new WrfData [n_fcst];
   obs_shape = new WrfData [n_obs];

   if(!fcst_shape || !obs_shape) {
      cerr << "\n\nERROR: Engine::do_match_only() -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Do the single features
   //
   for(j=0; j<n_fcst; j++) {
      fcst_shape[j] = select(*fcst_split, j+1);
      fcst_single[j].set(*fcst_filter, fcst_shape[j],
                         wconf.intensity_percentile().ival());
      fcst_single[j].object_number = j+1;
   }

   for(j=0; j<n_obs; j++) {
      obs_shape[j] = select(*obs_split, j+1);
      obs_single[j].set(*obs_filter, obs_shape[j],
                        wconf.intensity_percentile().ival());
      obs_single[j].object_number = j+1;
   }

   //
   // Do the pair features
   //
   for(j=0; j<n_fcst; j++) {
      for(k=0; k<n_obs; k++) {

         n = two_to_one(j, k);

         pair[n].set(fcst_single[j], obs_single[k],
                     wconf.max_centroid_dist().ival());
         pair[n].pair_number = n;
      }
   }

   //
   // Calculate the interest values
   //
   for(j=0; j<n_fcst; j++) {
      for(k=0; k<n_obs; k++) {

         n = two_to_one(j, k);

         info[n].fcst_number = (j+1);
         info[n].obs_number = (k + 1);
         info[n].pair_number = n;
         info[n].interest_value = total_interest(wconf, pair[n]);
      }
   }

   //
   // Sort the interest values in decreasing order
   //
   n = n_fcst*n_obs;
   for(j=0; j<(n-1); j++) {
      for(k=(j+1); k<n; k++) {

         if(info[j].interest_value < info[k].interest_value) {

            junkinfo = info[j];
            info[j] = info[k];
            info[k] = junkinfo;
         }
      }
   }

   //
   // Form the sets
   //
   n = n_fcst*n_obs;

   for(j=0; j<n; j++) {

      if(info[j].interest_value < (wconf.total_interest_threshold().dval()))
         continue;

      //
      // Only add this pair if both the forecast and observation objects
      // are not already matched
      //
      if(!collection.is_fcst_matched(info[j].fcst_number) &&
         !collection.is_obs_matched(info[j].obs_number)) {
         collection.add_pair(info[j].fcst_number, info[j].obs_number);
      }
   }

   //
   // Clear our any empty sets
   //
   collection.clear_empty_sets();

   //
   // Assign the colors
   //
   if(collection.n_sets > ctable.n_entries()) {
      cerr << "\n\nERROR: Engine::do_match_only() -> "
           << "not enough colors ... need at least " << (collection.n_sets)
           << "\n\n" << flush;
      exit(1);
   }

   for(j=0; j<n_fcst; j++) {
      fcst_color[j] = unmatched_color;

      for(k=0; k<(collection.n_sets); k++) {
         if(collection.set[k].has_fcst(j+1)) fcst_color[j] = ctable.nearest(k+1);
      }
   }

   for(j=0; j<n_obs; j++) {
      obs_color[j] = unmatched_color;

      for(k=0; k<(collection.n_sets); k++) {
         if(collection.set[k].has_obs(j+1)) obs_color[j] = ctable.nearest(k+1);
      }
   }

   //
   // Done
   //

   delete [] fcst_shape;   fcst_shape = (WrfData *) 0;
   delete [] obs_shape;    obs_shape = (WrfData *) 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// Create a split composite field based on the merging of the single objects
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_fcst_comp_splitting() {
   int x, y, v_int, set_number;
   WrfData junk;

   //
   // Initialize fcst_comp_split
   //
   *fcst_comp_split = *fcst_split;

   for(x=0; x<fcst_comp_split->get_nx(); x++) {
      for(y=0; y<fcst_comp_split->get_ny(); y++) {

         v_int = fcst_split->get_xy_int(x, y);

         if(v_int == 0) continue;

         set_number = collection.fcst_set_number(v_int);

         fcst_comp_split->put_xy_int(set_number + 1, x, y);
      }  // end for y
   }  // end for x

   fcst_comp_split->calc_moments();

   //
   // Done
   //

   need_fcst_comp_split = 0;

   return;
}

///////////////////////////////////////////////////////////////////////////////
//
// Create a split composite field based on the merging of the single objects
//
///////////////////////////////////////////////////////////////////////////////

void Engine::do_obs_comp_splitting() {
   int x, y, v_int, set_number;
   WrfData junk;

   //
   // Initialize obs_comp_split
   //
   *obs_comp_split = *obs_split;

   for(x=0; x<obs_comp_split->get_nx(); x++) {
      for(y=0; y<obs_comp_split->get_ny(); y++) {

         v_int = obs_split->get_xy_int(x, y);

         if(v_int == 0) continue;

         set_number = collection.obs_set_number(v_int);

         obs_comp_split->put_xy_int(set_number + 1, x, y);
      }  // end for y
   }  // end for x

   obs_comp_split->calc_moments();

   //
   // Done
   //

   need_obs_comp_split = 0;

   return;
}

////////////////////////////////////////////////////////////////////////

void Engine::process_engine_config() {

   //
   // For each of the threshold values specified in the config file
   // parse out the threshold value and the threshold indicator
   //
   fcst_raw_threshold.set(wconf.fcst_raw_threshold().sval());
   obs_raw_threshold.set(wconf.obs_raw_threshold().sval());
   fcst_conv_threshold.set(wconf.fcst_conv_threshold().sval());
   obs_conv_threshold.set(wconf.obs_conv_threshold().sval());
   fcst_area_threshold.set(wconf.fcst_area_threshold().sval());
   obs_area_threshold.set(wconf.obs_area_threshold().sval());
   fcst_inten_perc_threshold.set(wconf.fcst_inten_perc_threshold().sval());
   obs_inten_perc_threshold.set(wconf.obs_inten_perc_threshold().sval());
   fcst_merge_threshold.set(wconf.fcst_merge_threshold().sval());
   obs_merge_threshold.set(wconf.obs_merge_threshold().sval());

   return;
}

///////////////////////////////////////////////////////////////////////////////

int Engine::get_info_index(int pair_n) const {
   int i;

   for(i=0; i<max_singles*max_singles; i++) {

      if(info[i].pair_number == pair_n) return(i);
   }

   return(-1);
}

///////////////////////////////////////////////////////////////////////////////

int Engine::get_unmatched_fcst() const {
   int i, count;

   count = 0;

   for(i=0; i<n_fcst; i++) {

      if(collection.fcst_set_number(i+1) == -1) count++;
   }

   return(count);
}

///////////////////////////////////////////////////////////////////////////////

int Engine::get_unmatched_obs() const {
   int i, count;

   count = 0;

   for(i=0; i<n_obs; i++) {

      if(collection.obs_set_number(i+1) == -1) count++;
   }

   return(count);
}

///////////////////////////////////////////////////////////////////////////////

double total_interest(WrfMode_Conf &fm, const PairFeature &p) {
   double t;

   t = total_interest_print(fm, p, (ostream *) 0);

   return(t);
}

///////////////////////////////////////////////////////////////////////////////

double total_interest_print(WrfMode_Conf &wmc, const PairFeature &p, ostream *out) {
   double attribute;
   double interest, weight;
   double confidence;
   double aspect_obs, aspect_fcst;
   double complexity_obs, complexity_fcst;
   double conf_obs, conf_fcst;
   double term, sum, weight_sum;
   double total;

   /////////////////////////////////////////////////////////////////////
   //
   // Check for centroid distance too large, if it is don't compute
   // the interest for this pair
   //
   /////////////////////////////////////////////////////////////////////

   if(p.centroid_dist > wmc.max_centroid_dist().ival()) {
      total = stat_missing_value;

      if(out) {
         (*out) << "Total Interest = " << total << "\n"
                << "Centroid Distance (" << p.centroid_dist
                << ") > Max Centroid Distance ("
                << wmc.max_centroid_dist().ival() << ")\n";
      }
      return(total);
   }

   sum = 0.0;
   weight_sum = 0.0;

   /////////////////////////////////////////////////////////////////////
   //
   // Centroid distance
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = p.centroid_dist;
   interest    = wmc.centroid_dist_if(attribute);
   confidence  = wmc.area_ratio_conf(p.area_ratio);
   weight      = wmc.centroid_dist_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Centroid Distance:\n"
             << "------------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Boundary distance
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = p.boundary_dist;
   interest    = wmc.boundary_dist_if(attribute);
   confidence  = 1.0;
   weight      = wmc.boundary_dist_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Boundary Distance:\n"
             << "------------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Convex Hull distance
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = p.convex_hull_dist;
   interest    = wmc.convex_hull_dist_if(attribute);
   confidence  = 1.0;
   weight      = wmc.convex_hull_dist_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Convex Hull Distance:\n"
             << "------------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Angle difference
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = p.angle_diff;
   interest    = wmc.angle_diff_if(attribute);
   aspect_obs  = p.Obs->aspect_ratio;
   aspect_fcst = p.Fcst->aspect_ratio;
   conf_obs    = wmc.aspect_ratio_conf(aspect_obs);
   conf_fcst   = wmc.aspect_ratio_conf(aspect_fcst);
   confidence  = sqrt(conf_obs*conf_fcst);
   weight      = wmc.angle_diff_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Angle Difference:\n"
             << "-----------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Area ratio
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = p.area_ratio;
   interest    = wmc.area_ratio_if(attribute);
   confidence  = 1.0;
   weight      = wmc.area_ratio_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Area Ratio:\n"
             << "-----------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Intersection/area ratio
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = (p.intersection_area)/(min(p.Obs->area, p.Fcst->area));
   interest    = wmc.int_area_ratio_if(attribute);
   confidence  = 1.0;
   weight      = wmc.int_area_ratio_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Intersection/Area Ratio:\n"
             << "----------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Complexity ratio
   //
   /////////////////////////////////////////////////////////////////////

   attribute       = p.complexity_ratio;
   complexity_obs  = p.Obs->complexity;
   complexity_fcst = p.Fcst->complexity;
   // Both complexities are non-zero
   if(complexity_obs > 0 && complexity_fcst > 0) {
      interest  = wmc.complexity_ratio_if(attribute);
   }
   // At least one complexity is zero
   else {
      interest  = wmc.ratio_if(attribute);
   }
   confidence  = 1.0;
   weight      = wmc.complexity_ratio_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Complexity Ratio:\n"
             << "-----------------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Nth Percentile intensity ratio
   //
   /////////////////////////////////////////////////////////////////////

   attribute   = p.percentile_intensity_ratio;
   interest    = wmc.intensity_ratio_if(attribute);
   confidence  = 1.0;
   weight      = wmc.intensity_ratio_weight().dval();
   term        = weight*interest*confidence;
   sum        += term;
   weight_sum += weight*confidence;

   if(out) {
      (*out) << "Percentile Intensity Ratio:\n"
             << "-----------------------\n"
             << "   Value      = " << attribute  << "\n"
             << "   Interest   = " << interest   << "\n"
             << "   Confidence = " << confidence << "\n"
             << "   Weight     = " << weight     << "\n"
             << "   Term       = " << term       << "\n"
             << "\n";
   }

   /////////////////////////////////////////////////////////////////////
   //
   // Done
   //
   /////////////////////////////////////////////////////////////////////

   total = sum/weight_sum;

   if(out) {
      (*out) << "Total Interest = (sum of terms)/(sum of weights*confidence)\n\n"
             << "               = " << sum << "/" << weight_sum << "\n\n"
             << "               = " << total << "\n\n";
   }

   return(total);
}

///////////////////////////////////////////////////////////////////////////////

void write_engine_stats(Engine &eng, const Grid &grid, ostream &out) {
   int i, j;

   model_width = max(strlen("MODEL"), strlen(eng.wconf.model().sval()));

   //
   // Column headers
   //
   write_header(eng, out);

   //
   // Single objects
   //
   for(i=0; i<eng.n_fcst; i++) {
      write_fcst_single(eng, i, grid, out);
   }

   for(i=0; i<eng.n_obs; i++) {
      write_obs_single(eng, i, grid, out);
   }

   //
   // If no matching was requested, don't write any more
   //
   if(eng.wconf.match_flag().ival() == 0) return;

   //
   // Object pairs
   //
   for(i=0; i<eng.n_fcst; i++) {
      for(j=0; j<eng.n_obs; j++) {
         write_pair(eng, i, j, out);
      }
   }

   //
   // Single composites
   //
   for(i=0; i<(eng.collection.n_sets); i++) {
      write_fcst_composite(eng, i, grid, out);
   }

   for(i=0; i<(eng.collection.n_sets); i++) {
      write_obs_composite(eng, i, grid, out);
   }

   //
   // Composite Pairs
   //
   for(i=0; i<(eng.collection.n_sets); i++) {
      write_composite_pair(eng, i, i, out);
   }

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_header(Engine &eng, ostream &out) {
   char line[max_line_len];

   out.setf(ios::left);

   // Model Name
   out.width(model_width); out << "MODEL" << sep;

   // Forecast lead time
   out.width(fcst_lead_width); out << "FCST_LEAD" << sep;
   // Forecast valid time
   out.width(fcst_valid_width); out << "FCST_VALID" << sep;
   // Forecast accumulation time
   out.width(fcst_accum_width); out << "FCST_ACCUM" << sep;

   // Observation lead time
   out.width(lead2_width); out << "LEAD2" << sep;
   // Observation valid time
   out.width(obs_valid_width); out << "OBS_VALID" << sep;
   // Observation accumulation time
   out.width(obs_accum_width); out << "OBS_ACCUM" << sep;

   // Forecast convolution radius in grid squares
   out.width(rad_width); out << "FCST_RAD" << sep;
   // Forecast convolution threshold
   out.width(thresh_width); out << "FCST_THR" << sep;
   // Observation convolution radius
   out.width(rad_width); out << "OBS_RAD" << sep;
   // Observation convolution threshold
   out.width(thresh_width); out << "OBS_THR" << sep;

   // Forecast variable being analyzed
   out.width(cts_width); out << "FCST_VAR" << sep;
   // Forecast variable level being analyzed
   out.width(cts_width); out << "FCST_LEV" << sep;
   // Observation variable being analyzed
   out.width(cts_width); out << "OBS_VAR" << sep;
   // Observation variable level being analyzed
   out.width(cts_width); out << "OBS_LEV" << sep;
   // Object ID
   out.width(object_id_width); out << "OBJECT_ID" << sep;
   // Object category
   out.width(object_cat_width); out << "OBJECT_CAT" << sep;

   out << "=" << sep;

   ///////////////////////////////////////////////////////////////////////////

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out << "CENTROID_X" << sep;
   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out << "CENTROID_Y" << sep;
   // Object centroid, latitude
   out.width(centroid_latlon_width); out << "CENTROID_LAT" << sep;
   // Object centroid, longitude
   out.width(centroid_latlon_width); out << "CENTROID_LON" << sep;
   // Axis angle in degrees from north
   out.width(axis_angle_width); out << "AXIS_ANG" << sep;
   // Length of the major axis of fitting ellipse
   out.width(length_width); out << "LENGTH" << sep;
   // Length of the minor axis of fitting ellipse
   out.width(width_width); out << "WIDTH" << sep;
   // Object area as a count of grid squares
   out.width(area_width); out << "AREA" << sep;
   // Area of filtered object
   out.width(area_filter_width); out << "AREA_FILTER" << sep;
   // Object curvature
   out.width(curvature_width); out << "CURVATURE" << sep;
   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out << "CURVATURE_X" << sep;
   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out << "CURVATURE_Y" << sep;
   // Object complexity
   out.width(complexity_width); out << "COMPLEXITY" << sep;
   // 10th percentile of object intensity
   out.width(intensity_width); out << "INTENSITY_10" << sep;
   // 25th percentile of object intensity
   out.width(intensity_width); out << "INTENSITY_25" << sep;
   // 50th percentile of object intensity
   out.width(intensity_width); out << "INTENSITY_50" << sep;
   // 75th percentile of object intensity
   out.width(intensity_width); out << "INTENSITY_75" << sep;
   // 90th percentile of object intensity
   out.width(intensity_width); out << "INTENSITY_90" << sep;
   // Specified percentile of object intensity
   sprintf(line, "INTENSITY_%d", nint(eng.wconf.intensity_percentile().ival()));
   out.width(intensity_width); out << line << sep;
   // Sum of the intensity values withint he object
   out.width(intensity_width); out << "INTENSITY_SUM" << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   out.width(centroid_dist_width); out << "CENTROID_DIST" << sep;
   // Distance between boundaries
   out.width(boundary_dist_width); out << "BOUNDARY_DIST" << sep;
   // Distance between convex hulls
   out.width(convex_hull_dist_width); out << "CONVEX_HULL_DIST" << sep;
   // Difference in angles in degrees
   out.width(angle_diff_width); out << "ANGLE_DIFF" << sep;
   // Area ratio
   out.width(area_ratio_width); out << "AREA_RATIO" << sep;
   // Intersection area
   out.width(intersection_area_width); out << "INTERSECTION_AREA" << sep;
   // Union area
   out.width(union_area_width); out << "UNION_AREA" << sep;
   // Symmetric difference area
   out.width(symmetric_diff_width); out << "SYMMETRIC_DIFF" << sep;
   // Intersection over area
   out.width(intersection_over_area_width); out << "INTERSECTION_OVER_AREA" << sep;
   // Complexity ratio
   out.width(complexity_ratio_width); out << "COMPLEXITY_RATIO" << sep;
   // Percentile intensity ratio
   out.width(percentile_intensity_ratio_width); out << "PERCENTILE_INTENSITY_RATIO" << sep;
   // Total interest value
   out.width(interest_width); out << "INTEREST" << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_header_columns(Engine &eng, ostream &out) {
   int mon, day, yr, hr, min, sec;
   char line[max_line_len];

   // Model Name
   out.width(model_width);
   out << eng.wconf.model().sval() << sep;

   // Forecast lead time
   sec_to_hms(eng.fcst_raw->get_lead_time(), hr, min, sec);
   sprintf(line, hms_fmt, hr, min, sec);
   out.width(fcst_lead_width); out << line << sep;

   // Forecast valid time
   unix_to_mdyhms(eng.fcst_raw->get_valid_time(), mon, day, yr, hr, min, sec);
   sprintf(line, ymd_hms_fmt, yr, mon, day, hr, min, sec);
   out.width(fcst_valid_width); out << line << sep;

   // Forecast accumulation time
   sec_to_hms(eng.fcst_raw->get_accum_time(), hr, min, sec);
   sprintf(line, hms_fmt, hr, min, sec);
   out.width(fcst_accum_width); out << line << sep;

   // Observation lead time
   sec_to_hms(eng.obs_raw->get_lead_time(), hr, min, sec);
   sprintf(line, hms_fmt, hr, min, sec);
   out.width(lead2_width); out << line << sep;

   // Observation valid time
   unix_to_mdyhms(eng.obs_raw->get_valid_time(), mon, day, yr, hr, min, sec);
   sprintf(line, ymd_hms_fmt, yr, mon, day, hr, min, sec);
   out.width(obs_valid_width); out << line << sep;

   // Observation accumulation time
   sec_to_hms(eng.obs_raw->get_accum_time(), hr, min, sec);
   sprintf(line, hms_fmt, hr, min, sec);
   out.width(obs_accum_width); out << line << sep;

   // Forecast convolution radius
   out.width(rad_width); out.precision(integer_precision);
   out << eng.wconf.fcst_conv_radius().ival() << sep;

   // Forecast convolution threshold
   eng.fcst_conv_threshold.get_str(line);
   out.width(thresh_width); out << line << sep;

   // Observation convolution radius
   out.width(rad_width); out.precision(integer_precision);
   out << eng.wconf.obs_conv_radius().ival() << sep;

   // Observation convolution threshold
   eng.obs_conv_threshold.get_str(line);
   out.width(thresh_width); out << line << sep;

   // Forecast Variable Name
   out.width(cts_width); out.precision(default_precision);
   out << eng.fcst_field_name << sep;

   // Forecast Variable Level
   out.width(cts_width); out.precision(default_precision);
   out << eng.fcst_level_name << sep;

   // Observation Variable Name
   out.width(cts_width); out.precision(default_precision);
   out << eng.obs_field_name << sep;

   // Observation Variable Level
   out.width(cts_width); out.precision(default_precision);
   out << eng.obs_level_name << sep;

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_fcst_single(Engine &eng, const int n, const Grid &grid, ostream &out) {
   int i;
   double xcurv, ycurv, curv, lat, lon;
   char line[max_line_len];
   WrfData mask;
   DistributionPercentiles ptile;

   if(n >= eng.n_fcst) {
      cerr << "\n\nERROR: write_fcst_single(const Engine &, int, ostream) -> "
           << n << " >= number of fcst, " << eng.n_fcst << "\n\n" << flush;
      exit(1);
   }

   // Write out the common header columns
   write_header_columns(eng, out);

   // Object ID
   sprintf(line, "F%03d", (n+1));
   out.width(object_id_width); out << line << sep;

   // Object category
   i = eng.collection.fcst_set_number(n + 1);
   sprintf(line, "CF%03d", (i + 1));
   out.width(object_cat_width); out << line << sep;

   out << "=" << sep;

   ////////////////////////////////////////////////////////////////////////////

   mask = select(*(eng.fcst_split), n+1);
   grid.xy_to_latlon(eng.fcst_single[n].xbar, eng.fcst_single[n].ybar, lat, lon);

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << eng.fcst_single[n].xbar << sep;

   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << eng.fcst_single[n].ybar << sep;

   // Object centroid, latitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << lat << sep;

   // Object centroid, longitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << -1.0*lon << sep;

   // Axis angle
   out.width(axis_angle_width); out.precision(default_precision);
   out << eng.fcst_single[n].angle_degrees << sep;

   // Object length
   out.width(length_width); out.precision(default_precision);
   out << mask.length() << sep;

   // Object width
   out.width(width_width); out.precision(default_precision);
   out << mask.width() << sep;

   // Area of object
   out.width(area_width); out.precision(integer_precision);
   out << eng.fcst_single[n].area << sep;

   // Area of filtered object
   out.width(area_filter_width); out.precision(integer_precision);
   out << eng.fcst_single[n].area_filter << sep;

   // Object curvature
   curv = mask.curvature(xcurv, ycurv);
   out.width(curvature_width); out.precision(default_precision);
   out << curv << sep;

   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out.precision(default_precision);
   out << xcurv << sep;

   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out.precision(default_precision);
   out << ycurv << sep;

   // Object complexity
   out.width(complexity_width); out.precision(default_precision);
   out << eng.fcst_single[n].complexity << sep;

   // Percentiles of intensity
   ptile = get_percentiles(mask, *(eng.fcst_filter), eng.wconf.intensity_percentile().ival());

   // 10th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p10) << sep;

   // 25th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p25) << sep;

   // 50th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p50) << sep;

   // 75th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p75) << sep;

   // 90th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p90) << sep;

   // Specified percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.pth) << sep;

   // Sum of the object intensity values
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.sum) << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   out.width(centroid_dist_width); out << stat_missing_value << sep;

   // Distance between boundaries
   out.width(boundary_dist_width); out << stat_missing_value << sep;

   // Distance between convex hulls
   out.width(convex_hull_dist_width); out << stat_missing_value << sep;

   // Difference in angles in degrees
   out.width(angle_diff_width); out << stat_missing_value << sep;

   // Area ratio
   out.width(area_ratio_width); out << stat_missing_value << sep;

   // Intersection area
   out.width(intersection_area_width); out << stat_missing_value << sep;

   // Union area
   out.width(union_area_width); out << stat_missing_value << sep;

   // Symmetric difference area
   out.width(symmetric_diff_width); out << stat_missing_value << sep;

   // Intersection over area
   out.width(intersection_over_area_width); out << stat_missing_value << sep;

   // Complexity ratio
   out.width(complexity_ratio_width); out << stat_missing_value << sep;

   // Percentile intensity ratio
   out.width(percentile_intensity_ratio_width); out << stat_missing_value << sep;

   // Total interest value
   out.width(interest_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_obs_single(Engine &eng, const int n, const Grid &grid, ostream &out) {
   int i;
   double xcurv, ycurv, curv, lat, lon;
   char line[max_line_len];
   WrfData mask;
   DistributionPercentiles ptile;

   if(n >= eng.n_obs) {
      cerr << "\n\nERROR: write_obs_single(const Engine &, int, ostream) -> "
           << n << " >= number of obs, " << eng.n_obs << "\n\n" << flush;
      exit(1);
   }

   // Write out the common header columns
   write_header_columns(eng, out);

   // Object ID
   sprintf(line, "O%03d", (n+1));
   out.width(object_id_width); out << line << sep;

   // Object category
   i = eng.collection.obs_set_number(n + 1);
   sprintf(line, "CO%03d", (i + 1));
   out.width(object_cat_width); out << line << sep;

   out << "=" << sep;

   ////////////////////////////////////////////////////////////////////////////

   mask = select(*(eng.obs_split), n+1);
   grid.xy_to_latlon(eng.obs_single[n].xbar, eng.obs_single[n].ybar, lat, lon);

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << eng.obs_single[n].xbar << sep;

   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << eng.obs_single[n].ybar << sep;

   // Object centroid, latitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << lat << sep;

   // Object centroid, longitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << -1.0*lon << sep;

   // Axis angle
   out.width(axis_angle_width); out.precision(default_precision);
   out << eng.obs_single[n].angle_degrees << sep;

   // Object length
   out.width(length_width); out.precision(default_precision);
   out << mask.length() << sep;

   // Object width
   out.width(width_width); out.precision(default_precision);
   out << mask.width() << sep;

   // Area of object
   out.width(area_width); out.precision(integer_precision);
   out << eng.obs_single[n].area << sep;

   // Area of filtered object
   out.width(area_filter_width); out.precision(integer_precision);
   out << eng.obs_single[n].area_filter << sep;

   // Object curvature
   curv = mask.curvature(xcurv, ycurv);
   out.width(curvature_width); out.precision(default_precision);
   out << curv << sep;

   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out.precision(default_precision);
   out << xcurv << sep;

   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out.precision(default_precision);
   out << ycurv << sep;

   // Object complexity
   out.width(complexity_width); out.precision(default_precision);
   out << eng.obs_single[n].complexity << sep;

   // Percentiles of intensity
   ptile = get_percentiles(mask, *(eng.obs_filter), eng.wconf.intensity_percentile().ival());

   // 10th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p10) << sep;

   // 25th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p25) << sep;

   // 50th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p50) << sep;

   // 75th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p75) << sep;

   // 90th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p90) << sep;

   // Specified percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.pth) << sep;

   // Sum of the object intensity values
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.sum) << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   out.width(centroid_dist_width); out << stat_missing_value << sep;

   // Distance between boundaries
   out.width(boundary_dist_width); out << stat_missing_value << sep;

   // Distance between convex hulls
   out.width(convex_hull_dist_width); out << stat_missing_value << sep;

   // Difference in angles in degrees
   out.width(angle_diff_width); out << stat_missing_value << sep;

   // Area ratio
   out.width(area_ratio_width); out << stat_missing_value << sep;

   // Intersection area
   out.width(intersection_area_width); out << stat_missing_value << sep;

   // Union area
   out.width(union_area_width); out << stat_missing_value << sep;

   // Symmetric difference area
   out.width(symmetric_diff_width); out << stat_missing_value << sep;

   // Intersection over area
   out.width(intersection_over_area_width); out << stat_missing_value << sep;

   // Complexity ratio
   out.width(complexity_ratio_width); out << stat_missing_value << sep;

   // Percentile intensity ratio
   out.width(percentile_intensity_ratio_width); out << stat_missing_value << sep;

   // Total interest value
   out.width(interest_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_pair(Engine &eng, const int n_f, const int n_o, ostream &out) {
   int n;
   int fcst_i, obs_i;
   char line[max_line_len];
   WrfData mask;

   if(n_f >= eng.n_fcst || n_o >= eng.n_obs) {
      cerr << "\n\nERROR: write_pair(const Engine &, int, int, ostream) -> "
           << n_f << " >= number of fcst, " << eng.n_fcst << " or "
           << n_o << " >= number of obs, " << eng.n_obs << "\n\n" << flush;
      exit(1);
   }

   n = eng.two_to_one(n_f, n_o);

   //
   // Only dump out pair features if the total interest value for the pair
   // is greater than the print interest threshold
   //
   if( eng.info[eng.get_info_index(n)].interest_value
     < eng.wconf.print_interest_threshold().dval())  return;

   // Write out the common header columns
   write_header_columns(eng, out);

   // Object ID
   sprintf(line, "F%03d_O%03d", (n_f+1), (n_o+1));
   out.width(object_id_width); out << line << sep;

   // Object category
   fcst_i = eng.collection.fcst_set_number(n_f+1);
   obs_i  = eng.collection.obs_set_number(n_o+1);
   sprintf(line, "CF%03d_CO%03d", (fcst_i+1), (obs_i+1));
   out.width(object_cat_width); out << line << sep;

   out << "=" << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out << stat_missing_value << sep;

   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out << stat_missing_value << sep;

   // Object centroid, latitude
   out.width(centroid_latlon_width); out << stat_missing_value << sep;

   // Object centroid, longitude
   out.width(centroid_latlon_width); out << stat_missing_value << sep;

   // Axis angle
   out.width(axis_angle_width); out << stat_missing_value << sep;

   // Object length
   out.width(length_width); out << stat_missing_value << sep;

   // Object width
   out.width(width_width); out << stat_missing_value << sep;

   // Area of object
   out.width(area_width); out << stat_missing_value << sep;

   // Area of filtered object
   out.width(area_filter_width); out << stat_missing_value << sep;

   // Object curvature
   out.width(curvature_width); out << stat_missing_value << sep;

   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out << stat_missing_value << sep;

   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out << stat_missing_value << sep;

   // Object complexity
   out.width(complexity_width); out << stat_missing_value << sep;

   // 10th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 25th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 50th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 75th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 90th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // Specified percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // Sum of the object intensity values
   out.width(intensity_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   out.width(centroid_dist_width); out.precision(default_precision);
   out << eng.pair[n].centroid_dist << sep;

   // Distance between boundaries
   out.width(boundary_dist_width); out.precision(default_precision);
   out << eng.pair[n].boundary_dist << sep;

   // Distance between convex hulls
   out.width(convex_hull_dist_width); out.precision(default_precision);
   out << eng.pair[n].convex_hull_dist << sep;

   // Difference in angles in degrees
   out.width(angle_diff_width); out.precision(default_precision);
   out << eng.pair[n].angle_diff << sep;

   // Area ratio
   out.width(area_ratio_width); out.precision(default_precision);
   out << eng.pair[n].area_ratio << sep;

   // Intersection area
   out.width(intersection_area_width); out.precision(integer_precision);
   out << nint(eng.pair[n].intersection_area) << sep;

   // Union area
   out.width(union_area_width); out.precision(integer_precision);
   out << nint(eng.pair[n].union_area) << sep;

   // Symmetric difference area
   out.width(symmetric_diff_width); out.precision(integer_precision);
   out << nint(eng.pair[n].union_area - eng.pair[n].intersection_area) << sep;

   // Intersection over area
   out.width(intersection_over_area_width); out.precision(default_precision);
   out << eng.pair[n].intersection_area/(min(eng.pair[n].Obs->area, eng.pair[n].Fcst->area)) << sep;

   // Complexity ratio
   out.width(complexity_ratio_width); out.precision(default_precision);
   out << eng.pair[n].complexity_ratio << sep;

   // Percentile intensity ratio
   out.width(percentile_intensity_ratio_width); out.precision(default_precision);
   out << eng.pair[n].percentile_intensity_ratio << sep;

   // Total interest value
   out.width(interest_width); out.precision(default_precision);
   out << eng.info[eng.get_info_index(n)].interest_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_fcst_composite(Engine &eng, const int n, const Grid &grid, ostream &out) {
   double xbar, ybar, xcurv, ycurv, curv, lat, lon;
   char line[max_line_len];
   WrfData comp;
   DistributionPercentiles ptile;

   // Make composite mask
   comp = select(*eng.fcst_comp_split, n+1);

   // Write out the common header columns
   write_header_columns(eng, out);

   // Object ID
   sprintf(line, "CF%03d", (n+1));
   out.width(object_id_width); out << line << sep;

   // Object category
   out.width(object_cat_width); out << line << sep;

   out << "=" << sep;

   ////////////////////////////////////////////////////////////////////////////

   comp.centroid(xbar, ybar);
   grid.xy_to_latlon(xbar, ybar, lat, lon);

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << xbar << sep;

   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << ybar << sep;

   // Object centroid, latitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << lat << sep;

   // Object centroid, longitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << -1.0*lon << sep;

   // Axis angle
   out.width(axis_angle_width); out.precision(default_precision);
   out << comp.angle_degrees() << sep;

   // Object length
   out.width(length_width); out.precision(default_precision);
   out << comp.length() << sep;

   // Object width
   out.width(width_width); out.precision(default_precision);
   out << comp.width() << sep;

   // Area of object
   out.width(area_width); out.precision(integer_precision);
   out << comp.area() << sep;

   // Area of filtered object
   out.width(area_filter_width); out.precision(integer_precision);
   out << wrfdata_intersection(*eng.fcst_filter, comp) << sep;

   // Object curvature
   curv = comp.curvature(xcurv, ycurv);
   out.width(curvature_width); out.precision(default_precision);
   out << curv << sep;

   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out.precision(default_precision);
   out << xcurv << sep;

   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out.precision(default_precision);
   out << ycurv << sep;

   // Object complexity
   out.width(complexity_width); out.precision(default_precision);
   out << comp.complexity() << sep;

   // Percentiles of intensity
   ptile = get_percentiles(comp, *(eng.fcst_filter), eng.wconf.intensity_percentile().ival());

   // 10th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p10) << sep;

   // 25th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p25) << sep;

   // 50th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p50) << sep;

   // 75th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p75) << sep;

   // 90th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p90) << sep;

   // Specified percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.pth) << sep;

   // Sum of the object intensity values
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.sum) << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   out.width(centroid_dist_width); out << stat_missing_value << sep;

   // Distance between boundaries
   out.width(boundary_dist_width); out << stat_missing_value << sep;

   // Distance between convex hulls
   out.width(convex_hull_dist_width); out << stat_missing_value << sep;

   // Difference in angles in degrees
   out.width(angle_diff_width); out << stat_missing_value << sep;

   // Area ratio
   out.width(area_ratio_width); out << stat_missing_value << sep;

   // Intersection area
   out.width(intersection_area_width); out << stat_missing_value << sep;

   // Union area
   out.width(union_area_width); out << stat_missing_value << sep;

   // Symmetric difference area
   out.width(symmetric_diff_width); out << stat_missing_value << sep;

   // Intersection over area
   out.width(intersection_over_area_width); out << stat_missing_value << sep;

   // Complexity ratio
   out.width(complexity_ratio_width); out << stat_missing_value << sep;

   // Percentile intensity ratio
   out.width(percentile_intensity_ratio_width); out << stat_missing_value << sep;

   // Total interest value
   out.width(interest_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_obs_composite(Engine &eng, const int n, const Grid &grid, ostream &out) {
   double xbar, ybar, xcurv, ycurv, curv, lat, lon;
   char line[max_line_len];
   WrfData comp;
   DistributionPercentiles ptile;

   // Make composite mask
   comp = select(*eng.obs_comp_split, n+1);

   // Write out the common header columns
   write_header_columns(eng, out);

   // Object ID
   sprintf(line, "CO%03d", (n+1));
   out.width(object_id_width); out << line << sep;

   // Object category
   out.width(object_cat_width); out << line << sep;

   out << "=" << sep;

   ////////////////////////////////////////////////////////////////////////////

   comp.centroid(xbar, ybar);
   grid.xy_to_latlon(xbar, ybar, lat, lon);

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << xbar << sep;

   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out.precision(default_precision);
   out << ybar << sep;

   // Object centroid, latitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << lat << sep;

   // Object centroid, longitude
   out.width(centroid_latlon_width); out.precision(default_precision);
   out << -1.0*lon << sep;

   // Axis angle
   out.width(axis_angle_width); out.precision(default_precision);
   out << comp.angle_degrees() << sep;

   // Object length
   out.width(length_width); out.precision(default_precision);
   out << comp.length() << sep;

   // Object width
   out.width(width_width); out.precision(default_precision);
   out << comp.width() << sep;

   // Area of object
   out.width(area_width); out.precision(integer_precision);
   out << comp.area() << sep;

   // Area of filtered object
   out.width(area_filter_width); out.precision(integer_precision);
   out << wrfdata_intersection(*eng.obs_filter, comp) << sep;

   // Object curvature
   curv = comp.curvature(xcurv, ycurv);
   out.width(curvature_width); out.precision(default_precision);
   out << curv << sep;

   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out.precision(default_precision);
   out << xcurv << sep;

   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out.precision(default_precision);
   out << ycurv << sep;

   // Object complexity
   out.width(complexity_width); out.precision(default_precision);
   out << comp.complexity() << sep;

   // Percentiles of intensity
   ptile = get_percentiles(comp, *(eng.obs_filter), eng.wconf.intensity_percentile().ival());

   // 10th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p10) << sep;

   // 25th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p25) << sep;

   // 50th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p50) << sep;

   // 75th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p75) << sep;

   // 90th percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.p90) << sep;

   // Specified percentile of object intensity
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.pth) << sep;

   // Sum of the object intensity values
   out.width(intensity_width); out.precision(default_precision);
   out << (ptile.sum) << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   out.width(centroid_dist_width); out << stat_missing_value << sep;

   // Distance between boundaries
   out.width(boundary_dist_width); out << stat_missing_value << sep;

   // Distance between convex hulls
   out.width(convex_hull_dist_width); out << stat_missing_value << sep;

   // Difference in angles in degrees
   out.width(angle_diff_width); out << stat_missing_value << sep;

   // Area ratio
   out.width(area_ratio_width); out << stat_missing_value << sep;

   // Intersection area
   out.width(intersection_area_width); out << stat_missing_value << sep;

   // Union area
   out.width(union_area_width); out << stat_missing_value << sep;

   // Symmetric difference area
   out.width(symmetric_diff_width); out << stat_missing_value << sep;

   // Intersection over area
   out.width(intersection_over_area_width); out << stat_missing_value << sep;

   // Complexity ratio
   out.width(complexity_ratio_width); out << stat_missing_value << sep;

   // Percentile intensity ratio
   out.width(percentile_intensity_ratio_width); out << stat_missing_value << sep;

   // Total interest value
   out.width(interest_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

void write_composite_pair(Engine &eng, const int n_f,
                          const int n_o, ostream &out) {
   int x, y, int_area, union_area, sym_diff_area, fcst_inside, obs_inside;
   double value, fcst_xbar, fcst_ybar, obs_xbar, obs_ybar, dx, dy;
   char line[max_line_len];
   WrfData fcst_comp, obs_comp;
   DistributionPercentiles fcst_ptile, obs_ptile;
   Polyline fcst_hull, obs_hull;
   Node fcst_bndy, obs_bndy;

   // Make composite masks
   fcst_comp = select(*eng.fcst_comp_split, n_f+1);
   obs_comp  = select(*eng.obs_comp_split, n_o+1);

   // Write out the common header columns
   write_header_columns(eng, out);

   // Object ID
   sprintf(line, "CF%03d_CO%03d", (n_f+1), (n_o+1));
   out.width(object_id_width); out << line << sep;

   // Object category
   out.width(object_cat_width); out << line << sep;

   out << "=" << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Object centroid, x-coordinate
   out.width(centroid_xy_width); out << stat_missing_value << sep;

   // Object centroid, y-coordinate
   out.width(centroid_xy_width); out << stat_missing_value << sep;

   // Object centroid, latitude
   out.width(centroid_latlon_width); out << stat_missing_value << sep;

   // Object centroid, longitude
   out.width(centroid_latlon_width); out << stat_missing_value << sep;

   // Axis angle
   out.width(axis_angle_width); out << stat_missing_value << sep;

   // Object length
   out.width(length_width); out << stat_missing_value << sep;

   // Object width
   out.width(width_width); out << stat_missing_value << sep;

   // Area of object
   out.width(area_width); out << stat_missing_value << sep;

   // Area of filtered object
   out.width(area_filter_width); out << stat_missing_value << sep;

   // Object curvature
   out.width(curvature_width); out << stat_missing_value << sep;

   // Center of curvature, x-coordinate
   out.width(curvature_xy_width); out << stat_missing_value << sep;

   // Center of curvature, y-coordiante
   out.width(curvature_xy_width); out << stat_missing_value << sep;

   // Object complexity
   out.width(complexity_width); out << stat_missing_value << sep;

   // 10th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 25th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 50th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 75th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // 90th percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // Specified percentile of object intensity
   out.width(intensity_width); out << stat_missing_value << sep;

   // Sum of the object intensity values
   out.width(intensity_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   // Distance between centroids
   fcst_comp.centroid(fcst_xbar, fcst_ybar);
   obs_comp.centroid(obs_xbar, obs_ybar);
   dx = obs_xbar - fcst_xbar;
   dy = obs_ybar - fcst_ybar;
   out.width(centroid_dist_width); out.precision(default_precision);
   out << sqrt(dx*dx + dy*dy) << sep;

   // Distance between boundaries
   fcst_comp.composite_boundary(fcst_bndy);
   obs_comp.composite_boundary(obs_bndy);
   out.width(boundary_dist_width); out.precision(default_precision);
   out << node_dist(obs_bndy, fcst_bndy) << sep;

   // Distance between convex hulls
   fcst_hull = fcst_comp.convex_hull();
   obs_hull = obs_comp.convex_hull();
   out.width(convex_hull_dist_width); out.precision(default_precision);
   out << polyline_dist(obs_hull, fcst_hull) << sep;

   // Difference in angles in degrees
   out.width(angle_diff_width); out.precision(default_precision);
   out << fabs(obs_comp.angle_degrees() - fcst_comp.angle_degrees()) << sep;

   // Area ratio
   value = min( ((double) obs_comp.area())/((double) fcst_comp.area()),
                ((double) fcst_comp.area())/((double) obs_comp.area()) );
   out.width(area_ratio_width); out.precision(default_precision);
   out << value << sep;

   // Compute areas
   int_area = union_area = sym_diff_area = 0;
   for(x=0; x<(eng.fcst_raw->get_nx()); x++) {
      for(y=0; y<(eng.fcst_raw->get_ny()); y++) {

         fcst_inside = fcst_comp.get_xy_int(x, y);
         obs_inside = obs_comp.get_xy_int(x, y);

         if(fcst_inside && obs_inside) int_area++;
         if(fcst_inside || obs_inside) union_area++;
         if( (fcst_inside && !obs_inside) ||
             (!fcst_inside && obs_inside) ) sym_diff_area++;

      } // end for y
   } // end for x

   // Intersection area
   out.width(intersection_area_width); out.precision(integer_precision);
   out << int_area << sep;

   // Union area
   out.width(union_area_width); out.precision(integer_precision);
   out << union_area << sep;

   // Symmetric difference area
   out.width(symmetric_diff_width); out.precision(integer_precision);
   out << sym_diff_area << sep;

   // Intersection over area
   out.width(intersection_over_area_width); out.precision(default_precision);
   out << ((double) (int_area))/min( (double) obs_comp.area(),
                                     (double) fcst_comp.area() ) << sep;

   // Complexity ratio

   // Both complexities are non-zero
   if(obs_comp.complexity() > 0.0 && fcst_comp.complexity() > 0.0) {
      value = min( (obs_comp.complexity())/(fcst_comp.complexity()),
                   (fcst_comp.complexity())/(obs_comp.complexity()) );
   }
   // At least one complexity is zero
   else {
      value = 1.0 - max(obs_comp.complexity(), fcst_comp.complexity());
   }
   out.width(complexity_ratio_width); out.precision(default_precision);
   out << value << sep;

   // Percentile intensity ratio
   fcst_ptile = get_percentiles(fcst_comp, *(eng.fcst_filter), eng.wconf.intensity_percentile().ival());
   obs_ptile = get_percentiles(obs_comp, *(eng.obs_filter), eng.wconf.intensity_percentile().ival());

   // Both percentile intensities are non-zero
   if(obs_ptile.pth > 0.0 && fcst_ptile.pth > 0.0) {
      value = min(obs_ptile.pth/fcst_ptile.pth, fcst_ptile.pth/obs_ptile.pth);
   }
   // At least one median intensity is zero
   else {
      value = 0.0;
   }
   out.width(percentile_intensity_ratio_width); out.precision(default_precision);
   out << value << sep;

   // Total interest value
   out.width(interest_width); out << stat_missing_value << sep;

   ////////////////////////////////////////////////////////////////////////////

   out << "\n" << flush;

   //
   // Done
   //

   return;
}

///////////////////////////////////////////////////////////////////////////////

DistributionPercentiles get_percentiles(const WrfData &mask, const WrfData &raw, const double p) {
   int i, x, y, count, n_values;
   int nx, ny;
   double *v = (double *) 0;
   DistributionPercentiles ptile;

   nx = raw.get_nx();
   ny = raw.get_ny();

   //
   // Count values
   //
   n_values = 0;
   for(x=0; x<nx; ++x) {
      for(y=0; y<ny; ++y) {
         if((mask.f_is_on(x, y)) &&
            (raw.is_valid_data(x, y)) &&
            (raw.get_xy_int(x, y) > 0))  ++n_values;
      }
   }

   //
   // Allocate memory
   //
   v = new double [n_values];

   if(!v) {
      cerr << "\n\nERROR: get_percentiles(const WrfData &mask, const WrfData &raw, const double p) -> "
           << "memory allocation error\n\n" << flush;
      exit(1);
   }

   //
   // Fill values
   //
   count = 0;
   for(x=0; x<nx; ++x) {
      for(y=0; y<ny; ++y) {

         if((mask.f_is_on(x, y)) &&
            (raw.is_valid_data(x, y)) &&
            (raw.get_xy_int(x, y) > 0)) {

            v[count++] = (double) (raw.get_xy_double(x, y));
         }
      }
   }

   //
   // Sort
   //
   sort(v, n_values);

   //
   // Get percentiles
   //
   ptile.p10 = percentile(v, n_values, 0.10);
   ptile.p25 = percentile(v, n_values, 0.25);
   ptile.p50 = percentile(v, n_values, 0.50);
   ptile.p75 = percentile(v, n_values, 0.75);
   ptile.p90 = percentile(v, n_values, 0.90);
   ptile.pth = percentile(v, n_values, (double) p/100.0);

   //
   // Compute the sum
   //
   ptile.sum = 0;
   for(i=0; i<n_values; i++) {
      ptile.sum += v[i];
   }

   //
   // Free memory
   //
   if(v) { delete [] v;  v = (double *) 0; }

   //
   // Done
   //

   return(ptile);
}

///////////////////////////////////////////////////////////////////////////////

void calc_fcst_comp_ch_mask(const Engine &eng, WrfData &mask) {
   int i, x, y;
   WrfData comp;
   Polyline poly;
   BoundingBox bb;

   if(eng.need_fcst_comp_split) {
      cerr << "\n\nERROR: calc_fcst_comp_ch_mask -> "
           << "should not be called with need_fcst_comp_split set to true\n\n"
           << flush;

      exit(1);
   }

   //
   // Initialize convex hull mask wrfdata object
   //
   mask = *(eng.fcst_comp_split);
   zero_field(mask);

   //
   // Calculate fcst convex hull mask field for composite objects
   //
   for(i=0; i<eng.collection.n_sets; i++) {
      comp = select(*eng.fcst_comp_split, i+1);
      poly = comp.convex_hull();
      poly.bounding_box(bb);

      for(x = (int) floor(bb.x_ll-1); x <= (int) ceil(bb.x_ur+1); x++) {
         for(y = (int) floor(bb.y_ll-1); y <= (int) ceil(bb.y_ur+1); y++) {

            if(poly.is_inside(x, y)) {
               mask.put_xy_int(wrfdata_int_data_max, x, y);
            }
         } // end for y
      } // end for x
   } // end for i

   //
   // Calculate fcst convex hull mask field for unmatched singles
   //
   for(i=0; i<eng.n_fcst; i++) {

      if(eng.collection.fcst_set_number(i+1) != -1) continue;

      comp = select(*eng.fcst_split, i+1);
      poly = comp.convex_hull();
      poly.bounding_box(bb);

      for(x = (int) floor(bb.x_ll-1); x <= (int) ceil(bb.x_ur+1); x++) {
         for(y = (int) floor(bb.y_ll-1); y <= (int) ceil(bb.y_ur+1); y++) {

            if(poly.is_inside(x, y)) {
               mask.put_xy_int(wrfdata_int_data_max, x, y);
            }
         } // end for y
      } // end for x
   }

   return;
}

///////////////////////////////////////////////////////////////////////////////

void calc_obs_comp_ch_mask(const Engine &eng, WrfData &mask) {
   int i, x, y;
   WrfData comp;
   Polyline poly;
   BoundingBox bb;

   if(eng.need_obs_comp_split) {
      cerr << "\n\nERROR: calc_obs_comp_ch_mask -> "
           << "should not be called with need_obs_comp_split set to true\n\n"
           << flush;
      exit(1);
   }

   //
   // Initialize convex hull mask wrfdata object
   //
   mask = *(eng.obs_comp_split);
   zero_field(mask);

   //
   // Calculate obs convex hull mask field for composite objects
   //
   for(i=0; i<eng.collection.n_sets; i++) {
      comp = select(*eng.obs_comp_split, i+1);
      poly = comp.convex_hull();
      poly.bounding_box(bb);

      for(x=(int) floor(bb.x_ll-1); x<=ceil(bb.x_ur+1); x++) {
         for(y=(int) floor(bb.y_ll-1); y<=ceil(bb.y_ur+1); y++) {

            if(poly.is_inside(x, y)) {
               mask.put_xy_int(wrfdata_int_data_max, x, y);
            }
         } // end for y
      } // end for x
   } // end for i

   //
   // Calculate obs convex hull mask field for unmatched singles
   //
   for(i=0; i<eng.n_obs; i++) {

      if(eng.collection.obs_set_number(i+1) != -1) continue;

      comp = select(*eng.obs_split, i+1);
      poly = comp.convex_hull();
      poly.bounding_box(bb);

      for(x=(int) floor(bb.x_ll-1); x<=ceil(bb.x_ur+1); x++) {
         for(y=(int) floor(bb.y_ll-1); y<=ceil(bb.y_ur+1); y++) {

            if(poly.is_inside(x, y)) {
               mask.put_xy_int(wrfdata_int_data_max, x, y);
            }
         } // end for y
      } // end for x
   }

   return;
}

///////////////////////////////////////////////////////////////////////////////

void calc_fcst_composite_mask(const Engine &eng, WrfData &comp, const int n_set) {
   int i, x, y;
   WrfData junk;

   comp.set_size(eng.fcst_raw->get_nx(), eng.fcst_raw->get_ny());

   for(i=0; i<eng.n_fcst; i++) {

      if(!(eng.collection.set[n_set].has_fcst(i + 1)))  continue;

      junk = select(*(eng.fcst_split), i + 1);

      for(x=0; x<(eng.fcst_raw->get_nx()); x++) {
         for(y=0; y<(eng.fcst_raw->get_ny()); y++) {

            if(junk.get_xy_int(x, y)) {
               comp.put_xy_int(wrfdata_int_data_max, x, y);
            }
         } // end for y
      } // end for x
   } // end for i

   comp.calc_moments();

   return;
}

///////////////////////////////////////////////////////////////////////////////

void calc_obs_composite_mask(const Engine &eng, WrfData &comp, const int n_set) {
   int i, x, y;
   WrfData junk;

   comp.set_size(eng.obs_raw->get_nx(), eng.obs_raw->get_ny());

   for(i=0; i<eng.n_obs; i++) {

      if(!(eng.collection.set[n_set].has_obs(i + 1)))  continue;

      junk = select(*(eng.obs_split), i + 1);

      for(x=0; x<(eng.obs_raw->get_nx()); x++) {
         for(y=0; y<(eng.obs_raw->get_ny()); y++) {

            if(junk.get_xy_int(x, y)) {
               comp.put_xy_int(wrfdata_int_data_max, x, y);
            }
         } // end for y
      } // end for x
   } // end for i

   comp.calc_moments();

   return;
}

///////////////////////////////////////////////////////////////////////////////
