// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*

///////////////////////////////////////////////////////////////////////
//
//   Filename:   vsdb_analysis.cc
//
//   Description:
//      Based on user specified options, this tool performs a variety
//      of analysis jobs on the VSDB output of the Grid-Stat and/or
//      Point-Stat tools.
//
//   Mod#   Date      Name            Description
//   ----   ----      ----            -----------
//   000    09/05/07  Halley Gotway   New
//   001    07/01/08  Halley Gotway   Add the rank_corr_flag to the
//                    config file to disable computing rank correlations.
//
////////////////////////////////////////////////////////////////////////

using namespace std;

#include <cmath>
#include <cstdio>
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "vsdb_analysis_Conf.h"

#include "vx_analysis_util/vx_analysis_util.h"
#include "vx_util/vx_util.h"
#include "vx_met_util/vx_met_util.h"
#include "vx_contable/vx_contable.h"

////////////////////////////////////////////////////////////////////////

static const char * program_name = "vsdb_analysis";

////////////////////////////////////////////////////////////////////////

//
// Pointer to the random number generator to be used when computing
// bootstrap confidence intervals
//
static gsl_rng *rng_ptr = (gsl_rng *) 0;

//
// Default temporary path which will be overridden by the contents
// of the config file.
//
static const char  *default_tmp_dir = "/tmp";
static       char   tmp_dir[max_str_len];
static       char   tmp_file[max_str_len];
static       char   tmp_path[max_str_len];

//
// Output file stream for the temporary file
//
static ofstream tmp_out;

//
// Output file stream for the VSDB Analysis output.  Default to
// standard out but may be used to create an output file if the -out
// option is used.
//
static ofstream * va_out = (ofstream *) 0;

static vsdb_analysis_Conf conf;
static VSDBAnalysisJob default_job;

//
// Search directories and the corresponding files found
//
static StringArray search_dirs;
static StringArray files;

static int n_jobs      = 0;
static int n_lines_in  = 0;
static int n_lines_out = 0;

//
// Information from command-line switches
//
static const char *config_file = (const char *) 0;
static const char *out_file    = (const char *) 0;
static char        command_line_job[max_str_len] = "";
static int verbosity = 1;

////////////////////////////////////////////////////////////////////////

//
// Structure to store data from a single MPR matched pair VSDB line
//
struct MPRData {
   int    gc;
   int    total,   index;
   double obs_lat, obs_lon, obs_lvl, obs_elv;
   double fcst,    obs,     climo;
};

//
// Structure to store data from a single ISC VSDB line
//
struct ISCData {
   int    total, nscale;
   double mse,   isc;
   double fen,   oen;
   double baser, fbias;
};

////////////////////////////////////////////////////////////////////////

static void parse_command_line(int &argc, char **argv);
static void sanity_check();
static void usage();
static void set_default_job();
static void process_search_dirs();
static void process_vsdb_file(const char *);
static void process_job(const char *);
static void clean_up();

static void do_job(const char *, VSDBAnalysisJob &);
static void do_job_filter(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);
static void do_job_summary(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);
static void do_job_aggr(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);
static void do_job_aggr_cts(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);
static void do_job_aggr_cnt(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);
static void do_job_aggr_mpr(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);
static void do_job_go_index(const char *, LineDataFile &,
               VSDBAnalysisJob &, int &, int &);

static void sum_contable_lines(const char *, LineDataFile &,
               VSDBAnalysisJob &, CTSInfo &,
               VSDBLineType, int &, int &);
static void sum_partial_sum_lines(const char *, LineDataFile &,
               VSDBAnalysisJob &, SL1L2Info &, VL1L2Info &,
               VSDBLineType, int &, int &);
static void sum_mpr_lines_ct(const char *, LineDataFile &,
               VSDBAnalysisJob &, CTSInfo &,
               int &, int &);
static void sum_mpr_lines_stats(const char *, LineDataFile &,
               VSDBAnalysisJob &, CNTInfo &, CTSInfo &,
               int &, int &);
static void sum_mpr_lines_psums(const char *, LineDataFile &,
               VSDBAnalysisJob &, SL1L2Info &,
               int &, int &);
static void sum_isc_lines(const char *, LineDataFile &,
               VSDBAnalysisJob &, ISCInfo &,
               int &, int &);

static void parse_fho_ctable(VSDBLine &, TTContingencyTable &);
static void parse_ctc_ctable(VSDBLine &, TTContingencyTable &);
static void parse_ctp_ctable(VSDBLine &, TTContingencyTable &);
static void parse_cfp_ctable(VSDBLine &, TTContingencyTable &);
static void parse_cop_ctable(VSDBLine &, TTContingencyTable &);
static void parse_nbrctc_ctable(VSDBLine &, TTContingencyTable &);

static void parse_sl1l2_line (VSDBLine &, SL1L2Info &);
static void parse_sal1l2_line(VSDBLine &, SL1L2Info &);
static void parse_vl1l2_line (VSDBLine &, VL1L2Info &);
static void parse_val1l2_line(VSDBLine &, VL1L2Info &);

static void parse_mpr_line(VSDBLine &, MPRData &);

static void parse_isc_line(VSDBLine &, ISCData &);
static int  parse_iscale(const char *);

static double compute_go_index(const char *, LineDataFile &,
                               VSDBAnalysisJob &, int &, int &);

static double compute_sl1l2_rmse(SL1L2Info &, const char *, const char *, int);

static void write_jobstring(const char *);
static void write_line(const char *);

//
// Routines for command-line switches
//
static void set_config(const char *);
static void set_search_dir(const char *);
static void set_out_file(const char *);
static void set_verbosity(int);
static int  is_switch(const char *);

////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv []) {
   int i, n;
   Result r;

   //
   // Parse the command line
   //
   parse_command_line(argc, argv);

   //
   // If a config file was specified, set up the default job using
   // the config file.
   //
   if(config_file != (char *) 0) {

      //
      // Read the config file
      //
      conf.read(config_file);

      //
      // Sanity check the command line and config file options
      //
      sanity_check();

      //
      // Setup the default job using the options specified in the config
      // file
      //
      set_default_job();
   }

   //
   // Enclose within a try block to catch any run time errors,
   // and delete the temp file before exiting.
   //
   try {

      //
      // Process the VSDB files found in the search directories.
      //
      process_search_dirs();

      //
      // If a config file was specified, process the jobs in the config
      // file.
      //
      if(config_file != (char *) 0) {

         n = conf.n_joblist_elements();

         for(i=0; i<n; i++) {
            r = conf.joblist(i);
            process_job(r.sval());
         }
      }
      //
      // Otherwise, process the job specified on the command line.
      //
      else {
         process_job(command_line_job);
      }

   }
   catch(int j) { // Catch an error

      cerr << "\n\nERROR: main() -> "
           << "encountered an error value of " << j
           << ".  Calling clean_up() and usage() before exiting.\n\n"
           << flush;

      clean_up();
      usage();
      exit(j);
   }

   //
   // Deallocate memory and clean up
   //
   clean_up();

   return(0);
}

////////////////////////////////////////////////////////////////////////

void parse_command_line(int &argc, char **argv) {
   int i;

   //
   // Check for the minimum number of required arguments
   //
   if(argc < 5) {
      usage();
      exit(1);
   }

   //
   // Parse the command line arguments
   //
   config_file = (char *) 0;

   for(i=1; i<argc; i++) {

      if(strcmp(argv[i], "-config") == 0){
         set_config(argv[i+1]);
         i++;
      }
      else if(strcmp(argv[i], "-lookin") == 0) {
         while(i+1<argc && !is_switch(argv[i+1])) {
            set_search_dir(argv[i+1]);
            i++;
         }
      }
      else if(strcmp(argv[i], "-out") == 0) {
         set_out_file(argv[i+1]);
         i++;
      }
      else if(strcmp(argv[i], "-v") == 0) {
         set_verbosity(atoi(argv[i+1]));
         i++;
      }
   } // end for

   //
   // If no config file was specified, parse out the VSDB Analysis job
   // command from the command line.
   //
   if(config_file == (char *) 0) {

      for(i=1; i<argc; i++) {

         //
         // Skip the command line options not specific to the
         // VSDB Analysis job
         //
         if(strcmp(argv[i], "-out") == 0 ||
            strcmp(argv[i], "-v") == 0) {
            i++;
            continue;
         }
         else if(strcmp(argv[i], "-lookin") == 0) {

            while(i+1<argc && !is_switch(argv[i+1])) i++;

            continue;
         }
         //
         // Store the remaining options as the VSDB Command Line job
         //
         else {
            if(strlen(command_line_job) == 0)
               sprintf(command_line_job, "%s", argv[i]);
            else
               sprintf(command_line_job, "%s %s", command_line_job, argv[i]);
         }
      } // end for
   } // end if

   return;
}

////////////////////////////////////////////////////////////////////////

void sanity_check() {
   unixtime ut_beg, ut_end;

   //
   // Check for at least one search file or directory
   //
   if(search_dirs.n_elements() == 0) {
      cerr << "\n\nERROR: sanity_check() -> "
           << "no VSDB search files or directories specified!\n\n"
           << flush;

      exit(1);
   }

   //
   // Check for valid_beg > valid_end
   //
   ut_beg = timestring_to_unix(conf.valid_beg().sval());
   ut_end = timestring_to_unix(conf.valid_end().sval());

   if((ut_beg > 0) && (ut_end > 0) && (ut_beg > ut_end)) {
      cerr << "\n\nERROR:sanity_check() -> "
           << "valid_beg is after valid_end: "
           << ut_beg << " > " << ut_end << "!\n\n";

      exit(1);
   }

   //
   // Check for init_beg > init_end
   //
   ut_beg = timestring_to_unix(conf.init_beg().sval());
   ut_end = timestring_to_unix(conf.init_end().sval());

   if((ut_beg > 0) && (ut_end > 0) && (ut_beg > ut_end)) {
      cerr << "\n\nERROR:sanity_check() -> "
           << "init_beg is after init_end: "
           << ut_beg << " > " << ut_end << "!\n\n";

      exit(1);
   }

   //
   // Conf: rank_corr_flag
   //
   if(conf.rank_corr_flag().ival() != 0 &&
      conf.rank_corr_flag().ival() != 1) {
      cerr << "\n\nERROR: sanity_check() -> "
           << "The rank_corr_flag (" << conf.rank_corr_flag().ival()
           << ") must be set to 0 or 1.\n\n"
           << flush;
      exit(1);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void set_config(const char *path) {

   config_file = path;

   return;
}

////////////////////////////////////////////////////////////////////////

void set_search_dir(const char *path) {

   search_dirs.add(path);

   return;
}

////////////////////////////////////////////////////////////////////////

void set_out_file(const char *path) {

   out_file = path;

   //
   // Create an output file and set the va_out ofstream to it.
   //
   va_out = new ofstream;
   va_out->open(out_file);

   if(!va_out) {
      cerr << "\n\nERROR: set_out_file()-> "
           << "can't open the output file \"" << out_file
           << "\" for writing|\n\n" << flush;

      throw(1);
   }

   if(verbosity > 0) {
      cout << "Creating " << program_name << " output file \""
           << out_file << "\"\n" << flush;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void set_verbosity(int i) {

   verbosity = i;

   return;
}

////////////////////////////////////////////////////////////////////////

int is_switch(const char *c) {
   int r;

   if(strlen(c) > 0 && c[0] == '-') r = 1;
   else                             r = 0;

   return(r);
}

////////////////////////////////////////////////////////////////////////

void set_default_job() {
   int i, n;
   Result r;

   //
   // Clear out the default job
   //
   default_job.clear();

   //
   // Get info from config file and store in "default_job"
   //

   //
   // model
   //
   n = conf.n_model_elements();

   for(i=0; i<n; i++) {
      r = conf.model(i);
      default_job.model.add(r.sval());
   }

   //
   // fcst_lead
   //
   n = conf.n_fcst_lead_elements();

   for(i=0; i<n; i++) {
      r = conf.fcst_lead(i);
      default_job.fcst_lead.add(timestring_to_sec(r.sval()));
   }

   //
   // init_beg
   //
   r = conf.init_beg();

   if(strlen(r.sval()) > 0)
      default_job.init_beg = timestring_to_unix(r.sval());

   //
   // init_end
   //
   r = conf.init_end();

   if(strlen(r.sval()) > 0)
      default_job.init_end = timestring_to_unix(r.sval());

   //
   // init_hour
   //
   n = conf.n_init_hour_elements();

   for(i=0; i<n; i++) {
      r = conf.init_hour(i);
      default_job.init_hour.add(timestring_to_sec(r.sval()));
   }

   //
   // valid_beg
   //
   r = conf.valid_beg();

   if(strlen(r.sval()) > 0)
      default_job.valid_beg = timestring_to_unix(r.sval());

   //
   // valid_end
   //
   r = conf.valid_end();

   if(strlen(r.sval()) > 0)
      default_job.valid_end = timestring_to_unix(r.sval());

   //
   // obtype
   //
   n = conf.n_obtype_elements();

   for(i=0; i<n; i++) {
      r = conf.obtype(i);
      default_job.obtype.add(r.sval());
   }

   //
   // mask
   //
   n = conf.n_vx_mask_elements();

   for(i=0; i<n; i++) {
      r = conf.vx_mask(i);
      default_job.mask.add(r.sval());
   }

   //
   // line_type
   //
   n = conf.n_line_type_elements();

   for(i=0; i<n; i++) {
      r = conf.line_type(i);
      default_job.line_type.add(r.sval());
   }

   //
   // var
   //
   n = conf.n_var_elements();

   for(i=0; i<n; i++) {
      r = conf.var(i);
      default_job.var.add(r.sval());
   }

   //
   // level
   //
   n = conf.n_level_elements();

   for(i=0; i<n; i++) {
      r = conf.level(i);
      default_job.level.add(r.sval());
   }

   //
   // interp_mthd
   //
   n = conf.n_interp_mthd_elements();

   for(i=0; i<n; i++) {
      r = conf.interp_mthd(i);
      default_job.interp_mthd.add(r.sval());
   }

   //
   // interp_pnts
   //
   n = conf.n_interp_pnts_elements();

   for(i=0; i<n; i++) {
      r = conf.interp_pnts(i);
      default_job.interp_pnts.add(r.ival());
   }

   //
   // alpha
   //
   default_job.alpha = conf.alpha().dval();

   //
   // boot_interval
   //
   default_job.boot_interval = conf.boot_interval().ival();

   //
   // boot_rep_prop
   //
   default_job.boot_rep_prop = conf.boot_rep_prop().dval();

   //
   // n_boot_rep
   //
   default_job.n_boot_rep = conf.n_boot_rep().ival();

   //
   // boot_rng
   //
   default_job.set_boot_rng(conf.boot_rng().sval());

   //
   // boot_seed
   //
   default_job.set_boot_seed(conf.boot_seed().sval());

   return;
}

////////////////////////////////////////////////////////////////////////

void process_search_dirs() {
   int n, i, j, max_len;
   pid_t p;

   //
   // Build the temp file name
   //
   p = getpid();
   sprintf(tmp_file, "tmp_%i.vsdb", p);

   //
   // Open up the temp file for storing the intermediate VSDB line data
   //
   if(config_file != (char *) 0) strcpy(tmp_dir, conf.tmp_dir().sval());
   else                          strcpy(tmp_dir, default_tmp_dir);

   sprintf(tmp_path, "%s/%s", tmp_dir, tmp_file);
   tmp_out.open(tmp_path);
   if(!tmp_out) {
      cerr << "\n\nERROR: process_search_dirs() -> "
           << "can't open the temporary file \"" << tmp_path
           << "\" for writing!\n\n" << flush;

      throw(1);
   }

   //
   // Get the list of vsdb files in the search directories
   //
   files = get_vsdb_filenames(search_dirs);

   n = files.n_elements();

   if(n == 0) {
      cerr << "\n\nERROR: process_search_dirs() -> "
           << "no VSDB files found in the directories specified!\n\n"
           << flush;

      throw(1);
   }

   //
   // Go through each input file
   //
   max_len = 0;

   for(i=0; i<n; i++)  {
      j = strlen(files[i]);
      if(j > max_len) max_len = j;
   }
   max_len += 3;

   for(i=0; i<n; i++) {
      if(verbosity > 2) {

         cout << "Processing VSDB file \"" << files[i] << "\" ";

         for(j=strlen(files[i]); j<max_len; j++) cout.put('.');

         cout << " " << i+1 << " of " << n << "\n";

         if((j%5) == 4) cout.put('\n');

         cout.flush();
      }

      process_vsdb_file(files[i]);
   }

   if(verbosity > 1) {
      cout << "VSDB Lines read     = " << n_lines_in << "\n";
      cout << "VSDB Lines retained = " << n_lines_out << "\n";
   }

   tmp_out.close();

   return;
}

////////////////////////////////////////////////////////////////////////

void process_vsdb_file(const char *filename) {
   LineDataFile f;
   VSDBLine line;

   if(!(f.open(filename))) {
      cerr << "\n\nERROR: process_vsdb_file() -> "
           << "unable to open input vsdb file \""
           << filename << "\"\n\n" << flush;

      throw(1);
   }

   while(f >> line) {

      n_lines_in++;

      if(default_job.is_keeper(line)) {

         n_lines_out++;

         tmp_out << line;
      }
   } // end while

   f.close();

   return;
}

////////////////////////////////////////////////////////////////////////

void process_job(const char * jobstring) {
   VSDBAnalysisJob job;
   char full_jobstring[1024];

   //
   // Increment the jobs counter
   //
   n_jobs++;

   //
   // Initialize to the default job
   //
   job = default_job;

   //
   // Parse the job command line options
   //
   job.parse_job_command(jobstring);

   //
   // Get the full jobstring
   //
   job.get_jobstring(full_jobstring);

   //
   // Do the job
   //
   do_job(full_jobstring, job);

   return;
}

////////////////////////////////////////////////////////////////////////

void do_job(const char *jobstring, VSDBAnalysisJob &j) {
   LineDataFile f;
   int n_in, n_out;

   //
   // Open up the temp file for reading the intermediate VSDB line data
   //
   if(!f.open(tmp_path)) {
      cerr << "\n\nERROR: do_job() -> "
           << "can't open the temporary file \"" << tmp_path
           << "\" for reading!\n\n" << flush;

      throw(1);
   }

   //
   // Initialize n_in and n_out to keep track of the number of lines
   // read and retained.
   //
   n_in = n_out = 0;

   if(verbosity > 1) {
      cout << "Job " << n_jobs << ": " << jobstring << "\n" << flush;
   }

   //
   // Switch on the job type
   //
   switch(j.job_type) {

      case(vsdb_job_filter):
         do_job_filter(jobstring, f, j, n_in, n_out);
         break;

      case(vsdb_job_summary):
         do_job_summary(jobstring, f, j, n_in, n_out);
         break;

      case(vsdb_job_aggr):
         do_job_aggr(jobstring, f, j, n_in, n_out);
         break;

      case(vsdb_job_aggr_cts):
         do_job_aggr_cts(jobstring, f, j, n_in, n_out);
         break;

      case(vsdb_job_aggr_cnt):
         do_job_aggr_cnt(jobstring, f, j, n_in, n_out);
         break;

      case(vsdb_job_aggr_mpr):
         do_job_aggr_mpr(jobstring, f, j, n_in, n_out);
         break;

      case(vsdb_job_go_index):
         do_job_go_index(jobstring, f, j, n_in, n_out);
         break;

      default:
         cerr << "\n\nERROR: do_job() -> "
              << "jobtype value of " << j.job_type
              << " not currently supported!\n\n" << flush;

         throw(1);
   }

   if(verbosity > 1) {
      cout << "Job " << n_jobs << " used " << n_out << " out of "
           << n_in << " VSDB lines.\n" << flush;
   }

   //
   // Close the input file stream
   //
   f.close();

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_filter() routine should only be called when the -dump_row
// option has been selected to dump out the filtered VSDB lines.
//
////////////////////////////////////////////////////////////////////////

void do_job_filter(const char *jobstring, LineDataFile &f,
                   VSDBAnalysisJob &j, int &n_in, int &n_out) {
   char out_line[max_line_len];
   ofstream * dr_out = (ofstream *) 0;
   VSDBLine line;

   //
   // Check that the -dump_row option has been supplied
   //
   if(j.dump_row.n_elements() == 0) {
      cerr << "\n\nERROR: do_job_filter()-> "
           << "this function may only be called when using the "
           << "-dump_row option in the job command line: "
           << jobstring << "\n\n" << flush;

      throw(1);
   }

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: do_job_filter()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

         throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   //
   // Build a simple output line
   //
   sprintf(out_line, "%-15s %s", "FILTER:", jobstring);

   //
   // Write the filter line generated
   //
   write_line(out_line);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_summary() routine should only be called when the
// -line_type option has been used exactly once and the -column option
// has been used to specify the column of data to summarize.
//
////////////////////////////////////////////////////////////////////////

void do_job_summary(const char *jobstring, LineDataFile &f,
                    VSDBAnalysisJob &j, int &n_in, int &n_out) {
   VSDBLine line;
   VSDBLineType lt;
   NumArray v_array;
   int offset;
   double v, min, v10, v25, v50, v75, v90, max;
   CIInfo mean_ci, stdev_ci;
   char fmt_str[max_str_len];
   char hdr_line[max_line_len], col_names[max_line_len];
   char out_line[max_line_len], col_str[max_line_len];
   char var_str[max_vsdb_col][vsdb_col_len];
   ofstream *dr_out = (ofstream *) 0;

   //
   // Check that the -line_type option has been supplied only once
   //
   if(j.line_type.n_elements() != 1) {
      cerr << "\n\nERROR: do_job_summary()-> "
           << "this function may only be called when the "
           << "\"-line_type\" option has been used exactly once to "
           << "specify a single line type from which to select a "
           << "statistic to summarize: " << jobstring << "\n\n"
           << flush;

         throw(1);
   }

   //
   // Check that the -column option has been supplied
   //
   if(j.col_name == (char *) 0) {
      cerr << "\n\nERROR: do_job_summary()-> "
           << "this function may only be called when the "
           << "\"-column\" option has been used to specify a "
           << "single column from which to select a statistic "
           << "to summarize: " << jobstring << "\n\n"
           << flush;

         throw(1);
   }

   //
   // Determine the line type
   //
   lt = determine_line_type(j.line_type[0]);

   //
   // Based on the line type and the column name selected, determine
   // the column offset to use.
   //
   offset = determine_column_offset(lt, j.col_name);

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: do_job_summary()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

         throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         v = atof(line.get_item(offset));

         if(v != bad_data_double) v_array.add(v);

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   //
   // Check for no matching VSDB lines
   //
   if(v_array.n_elements() == 0) {
      cout << "WARNING: do_job_summary() -> "
           << "no valid data found in the VSDB lines for job: " << jobstring
           << "\n" << flush;
      return;
   }

   //
   // Compute the summary information for this collection of values:
   // min, max, v10, v25, v50, 75, v90
   //
   v_array.sort_array();
   min = v_array.percentile_array(0.00);
   v10 = v_array.percentile_array(0.10);
   v25 = v_array.percentile_array(0.25);
   v50 = v_array.percentile_array(0.50);
   v75 = v_array.percentile_array(0.75);
   v90 = v_array.percentile_array(0.90);
   max = v_array.percentile_array(1.00);

   //
   // Set up CIInfo objects
   //
   mean_ci.allocate_n_alpha(1);
   stdev_ci.allocate_n_alpha(1);

   //
   // Set up the random number generator and seed value
   //
   rng_set(rng_ptr, j.boot_rng, j.boot_seed);

   //
   // Compute a bootstrap confidence interval for the mean of this
   // array of values.
   //
   if(j.boot_interval == boot_bca_flag) {
      compute_mean_stdev_ci_bca(rng_ptr, v_array,
                                j.n_boot_rep, j.alpha,
                                mean_ci, stdev_ci);
   }
   else {
      compute_mean_stdev_ci_perc(rng_ptr, v_array,
                                 j.n_boot_rep, j.boot_rep_prop, j.alpha,
                                 mean_ci, stdev_ci);
   }

   //
   // Convert to strings
   //
   sprintf(var_str[0], "%i",      v_array.n_elements());
   dbl_to_str(mean_ci.v,          var_str[1]);
   dbl_to_str(mean_ci.v_ncl[0],   var_str[2]);
   dbl_to_str(mean_ci.v_ncu[0],   var_str[3]);
   dbl_to_str(mean_ci.v_bcl[0],   var_str[4]);
   dbl_to_str(mean_ci.v_bcu[0],   var_str[5]);
   dbl_to_str(stdev_ci.v,         var_str[6]);
   dbl_to_str(stdev_ci.v_bcl[0],  var_str[7]);
   dbl_to_str(stdev_ci.v_bcu[0],  var_str[8]);
   dbl_to_str(min,                var_str[9]);
   dbl_to_str(v10,                var_str[10]);
   dbl_to_str(v25,                var_str[11]);
   dbl_to_str(v50,                var_str[12]);
   dbl_to_str(v75,                var_str[13]);
   dbl_to_str(v90,                var_str[14]);
   dbl_to_str(max,                var_str[15]);

   //
   // Get the column names
   //
   get_column_names(job_sum_columns, n_job_sum_columns, col_names);
   sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);

   //
   // Setup the format string
   //
   get_column_format(job_sum_columns, n_job_sum_columns, fmt_str);

   //
   // Build the job summary output line
   //
   sprintf(col_str, fmt_str,
           var_str[0],  var_str[1],  var_str[2],
           var_str[3],  var_str[4],  var_str[5],
           var_str[6],  var_str[7],  var_str[8],
           var_str[9],  var_str[10], var_str[11],
           var_str[12], var_str[13], var_str[13],
           var_str[14]);
   sprintf(out_line, "%-15s %s", "SUMMARY:", col_str);

   //
   // Write the column names line, the summary line and job
   // command line
   //
   write_line(hdr_line);
   write_line(out_line);
   write_jobstring(jobstring);

   //
   // Deallocate memory for the random number generator
   //
   rng_free(rng_ptr);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_aggr() routine should only be called when the -line_type
// option has been used exactly once.
//
////////////////////////////////////////////////////////////////////////

void do_job_aggr(const char *jobstring, LineDataFile &f,
                 VSDBAnalysisJob &j, int &n_in, int &n_out) {
   VSDBLine line;
   VSDBLineType lt;
   CTSInfo cts_info;
   NBRCTSInfo nbrcts_info;
   SL1L2Info sl1l2_info;
   VL1L2Info vl1l2_info;
   ISCInfo   isc_info;
   char hdr_line[max_line_len], col_names[max_line_len];
   char out_line[max_line_len], col_str[max_line_len];
   char line_type[max_str_len];

   //
   // Check that the -line_type option has been supplied only once
   //
   if(j.line_type.n_elements() != 1) {
      cerr << "\n\nERROR: do_job_aggr()-> "
           << "this function may only be called when the "
           << "\"-line_type\" option has been used exactly once to "
           << "specify a single line type over which to perform the "
           << "aggregation: " << jobstring << "\n\n"
           << flush;

         throw(1);
   }

   //
   // Determine the line type specified for this job
   //
   lt = determine_line_type(j.line_type[0]);

   //
   // Check that a valid line type has been selected
   //
   if(lt != vsdb_fho    && lt != vsdb_ctc &&
      lt != vsdb_ctp    && lt != vsdb_cfp &&
      lt != vsdb_cop    &&
      lt != vsdb_sl1l2  && lt != vsdb_sal1l2 &&
      lt != vsdb_vl1l2  && lt != vsdb_val1l2 &&
      lt != vsdb_nbrctc && lt != vsdb_isc) {
      cerr << "\n\nERROR: do_job_aggr()-> "
           << "the \"-line_type\" option must be set to one of:\n"
           << "\tFHO, CTC, CTP, CFP, COP\n"
           << "\tSL1L2, SAL1L2, VL1L2, VAL1L2, NBRCTC, or ISC\n\n"
           << flush;

      throw(1);
   }

   //
   // Sum up the contingency table type lines
   // FHO, CTC, CTP, CFP, COP, and NBRCTC
   //
   if(lt == vsdb_fho || lt == vsdb_ctc ||
      lt == vsdb_ctp || lt == vsdb_cfp ||
      lt == vsdb_cop || lt == vsdb_nbrctc) {
      sum_contable_lines(jobstring, f, j, cts_info, lt, n_in, n_out);
   }
   //
   // Sum the partial sum line types
   // SL1L2, SAL1L2, VL1L2, and VAL1L2
   //
   else if(lt == vsdb_sl1l2 || lt == vsdb_sal1l2 ||
           lt == vsdb_vl1l2 || lt == vsdb_val1l2) {
      sum_partial_sum_lines(jobstring, f, j, sl1l2_info, vl1l2_info, lt, n_in, n_out);
   }
   //
   // Sum the ISC line types
   //
   else if(lt == vsdb_isc) {
      sum_isc_lines(jobstring, f, j, isc_info, n_in, n_out);
   }

   //
   // For NBRCTS line type, copy the CTSInfo over into the NBRCTSInfo object
   //
   if(lt == vsdb_nbrctc) nbrcts_info.cts_info = cts_info;

   //
   // Check for no matching VSDB lines
   //
   if(n_out == 0) {
      cout << "WARNING: do_job_aggr() -> "
           << "no matching VSDB lines found for job: " << jobstring
           << "\n" << flush;
      return;
   }

   //
   // Switch on the line type to determine the format of the output
   // line
   //
   switch(lt) {

      case(vsdb_fho):

         //
         // Build the column names line and the output FHO line
         //
         get_column_names(fho_columns, n_fho_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_fho_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "FHO:", col_str);

         break;

      case(vsdb_ctc):

         //
         // Build the column names line and the output CTC line
         //
         get_column_names(ctc_columns, n_ctc_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_ctc_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "CTC:", col_str);

         break;

      case(vsdb_ctp):

         //
         // Build the column names line and the output CTP line
         //
         get_column_names(ctp_columns, n_ctp_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_ctp_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "CTP:", col_str);

         break;

      case(vsdb_cfp):

         //
         // Build the column names line and the output CFP line
         //
         get_column_names(cfp_columns, n_cfp_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_cfp_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "CFP:", col_str);

         break;

      case(vsdb_cop):

         //
         // Build the column names line and the output COP line
         //
         get_column_names(cop_columns, n_cop_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_cop_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "COP:", col_str);

         break;

      case(vsdb_sl1l2):

         //
         // Build the column names line and the output SL1L2 line
         //
         get_column_names(sl1l2_columns, n_sl1l2_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_sl1l2_str(col_str, sl1l2_info);
         sprintf(out_line, "%-15s %s", "SL1L2:", col_str);

         break;

      case(vsdb_sal1l2):

         //
         // Build the column names line and the output SAL1L2 line
         //
         get_column_names(sal1l2_columns, n_sal1l2_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_sal1l2_str(col_str, sl1l2_info);
         sprintf(out_line, "%-15s %s", "SAL1L2:", col_str);

         break;

      case(vsdb_vl1l2):

         //
         // Build the column names line and the output VL1L2 line
         //
         get_column_names(vl1l2_columns, n_vl1l2_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_vl1l2_str(col_str, vl1l2_info);
         sprintf(out_line, "%-15s %s", "VL1L2:", col_str);

         break;

      case(vsdb_val1l2):

         //
         // Build the column names line and the output VAL1L2 line
         //
         get_column_names(val1l2_columns, n_val1l2_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_val1l2_str(col_str, vl1l2_info);
         sprintf(out_line, "%-15s %s", "VAL1L2:", col_str);

         break;

      case(vsdb_nbrctc):

         //
         // Build the column names line and the output NBRCTC line
         //
         get_column_names(nbrctc_columns, n_nbrctc_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_nbrctc_str(col_str, nbrcts_info);
         sprintf(out_line, "%-15s %s", "NBRCTC:", col_str);

         break;

      case(vsdb_isc):

         //
         // Build the column names line and the output ISC line
         //
         get_column_names(isc_columns, n_isc_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_isc_str(col_str, isc_info, -1);
         sprintf(out_line, "%-15s %s", "ISC:", col_str);

         break;

      default:

         vsdblinetype_to_string(line.type(), line_type);
         cerr << "\n\nERROR: do_job_aggr() -> "
              << "line type value of " << line_type
              << " not currently supported for the aggregation "
              << "job!\n\n" << flush;

         throw(1);
   } // end switch

   //
   // Write the column names line, the aggregated line, and the
   // job command line
   //
   write_line(hdr_line);
   write_line(out_line);
   write_jobstring(jobstring);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_aggr_cts() routine should only be called when the
// -line_type option has been used exactly once for line types:
// FHO, CTC, CTP, CFP, and COP
//
////////////////////////////////////////////////////////////////////////

void do_job_aggr_cts(const char *jobstring, LineDataFile &f,
                     VSDBAnalysisJob &j, int &n_in, int &n_out) {
   VSDBLine line;
   VSDBLineType lt;
   CTSInfo cts_info;
   char hdr_line[max_line_len], col_names[max_line_len];
   char out_line[max_line_len], col_str[max_line_len];

   //
   // Check that the -line_type option has been supplied only once
   //
   if(j.line_type.n_elements() != 1) {
      cerr << "\n\nERROR: do_job_aggr_cts()-> "
           << "this function may only be called when the "
           << "\"-line_type\" option has been used exactly once to "
           << "specify a single contingency table line type from "
           << "which to derive the statistics: " << jobstring << "\n\n"
           << flush;

      throw(1);
   }

   //
   // Determine the line type specified for this job
   //
   lt = determine_line_type(j.line_type[0]);

   //
   // Check that a valid line type has been selected
   //
   if(lt != vsdb_fho && lt != vsdb_ctc &&
      lt != vsdb_ctp && lt != vsdb_cfp &&
      lt != vsdb_cop) {
      cerr << "\n\nERROR: do_job_aggr_cts()-> "
           << "the \"-line_type\" option must be set to one of:\n"
           << "\tFHO, CTC, CTP, CFP, or COP\n\n"
           << flush;

      throw(1);
   }

   //
   // Sum up the contingency table type lines
   //
   sum_contable_lines(jobstring, f, j, cts_info, lt, n_in, n_out);

   //
   // Check for no matching VSDB lines
   //
   if(n_out == 0) {
      cout << "WARNING: do_job_aggr_cts() -> "
           << "no matching VSDB lines found for job: " << jobstring
           << "\n" << flush;
      return;
   }

   //
   // Store the alpha information in the CTSInfo object
   //
   cts_info.allocate_n_alpha(1);
   cts_info.alpha[0] = j.alpha;

   //
   // Compute the stats and confidence intervals for this
   // CTSInfo object
   //
   cts_info.compute_stats();
   cts_info.compute_ci();

   //
   // Build the column names line and the output CTS line
   //
   get_column_names(cts_columns, n_cts_columns, col_names);
   sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
   build_cts_str(col_str, cts_info, 0);
   sprintf(out_line, "%-15s %s", "CTS:", col_str);

   //
   // Write the column names line, the aggregated CTS line, and the
   // job command line
   //
   write_line(hdr_line);
   write_line(out_line);
   write_jobstring(jobstring);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_aggr_cnt() routine should only be called when the
// -line_type option has been used exactly once for line types:
// SL1L2 or SAL1L2
//
////////////////////////////////////////////////////////////////////////

void do_job_aggr_cnt(const char *jobstring, LineDataFile &f,
                     VSDBAnalysisJob &j, int &n_in, int &n_out) {
   VSDBLine line;
   VSDBLineType lt;
   SL1L2Info sl1l2_info;
   VL1L2Info vl1l2_info;
   CNTInfo cnt_info;
   char hdr_line[max_line_len], col_names[max_line_len];
   char out_line[max_line_len], col_str[max_line_len];


   //
   // Check that the -line_type option has been supplied only once
   //
   if(j.line_type.n_elements() != 1) {
      cerr << "\n\nERROR: do_job_aggr_cnt()-> "
           << "this function may only be called when the "
           << "\"-line_type\" option has been used exactly once to "
           << "specify the line type from which to derive the "
           << "statistics: " << jobstring << "\n\n"
           << flush;

      throw(1);
   }

   //
   // Determine the line type specified for this job
   //
   lt = determine_line_type(j.line_type[0]);

   //
   // Check that a valid line type has been selected
   //
   if(lt != vsdb_sl1l2 && lt != vsdb_sal1l2) {
      cerr << "\n\nERROR: do_job_aggr_cnt()-> "
           << "the \"-line_type\" option must be set to one of:\n"
           << "\tSL1L2 or SAL1L2\n\n"
           << flush;

      throw(1);
   }

   //
   // Sum up the partial sums line type
   //
   sum_partial_sum_lines(jobstring, f, j, sl1l2_info, vl1l2_info, lt, n_in, n_out);

   //
   // Check for no matching VSDB lines
   //
   if(n_out == 0) {
      cout << "WARNING: do_job_aggr_cnt() -> "
           << "no matching VSDB lines found for job: " << jobstring
           << "\n" << flush;
      return;
   }

   //
   // Compute CNTInfo statistics from the aggregated SL1L2Info object
   //
   cnt_info.allocate_n_alpha(1);
   cnt_info.alpha[0] = j.alpha;

   // SL1L2 line type
   if(lt == vsdb_sl1l2) compute_cntinfo(sl1l2_info, 0, cnt_info);
   // SAL1L2 line type
   else                 compute_cntinfo(sl1l2_info, 1, cnt_info);

   //
   // Build the column names line and the output CNT line
   //
   get_column_names(cnt_columns, n_cnt_columns, col_names);
   sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
   build_cnt_str(col_str, cnt_info, 0);
   sprintf(out_line, "%-15s %s", "CNT:", col_str);

   //
   // Write the column names line, the aggregated CNT line, and the
   // job command line
   //
   write_line(hdr_line);
   write_line(out_line);
   write_jobstring(jobstring);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_aggr_mpr() routine should only be called when...
// The -var option has been used exactly once.
// The -out_line_type option has been used exactly once for line types:
//       FHO, CTC, CTP, CFP, COP, CTS, CNT, SL1L2, SAL1L2
// The -threshold option must be specified for out line types:
//       FHO, CTC, CTP, CFP, COP, CTS
// The -alpha, -boot_interval, -boot_rep_prop, -n_boot_rep,
// -boot_rng and -boot_seed arguments may be used to specify how to
// preform bootstrapping.
// The -mask_grid and/or -mask_poly optional arguments may be used to
// define a masking region.
//
////////////////////////////////////////////////////////////////////////

void do_job_aggr_mpr(const char *jobstring, LineDataFile &f,
                     VSDBAnalysisJob &j, int &n_in, int &n_out) {
   VSDBLine line;
   NumArray f_na, o_na, c_na;
   CTSInfo cts_info;
   CNTInfo cnt_info;
   SL1L2Info sl1l2_info;
   VL1L2Info vl1l2_info;
   char hdr_line[max_line_len], col_names[max_line_len];
   char out_line[max_line_len], col_str[max_line_len];
   char line_type[max_str_len];

   //
   // Check that the -var option has been used exactly once
   //
   if(j.var.n_elements() != 1) {
      cerr << "\n\nERROR: do_job_aggr_mpr()-> "
           << "this function may only be called when the "
           << "\"-var\" option has been used exactly once to "
           << "specify a variable type from which to derive the "
           << "statistics: " << jobstring << "\n\n"
           << flush;

      throw(1);
   }

   //
   // Check that the -out_line_type option has been used exactly once
   //
   if(j.out_line_type == no_vsdb_line_type) {
      cerr << "\n\nERROR: do_job_aggr_mpr()-> "
           << "this function may only be called when the "
           << "\"-out_line_type\" option has been used exactly once to "
           << "specify the line type from which to derive the "
           << "statistics: " << jobstring << "\n\n"
           << flush;

      throw(1);
   }

   //
   // Check that out_line_type is set to an allowed value
   //
   if(j.out_line_type != vsdb_fho && j.out_line_type != vsdb_ctc &&
      j.out_line_type != vsdb_ctp && j.out_line_type != vsdb_cfp &&
      j.out_line_type != vsdb_cop && j.out_line_type != vsdb_cts &&
      j.out_line_type != vsdb_cnt && j.out_line_type != vsdb_sl1l2 &&
      j.out_line_type != vsdb_sal1l2) {

      cerr << "\n\nERROR: do_job_aggr_mpr()-> "
              << "when \"-out_line_type\" must be set to one of: "
              << "FHO, CTC, CTP, CFP, COP, CTS, CNT, SL1L2, or SAL1L2: "
              << jobstring << "\n\n"
              << flush;

      throw(1);
   }

   //
   // Set the -line_type option to MPR
   //
   j.line_type.clear();
   j.line_type.add("MPR");

   //
   // Check for a required threshold value based on out_line_type
   //
   if(j.out_line_type == vsdb_fho || j.out_line_type == vsdb_ctc ||
      j.out_line_type == vsdb_ctp || j.out_line_type == vsdb_cfp ||
      j.out_line_type == vsdb_cop || j.out_line_type == vsdb_cts) {

      if(j.threshold.n_elements() != 1) {
         cerr << "\n\nERROR: do_job_aggr_mpr()-> "
              << "when \"-out_line_type\" is set to FHO, CTC, "
              << "CTP, CFP, COP, or CTS, the \"-threshold\" option "
              << "must also be specified: " << jobstring << "\n\n"
              << flush;

         throw(1);
      }
   }

   //
   // Handle the contingency table out line types which require
   // no bootstrapping
   //
   if(j.out_line_type == vsdb_fho || j.out_line_type == vsdb_ctc ||
      j.out_line_type == vsdb_ctp || j.out_line_type == vsdb_cfp ||
      j.out_line_type == vsdb_cop) {

      sum_mpr_lines_ct(jobstring, f, j, cts_info, n_in, n_out);
   }
   //
   // Handle the statistics lines types which do require
   // bootstrapping
   //
   else if(j.out_line_type == vsdb_cts || j.out_line_type == vsdb_cnt) {

      sum_mpr_lines_stats(jobstring, f, j, cnt_info, cts_info, n_in, n_out);
   }
   //
   // Handle the SL1L2 and SAL1L2 out line types
   //
   else if(j.out_line_type == vsdb_sl1l2 || j.out_line_type == vsdb_sal1l2) {

      sum_mpr_lines_psums(jobstring, f, j, sl1l2_info, n_in, n_out);
   }

   //
   // Check for no matching VSDB lines
   //
   if(n_out == 0) {
      cout << "WARNING: do_job_aggr_mpr() -> "
           << "no matching VSDB lines found for job: " << jobstring
           << "\n" << flush;
      return;
   }

   //
   // Construct the output line
   //
   switch(j.out_line_type) {

      case(vsdb_fho):

         //
         // Build the column names line and the output FHO line
         //
         get_column_names(fho_columns, n_fho_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_fho_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "FHO:", col_str);

         break;

      case(vsdb_ctc):

         //
         // Build the column names line and the output CTC line
         //
         get_column_names(ctc_columns, n_ctc_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_ctc_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "CTC:", col_str);

         break;

      case(vsdb_ctp):

         //
         // Build the column names line and the output CTP line
         //
         get_column_names(ctp_columns, n_ctp_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_ctp_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "CTP:", col_str);

         break;

      case(vsdb_cfp):

         //
         // Build the column names line and the output CFP line
         //
         get_column_names(cfp_columns, n_cfp_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_cfp_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "CFP:", col_str);

         break;

      case(vsdb_cop):

         //
         // Build the column names line and the output COP line
         //
         get_column_names(cop_columns, n_cop_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_cop_str(col_str, cts_info);
         sprintf(out_line, "%-15s %s", "COP:", col_str);

         break;

      case(vsdb_cts):

         //
         // Build the column names line and the output CTS line
         //
         get_column_names(cts_columns, n_cts_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_cts_str(col_str, cts_info, 0);
         sprintf(out_line, "%-15s %s", "CTS:", col_str);

         break;

      case(vsdb_cnt):

         //
         // Build the column names line and the output CNT line
         //
         get_column_names(cnt_columns, n_cnt_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_cnt_str(col_str, cnt_info, 0);
         sprintf(out_line, "%-15s %s", "CNT:", col_str);

         break;

      case(vsdb_sl1l2):

         //
         // Build the column names line and the output SL1L2 line
         //
         get_column_names(sl1l2_columns, n_sl1l2_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_sl1l2_str(col_str, sl1l2_info);
         sprintf(out_line, "%-15s %s", "SL1L2:", col_str);

         break;

      case(vsdb_sal1l2):

         //
         // Build the column names line and the output SAL1L2 line
         //
         get_column_names(sal1l2_columns, n_sal1l2_columns, col_names);
         sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
         build_sal1l2_str(col_str, sl1l2_info);
         sprintf(out_line, "%-15s %s", "SAL1L2:", col_str);

         break;

      default:
         vsdblinetype_to_string(line.type(), line_type);
         cerr << "\n\nERROR: do_job_aggr_mpr() -> "
              << "line type value of " << line_type
              << " not currently supported for the aggregation "
              << "job!\n\n" << flush;

         throw(1);
   } // end switch

   //
   // Write the column names line, the aggregated line, and the
   // job command line
   //
   write_line(hdr_line);
   write_line(out_line);
   write_jobstring(jobstring);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The do_job_go_index() routine is used to compute the GO Index.
// The GO Index is a weighted sum of the RMSE's for 12 different
// forecast parameters with lead times of 12, 24, 36, and 48 hours.
//
// The weights are based on lead time and are one of two sets of
// values:
// a: 4, 3, 2, 1 for lead times of 12, 24, 36, and 48 hours
// b: 8, 6, 4, 2 for lead times of 12, 24, 36, and 48 hours
//
// The 12 forecast parameters and corresponding sets of weights are
// as follows:
// (4) Wind Speed at the surface (b), 850 mb (a), 400 mb (a), and 250 mb (a)
// (4) Dewpoint at the surface (b), 850 mb (b), 700 mb (b), and 400 mb (b)
// (2) Temperature at the surface(b) and 400 mb (a)
// (1) Geopotential height at 400 mb (a)
// (1) Sea Level Pressure (b)
//
// The RMSE values are computed from the partial sums in the SL1L2
// VSDB lines as follows:
// rmse = sqrt( (sum(f*f) + sum(o*o) - 2*sum(f*o))/N )
//
////////////////////////////////////////////////////////////////////////

void do_job_go_index(const char *jobstring, LineDataFile &f,
                     VSDBAnalysisJob &j, int &n_in, int &n_out) {
   VSDBLine line;
   char hdr_line[max_line_len], col_names[max_line_len];
   char out_line[max_line_len], col_str[max_line_len];
   double go_index;

   //
   // When computing the GO Index, a single initialization time must
   // be set for the job. i.e. init_beg = init_end
   //
   if(j.init_beg == 0 ||
      j.init_end == 0 ||
      j.init_beg != j.init_end) {
      cerr << "\n\nERROR: do_job_go_index() -> "
           << "when computing the GO Index, the -init_beg and "
           << "-init_end job command line options must be set to "
           << "the same value.\n\n"
           << "\n\n" << flush;

      throw(1);
   }

   //
   // Compute GO Index
   //
   go_index = compute_go_index(jobstring, f, j, n_in, n_out);

   //
   // Check for no matching VSDB lines
   //
   if(n_out == 0) {
      cout << "WARNING: do_job_go_index() -> "
           << "no matching VSDB lines found for job: " << jobstring
           << "\n" << flush;
      return;
   }

   //
   // Build the column names line and the output GO Index line
   //
   get_column_names(job_go_columns, n_job_go_columns, col_names);
   sprintf(hdr_line, "%-15s %s", "COL_NAME:", col_names);
   dbl_to_str(go_index, col_str);
   sprintf(out_line, "%-15s %s", "GO_INDEX:", col_str);

   //
   // Write the column names line, the GO Index line, and the
   // job command line
   //
   write_line(hdr_line);
   write_line(out_line);
   write_jobstring(jobstring);

   return;
}

////////////////////////////////////////////////////////////////////////
//
// The sum_contable() routine should only be called when the -line_type
// option has been used exactly once.
//
////////////////////////////////////////////////////////////////////////

void sum_contable_lines(const char *jobstring, LineDataFile &f,
                        VSDBAnalysisJob &j, CTSInfo &cts_info,
                        VSDBLineType lt, int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   TTContingencyTable ct;
   int fy_oy, fy_on, fn_oy, fn_on;
   char line_type[max_str_len];

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: sum_contable_lines()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Initialize the Contingency Table counts
   //
   fy_oy = fy_on = fn_oy = fn_on = 0;
   cts_info.clear();

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // Zero out the contingecy table object
         //
         ct.zero_out();

         //
         // Switch on the line type looking only for contingency
         // table types of lines
         //
         switch(line.type()) {

            case(vsdb_fho):
               parse_fho_ctable(line, ct);
               break;

            case(vsdb_ctc):
               parse_ctc_ctable(line, ct);
               break;

            case(vsdb_ctp):
               parse_ctp_ctable(line, ct);
               break;

            case(vsdb_cfp):
               parse_cfp_ctable(line, ct);
               break;

            case(vsdb_cop):
               parse_cop_ctable(line, ct);
               break;

            case(vsdb_nbrctc):
               parse_nbrctc_ctable(line, ct);
               break;

            default:
               vsdblinetype_to_string(line.type(), line_type);
               cerr << "\n\nERROR: sum_contable_lines() -> "
                    << "line type value of " << line_type
                    << " not currently supported for the aggregation "
                    << "job!\n\n" << flush;

               throw(1);
         } // end switch

         //
         // Increment the ctable counts
         //
         fy_oy += ct.fy_oy();
         fy_on += ct.fy_on();
         fn_oy += ct.fn_oy();
         fn_on += ct.fn_on();

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // Store the ctable sums
   //
   cts_info.cts.set_fy_oy(fy_oy);
   cts_info.cts.set_fy_on(fy_on);
   cts_info.cts.set_fn_oy(fn_oy);
   cts_info.cts.set_fn_on(fn_on);

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void sum_partial_sum_lines(const char *jobstring, LineDataFile &f,
                           VSDBAnalysisJob &j,
                           SL1L2Info &sl1l2_info, VL1L2Info &vl1l2_info,
                           VSDBLineType lt, int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   SL1L2Info s;
   VL1L2Info v;

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: sum_partial_sum_lines()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Initialize the partial sums
   //
   sl1l2_info.clear();
   vl1l2_info.clear();

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // Switch on the line type.
         // For each partial sum line type, clear out the object,
         // parse the new line, and add it to running sum.
         //
         switch(line.type()) {

            case(vsdb_sl1l2):
               s.clear();
               parse_sl1l2_line(line, s);
               sl1l2_info += s;
               break;

            case(vsdb_sal1l2):
               s.clear();
               parse_sal1l2_line(line, s);
               sl1l2_info += s;
               break;

            case(vsdb_vl1l2):
               v.clear();
               parse_vl1l2_line(line, v);
               vl1l2_info += v;
               break;

            case(vsdb_val1l2):
               v.clear();
               parse_val1l2_line(line, v);
               vl1l2_info += v;
               break;

            default:
               cerr << "\n\nERROR: sum_partial_sum_lines() -> "
                    << "should only encounter partial sum line types!\n\n"
                    << flush;
               throw(1);
         } // end switch

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void sum_mpr_lines_ct(const char *jobstring, LineDataFile &f,
                            VSDBAnalysisJob &j,
                            CTSInfo &cts_info,
                            int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   MPRData m;

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: sum_mpr_lines_ct()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // Switch on the line type.
         //
         switch(line.type()) {

            case(vsdb_mpr):

               parse_mpr_line(line, m);

               //
               // Check the grid and poly masks if specified.
               // Convert degrees_east to degrees_west.
               //
               if(!j.is_in_grid_mask(m.obs_lat, (-1.0)*m.obs_lon) ||
                  !j.is_in_poly_mask(m.obs_lat, (-1.0)*m.obs_lon)) continue;

               //
               // Check for bad data.
               //
               if(is_bad_data(m.fcst) || is_bad_data(m.obs)) continue;

               //
               // Update the contingency table counts for this MPR line
               //
               if     ( j.thresh.check(m.fcst) &&  j.thresh.check(m.obs))
                  cts_info.cts.inc_fy_oy();
               else if( j.thresh.check(m.fcst) && !j.thresh.check(m.obs))
                  cts_info.cts.inc_fy_on();
               else if(!j.thresh.check(m.fcst) &&  j.thresh.check(m.obs))
                  cts_info.cts.inc_fn_oy();
               else if(!j.thresh.check(m.fcst) && !j.thresh.check(m.obs))
                  cts_info.cts.inc_fn_on();

               break;

            default:
               cerr << "\n\nERROR: sum_mpr_lines_ct() -> "
                    << "should only encounter MPR line types!\n\n"
                    << flush;
               throw(1);
         } // end switch

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void sum_mpr_lines_stats(const char *jobstring, LineDataFile &f,
                         VSDBAnalysisJob &j,
                         CNTInfo &cnt_info, CTSInfo &cts_info,
                         int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   MPRData m;
   NumArray f_na, o_na;
   CTSInfo *cts_info_ptr = &cts_info;

   //
   // Store the alpha value
   //
   cnt_info.allocate_n_alpha(1);
   cnt_info.alpha[0] = j.alpha;
   cts_info.allocate_n_alpha(1);
   cts_info.alpha[0] = j.alpha;

   //
   // Store the threshold to be applied.
   // If one is not specified, just set it >=0
   //
   if(j.thresh.type == thresh_na)
      cts_info.cts_thresh.set(0.0, thresh_ge);
   else
      cts_info.cts_thresh = j.thresh;

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: sum_mpr_lines_stats()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // Switch on the line type.
         //
         switch(line.type()) {

            case(vsdb_mpr):

               parse_mpr_line(line, m);

               //
               // Check the grid and poly masks if specified.
               // Convert degrees_east to degrees_west.
               //
               if(!j.is_in_grid_mask(m.obs_lat, (-1.0)*m.obs_lon) ||
                  !j.is_in_poly_mask(m.obs_lat, (-1.0)*m.obs_lon)) continue;

               //
               // Check for bad data.
               //
               if(is_bad_data(m.fcst) || is_bad_data(m.obs)) continue;

               //
               // Store the fcst and obs pair
               //
               f_na.add(m.fcst);
               o_na.add(m.obs);

               break;

            default:
               cerr << "\n\nERROR: sum_mpr_lines_stats() -> "
                    << "should only encounter MPR line types!\n\n"
                    << flush;
               throw(1);
         } // end switch

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // Set up the random number generator and seed value
   //
   rng_set(rng_ptr, j.boot_rng, j.boot_seed);

   //
   // Compute the stats, normal confidence intervals, and bootstrap
   // bootstrap confidence intervals
   //
   if(j.boot_interval == boot_bca_flag) {
      compute_stats_ci_bca(rng_ptr, f_na, o_na, -1,
         j.n_boot_rep,
         cnt_info, cts_info_ptr, 1, conf.rank_corr_flag().ival(),
         tmp_dir);
   }
   else {
      compute_stats_ci_perc(rng_ptr, f_na, o_na, -1,
         j.n_boot_rep, j.boot_rep_prop,
         cnt_info, cts_info_ptr, 1, conf.rank_corr_flag().ival(),
         tmp_dir);
   }

   //
   // Deallocate memory for the random number generator
   //
   rng_free(rng_ptr);

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void sum_mpr_lines_psums(const char *jobstring, LineDataFile &ldf,
               VSDBAnalysisJob &j, SL1L2Info &s_info,
               int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   MPRData m;

   int scount, sacount;
   double f, o, c;
   double f_sum,  o_sum,  ff_sum,  oo_sum,  fo_sum;
   double fa_sum, oa_sum, ffa_sum, ooa_sum, foa_sum;

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: sum_mpr_lines_psums()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Initialize the SL1L2Info object and counts
   //
   s_info.clear();
   scount = sacount = 0;
   f_sum  = o_sum  =  ff_sum  = oo_sum  = fo_sum  = 0.0;
   fa_sum = oa_sum =  ffa_sum = ooa_sum = foa_sum = 0.0;

   //
   // Process the VSDB lines
   //
   while(ldf >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // Switch on the line type.
         //
         switch(line.type()) {

            case(vsdb_mpr):

               parse_mpr_line(line, m);

               //
               // Check the grid and poly masks if specified.
               // Convert degrees_east to degrees_west.
               //
               if(!j.is_in_grid_mask(m.obs_lat, (-1.0)*m.obs_lon) ||
                  !j.is_in_poly_mask(m.obs_lat, (-1.0)*m.obs_lon)) continue;

               //
               // Check for bad data.
               //
               if(is_bad_data(m.fcst) || is_bad_data(m.obs)) continue;

               //
               // Update the counts for this MPR line
               //
               f = m.fcst;
               o = m.obs;
               c = m.climo;

               f_sum   += f;
               o_sum   += o;
               ff_sum  += f*f;
               oo_sum  += o*o;
               fo_sum  += f*o;
               scount  += 1;

               //
               // Check c for valid data
               //
               if(is_bad_data(c)) break;

               fa_sum  += f-c;
               oa_sum  += o-c;
               ffa_sum += (f-c)*(f-c);
               ooa_sum += (o-c)*(o-c);
               foa_sum += (f-c)*(o-c);
               sacount += 1;

               break;

            default:
               cerr << "\n\nERROR: sum_mpr_lines_psums() -> "
                    << "should only encounter MPR line types!\n\n"
                    << flush;
               throw(1);
         } // end switch

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   if(scount != 0) {
      s_info.scount  = scount;
      s_info.fbar    = f_sum/scount;
      s_info.obar    = o_sum/scount;
      s_info.fobar   = fo_sum/scount;
      s_info.ffbar   = ff_sum/scount;
      s_info.oobar   = oo_sum/scount;
   }

   if(sacount != 0) {
      s_info.sacount = sacount;
      s_info.fabar   = fa_sum/sacount;
      s_info.oabar   = oa_sum/sacount;
      s_info.foabar  = foa_sum/sacount;
      s_info.ffabar  = ffa_sum/sacount;
      s_info.ooabar  = ooa_sum/sacount;
   }

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void sum_isc_lines(const char *jobstring, LineDataFile &ldf,
               VSDBAnalysisJob &j, ISCInfo &isc_aggr,
               int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   ISCData isc_data;

   int i, iscale;
   double baser_fbias_sum, den;
   NumArray total_na, mse_na, fen_na, oen_na, baser_na, fbias_na;
   NumArray w_na;

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: sum_isc_lines()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Determine the scale being aggregated from the line type
   //
   iscale = parse_iscale(j.line_type[0]);

   //
   // Initialize the ISCInfo object
   //
   isc_aggr.clear();
   isc_aggr.zero_out();

   //
   // Process the VSDB lines
   //
   while(ldf >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // Switch on the line type.
         //
         switch(line.type()) {

            case(vsdb_isc):

               parse_isc_line(line, isc_data);

               //
               // Check for bad data
               //
               if(isc_data.total == bad_data_int ||
                  isc_data.mse   == bad_data_double ||
                  isc_data.fen   == bad_data_double ||
                  isc_data.oen   == bad_data_double ||
                  isc_data.baser == bad_data_double ||
                  isc_data.fbias == bad_data_double) continue;

               //
               // Store the data for this ISC line
               //
               isc_aggr.total += isc_data.total;

               total_na.add(isc_data.total);
               mse_na.add(isc_data.mse);
               fen_na.add(isc_data.fen);
               oen_na.add(isc_data.oen);
               baser_na.add(isc_data.baser);
               fbias_na.add(isc_data.fbias);

               //
               // Store the number of scales if not set or check
               // to make sure that it doesn't change.
               //
               if(isc_aggr.nscale == 0) {
                  isc_aggr.nscale = isc_data.nscale;
               }
               else {
                  if(isc_aggr.nscale != isc_data.nscale) {
                     cerr << "\n\nERROR: sum_isc_lines() -> "
                          << "the number of scales must remain constant "
                          << "when aggregating ISC lines!\n\n"
                          << flush;
                     throw(1);
                  }
               }

               break;

            default:
               cerr << "\n\nERROR: sum_isc_lines() -> "
                    << "should only encounter MPR line types!\n\n"
                    << flush;
               throw(1);
         } // end switch

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // Get the sum of the totals and compute weights
   //
   for(i=0; i<total_na.n_elements(); i++) w_na.add(total_na[i]/isc_aggr.total);

   //
   // Compute weighted means
   //
   baser_fbias_sum = 0.0;
   for(i=0; i<w_na.n_elements(); i++) {
      isc_aggr.mse    += w_na[i]*mse_na[i];
      isc_aggr.fen    += w_na[i]*fen_na[i];
      isc_aggr.oen    += w_na[i]*oen_na[i];
      isc_aggr.baser  += w_na[i]*baser_na[i];
      baser_fbias_sum += w_na[i]*baser_na[i]*fbias_na[i];
   }

   //
   // Compute the aggregated FBIAS
   //
   isc_aggr.fbias = baser_fbias_sum/isc_aggr.baser;

   //
   // Compute the aggregated ISC score.  For scale 0 which contains the
   // the whole field, do no divide by the number of scales.  For all
   // scales, divide by the number of scales.
   //
   if(iscale == 0) {
      den = (isc_aggr.fbias*isc_aggr.baser*(1.0 - isc_aggr.baser) +
             isc_aggr.baser*(1.0 - isc_aggr.fbias*isc_aggr.baser));
   }
   else {
      den = (isc_aggr.fbias*isc_aggr.baser*(1.0 - isc_aggr.baser) +
             isc_aggr.baser*(1.0 - isc_aggr.fbias*isc_aggr.baser))
            /(isc_aggr.nscale+1);
   }

   if(isc_aggr.fbias == bad_data_double ||
      isc_aggr.baser == bad_data_double ||
      den            == 0.0) isc_aggr.isc = bad_data_double;
   else                      isc_aggr.isc = 1.0 - isc_aggr.mse/den;

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_fho_ctable(VSDBLine &l, TTContingencyTable &ct) {
   int n, fy, fy_oy, oy;
   double f_rate, h_rate, o_rate;

   n      = atoi(l.get_item(fho_total_offset));
   f_rate = atof(l.get_item(fho_f_rate_offset));
   h_rate = atof(l.get_item(fho_h_rate_offset));
   o_rate = atof(l.get_item(fho_o_rate_offset));

   fy    = nint(n * f_rate);
   fy_oy = nint(n * h_rate);
   oy    = nint(n * o_rate);

   // FY_OY
   ct.set_fy_oy(fy_oy);

   // FY_ON
   ct.set_fy_on(fy - fy_oy);

   // FN_OY
   ct.set_fn_oy(oy - fy_oy);

   // FN_ON
   ct.set_fn_on(n - fy - oy + fy_oy);

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_ctc_ctable(VSDBLine &l, TTContingencyTable &ct) {

   // FY_OY
   ct.set_fy_oy(atoi(l.get_item(ctc_fy_oy_offset)));

   // FY_ON
   ct.set_fy_on(atoi(l.get_item(ctc_fy_on_offset)));

   // FN_OY
   ct.set_fn_oy(atoi(l.get_item(ctc_fn_oy_offset)));

   // FN_ON
   ct.set_fn_on(atoi(l.get_item(ctc_fn_on_offset)));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_ctp_ctable(VSDBLine &l, TTContingencyTable &ct) {
   int n;
   double fy_oy_tp, fy_on_tp, fn_oy_tp, fn_on_tp;

   n        = atoi(l.get_item(ctp_total_offset));
   fy_oy_tp = atof(l.get_item(ctp_fy_oy_tp_offset));
   fy_on_tp = atof(l.get_item(ctp_fy_on_tp_offset));
   fn_oy_tp = atof(l.get_item(ctp_fn_oy_tp_offset));
   fn_on_tp = atof(l.get_item(ctp_fn_on_tp_offset));

   // FY_OY
   ct.set_fy_oy(nint(n * fy_oy_tp));

   // FY_ON
   ct.set_fy_on(nint(n * fy_on_tp));

   // FN_OY
   ct.set_fn_oy(nint(n * fn_oy_tp));

   // FN_ON
   ct.set_fn_on(nint(n * fn_on_tp));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_cfp_ctable(VSDBLine &l, TTContingencyTable &ct) {
   int fy, fn;
   double fy_oy_fp, fy_on_fp, fn_oy_fp, fn_on_fp;

   fy       = atoi(l.get_item(cfp_fy_offset));
   fn       = atoi(l.get_item(cfp_fn_offset));
   fy_oy_fp = atof(l.get_item(cfp_fy_oy_fp_offset));
   fy_on_fp = atof(l.get_item(cfp_fy_on_fp_offset));
   fn_oy_fp = atof(l.get_item(cfp_fn_oy_fp_offset));
   fn_on_fp = atof(l.get_item(cfp_fn_on_fp_offset));

   // FY_OY
   ct.set_fy_oy(nint(fy * fy_oy_fp));

   // FY_ON
   ct.set_fy_on(nint(fy * fy_on_fp));

   // FN_OY
   ct.set_fn_oy(nint(fn * fn_oy_fp));

   // FN_ON
   ct.set_fn_on(nint(fn * fn_on_fp));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_cop_ctable(VSDBLine &l, TTContingencyTable &ct) {
   int oy, on;
   double fy_oy_op, fy_on_op, fn_oy_op, fn_on_op;

   oy       = atoi(l.get_item(cop_oy_offset));
   on       = atoi(l.get_item(cop_on_offset));
   fy_oy_op = atof(l.get_item(cop_fy_oy_op_offset));
   fy_on_op = atof(l.get_item(cop_fy_on_op_offset));
   fn_oy_op = atof(l.get_item(cop_fn_oy_op_offset));
   fn_on_op = atof(l.get_item(cop_fn_on_op_offset));

   // FY_OY
   ct.set_fy_oy(nint(oy * fy_oy_op));

   // FY_ON
   ct.set_fy_on(nint(on * fy_on_op));

   // FN_OY
   ct.set_fn_oy(nint(oy * fn_oy_op));

   // FN_ON
   ct.set_fn_on(nint(on * fn_on_op));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_nbrctc_ctable(VSDBLine &l, TTContingencyTable &ct) {

   // FY_OY
   ct.set_fy_oy(atoi(l.get_item(nbrctc_fy_oy_offset)));

   // FY_ON
   ct.set_fy_on(atoi(l.get_item(nbrctc_fy_on_offset)));

   // FN_OY
   ct.set_fn_oy(atoi(l.get_item(nbrctc_fn_oy_offset)));

   // FN_ON
   ct.set_fn_on(atoi(l.get_item(nbrctc_fn_on_offset)));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_sl1l2_line(VSDBLine &l, SL1L2Info &s_info) {

   s_info.clear();

   s_info.scount = atoi(l.get_item(sl1l2_total_offset));
   s_info.fbar   = atof(l.get_item(sl1l2_fbar_offset));
   s_info.obar   = atof(l.get_item(sl1l2_obar_offset));
   s_info.fobar  = atof(l.get_item(sl1l2_fobar_offset));
   s_info.ffbar  = atof(l.get_item(sl1l2_ffbar_offset));
   s_info.oobar  = atof(l.get_item(sl1l2_oobar_offset));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_sal1l2_line(VSDBLine &l, SL1L2Info &s_info) {

   s_info.clear();

   s_info.sacount = atoi(l.get_item(sal1l2_total_offset));
   s_info.fabar   = atof(l.get_item(sal1l2_fabar_offset));
   s_info.oabar   = atof(l.get_item(sal1l2_oabar_offset));
   s_info.foabar  = atof(l.get_item(sal1l2_foabar_offset));
   s_info.ffabar  = atof(l.get_item(sal1l2_ffabar_offset));
   s_info.ooabar  = atof(l.get_item(sal1l2_ooabar_offset));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_vl1l2_line(VSDBLine &l, VL1L2Info &v_info) {

   v_info.clear();

   v_info.vcount  = atoi(l.get_item(vl1l2_total_offset));
   v_info.ufbar   = atof(l.get_item(vl1l2_ufbar_offset));
   v_info.vfbar   = atof(l.get_item(vl1l2_vfbar_offset));
   v_info.uobar   = atof(l.get_item(vl1l2_uobar_offset));
   v_info.vobar   = atof(l.get_item(vl1l2_vobar_offset));
   v_info.uvfobar = atof(l.get_item(vl1l2_uvfobar_offset));
   v_info.uvffbar = atof(l.get_item(vl1l2_uvffbar_offset));
   v_info.uvoobar = atof(l.get_item(vl1l2_uvoobar_offset));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_val1l2_line(VSDBLine &l, VL1L2Info &v_info) {

   v_info.clear();

   v_info.vacount  = atoi(l.get_item(val1l2_total_offset));
   v_info.ufabar   = atof(l.get_item(val1l2_ufabar_offset));
   v_info.vfabar   = atof(l.get_item(val1l2_vfabar_offset));
   v_info.uoabar   = atof(l.get_item(val1l2_uoabar_offset));
   v_info.voabar   = atof(l.get_item(val1l2_voabar_offset));
   v_info.uvfoabar = atof(l.get_item(val1l2_uvfoabar_offset));
   v_info.uvffabar = atof(l.get_item(val1l2_uvffabar_offset));
   v_info.uvooabar = atof(l.get_item(val1l2_uvooabar_offset));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_mpr_line(VSDBLine &l, MPRData &m_data) {

   m_data.gc      = str_to_grib_code(l.get_item(var_offset));
   m_data.total   = atoi(l.get_item(mpr_total_offset));
   m_data.index   = atoi(l.get_item(mpr_index_offset));
   m_data.obs_lat = atof(l.get_item(mpr_obs_lat_offset));
   m_data.obs_lon = atof(l.get_item(mpr_obs_lon_offset));
   m_data.obs_lvl = atof(l.get_item(mpr_obs_lvl_offset));
   m_data.obs_elv = atof(l.get_item(mpr_obs_elv_offset));
   m_data.fcst    = atof(l.get_item(mpr_fcst_offset));
   m_data.obs     = atof(l.get_item(mpr_obs_offset));
   m_data.climo   = atof(l.get_item(mpr_climo_offset));

   return;
}

////////////////////////////////////////////////////////////////////////

void parse_isc_line(VSDBLine &l, ISCData &i_data) {

   i_data.total  = atoi(l.get_item(isc_total_offset));
   i_data.nscale = atoi(l.get_item(isc_nscale_offset));
   i_data.mse    = atof(l.get_item(isc_mse_offset));
   i_data.isc    = atof(l.get_item(isc_isc_offset));
   i_data.fen    = atof(l.get_item(isc_fenergy2_offset));
   i_data.oen    = atof(l.get_item(isc_oenergy2_offset));
   i_data.baser  = atof(l.get_item(isc_baser_offset));
   i_data.fbias  = atof(l.get_item(isc_fbias_offset));

   return;
}

////////////////////////////////////////////////////////////////////////
//
// Parse the ISCALE value out of the specified line type of the form:
//    "ISC>=0.0000/i" where "i" is the iscale value
//
////////////////////////////////////////////////////////////////////////

int parse_iscale(const char *c) {
   int iscale;
   char *ptr;

   if((ptr = strchr(c, '/')) == NULL) {
      cerr << "\n\nERROR: parse_iscale()-> "
           << "The ISC line type string (" << c << ") should contain "
           << "a \"/\" character followed by a scale value|\n\n"
           << flush;

      throw(1);
   }

   iscale = atoi(ptr+1);

   return(iscale);
}

////////////////////////////////////////////////////////////////////////

double compute_go_index(const char *jobstring, LineDataFile &f,
                        VSDBAnalysisJob &j, int &n_in, int &n_out) {
   ofstream *dr_out = (ofstream *) 0;
   VSDBLine line;
   SL1L2Info s_info;
   int i;
   double go_index, go_sum, weight_sum;

   //
   // SL1L2Info objects to hold the counts for each variable and level
   // combination and each of the 4 lead times (12, 24, 36, and 48 hr)
   //
   SL1L2Info ws_sfc[4],      ws_850[4],      ws_400[4],      ws_250[4];
   double    ws_sfc_rmse[4], ws_850_rmse[4], ws_400_rmse[4], ws_250_rmse[4];
   SL1L2Info td_sfc[4],      td_850[4],      td_700[4],      td_400[4];
   double    td_sfc_rmse[4], td_850_rmse[4], td_700_rmse[4], td_400_rmse[4];
   SL1L2Info tt_sfc[4],      tt_400[4];
   double    tt_sfc_rmse[4], tt_400_rmse[4];
   SL1L2Info ht_400[4];
   double    ht_400_rmse[4];
   SL1L2Info slp[4];
   double    slp_rmse[4];

   const double a_weight[4] = { 4.0, 3.0, 2.0, 1.0 };
   const double b_weight[4] = { 8.0, 6.0, 4.0, 2.0 };

   //
   // Initialize the RMSE values
   //
   for(i=0; i<4; i++) {
      ws_sfc_rmse[i] = ws_850_rmse[i] = ws_400_rmse[i] = ws_250_rmse[i] = 0.0;
      td_sfc_rmse[i] = td_850_rmse[i] = td_700_rmse[i] = td_400_rmse[i] = 0.0;
      tt_sfc_rmse[i] = tt_400_rmse[i] = ht_400_rmse[i] = slp_rmse   [i] = 0.0;
   }

   //
   // If the -dump_row option was supplied open up the output file
   //
   if(j.dump_row.n_elements() > 0) {

      dr_out = new ofstream;
      dr_out->open(j.dump_row[0]);

      if(!dr_out) {
         cerr << "\n\nERROR: compute_go_index()-> "
              << "can't open the output file \"" << j.dump_row[0]
              << "\" for writing|\n\n" << flush;

            throw(1);
      }

      if(verbosity > 0) {
         cout << "Creating VSDB output file \"" << j.dump_row[0]
              << "\"\n" << flush;
      }
   }

   //
   // Process the VSDB lines
   //
   while(f >> line) {

      n_in++;

      if(j.is_keeper(line)) {

         //
         // The GO Index is computed only from VSDB line type SL1L2.
         //
         if(determine_line_type(line.line_type()) != vsdb_sl1l2 )
            continue;

         //
         // Determine the lead time.  The GO Index is only computed
         // using lead times of 12, 24, 36, and 48 hours.
         //
         if(     line.fcst_lead() == sec_per_hour * 12) i = 0;
         else if(line.fcst_lead() == sec_per_hour * 24) i = 1;
         else if(line.fcst_lead() == sec_per_hour * 36) i = 2;
         else if(line.fcst_lead() == sec_per_hour * 48) i = 3;
         else                                          continue;

         //
         // The GO Index is computed from the following 12 combinations
         // of variable type and level:
         // Wind Speed:    (4) WIND SFC (Z10), WIND P850, WIND P400, WIND P250
         // Dew Point:     (4) DPT  SFC (Z2),  DPT  P850, DPT  P700, DPT  P400
         // Temperature:   (2) TMP  SFC (Z2),  TMP  P400
         // Height:        (1) HGT  P400
         // Sea Level Prs: (1) PRMSL SFC (Z0)
         //

         //
         // Parse the line into an SL1L2Info object
         //
         parse_sl1l2_line(line, s_info);

         //
         // Handle the WIND variable type
         //
         if(strcmp(line.var(), "WIND") == 0) {

            if(     strncmp(line.level(), "Z10", strlen("Z10")) == 0)
               ws_sfc[i] += s_info;
            else if(strcmp(line.level(), "P850") == 0)
               ws_850[i] += s_info;
            else if(strcmp(line.level(), "P400") == 0)
               ws_400[i] += s_info;
            else if(strcmp(line.level(), "P250") == 0)
               ws_250[i] += s_info;
            else
               continue;
         }

         //
         // Handle the DPT variable type
         //
         else if(strcmp(line.var(), "DPT") == 0) {

            if(     strncmp(line.level(), "Z2", strlen("Z2")) == 0)
               td_sfc[i] += s_info;
            else if(strcmp(line.level(), "P850") == 0)
               td_850[i] += s_info;
            else if(strcmp(line.level(), "P700") == 0)
               td_700[i] += s_info;
            else if(strcmp(line.level(), "P400") == 0)
               td_400[i] += s_info;
            else
               continue;
         }

         //
         // Handle the TMP variable type
         //
         else if(strcmp(line.var(), "TMP") == 0) {

            if(     strncmp(line.level(), "Z2", strlen("Z2")) == 0)
               tt_sfc[i] += s_info;
            else if(strcmp(line.level(), "P400") == 0)
               tt_400[i] += s_info;
            else
               continue;
         }

         //
         // Handle the HGT variable type
         //
         else if(strcmp(line.var(), "HGT") == 0) {

            if(     strcmp(line.level(), "P400") == 0)
               ht_400[i] += s_info;
            else
               continue;
         }

         //
         // Handle the PRMSL variable type
         //
         else if(strcmp(line.var(), "PRMSL") == 0) {

            if(     strncmp(line.level(), "Z0", strlen("Z0")) == 0)
               slp[i] += s_info;
            else
               continue;
         }

         //
         // Handle all other variable types
         //
         else
            continue;

         if(dr_out) *(dr_out) << line;

         n_out++;
      }
   } // end while

   //
   // Compute the RMSE values for each SL1L2Info object
   //
   for(i=0; i<4; i++) {
      ws_sfc_rmse[i] = compute_sl1l2_rmse(ws_sfc[i], "WIND",  "Z10",  (i+1)*12);
      ws_850_rmse[i] = compute_sl1l2_rmse(ws_850[i], "WIND",  "P850", (i+1)*12);
      ws_400_rmse[i] = compute_sl1l2_rmse(ws_400[i], "WIND",  "P400", (i+1)*12);
      ws_250_rmse[i] = compute_sl1l2_rmse(ws_250[i], "WIND",  "P250", (i+1)*12);
      td_sfc_rmse[i] = compute_sl1l2_rmse(td_sfc[i], "DPT",   "Z2",   (i+1)*12);
      td_850_rmse[i] = compute_sl1l2_rmse(td_850[i], "DPT",   "P850", (i+1)*12);
      td_700_rmse[i] = compute_sl1l2_rmse(td_700[i], "DPT",   "P700", (i+1)*12);
      td_400_rmse[i] = compute_sl1l2_rmse(td_400[i], "DPT",   "P400", (i+1)*12);
      tt_sfc_rmse[i] = compute_sl1l2_rmse(tt_sfc[i], "TMP",   "Z2",   (i+1)*12);
      tt_400_rmse[i] = compute_sl1l2_rmse(tt_400[i], "TMP",   "P400", (i+1)*12);
      ht_400_rmse[i] = compute_sl1l2_rmse(ht_400[i], "HGT",   "P400", (i+1)*12);
      slp_rmse[i]    = compute_sl1l2_rmse(slp[i],    "PRMSL", "Z0",   (i+1)*12);
   }

   //
   // Compute GO Index value
   //
   go_sum     = 0.0;
   weight_sum = 0.0;
   go_index   = 0.0;

   for(i=0; i<4; i++) {

      if(ws_sfc[i].scount > 0) {
         go_sum     += ws_sfc_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }

      if(ws_850[i].scount > 0) {
         go_sum     += ws_850_rmse[i] * a_weight[i];
         weight_sum += a_weight[i];
      }

      if(ws_400[i].scount > 0) {
         go_sum     += ws_400_rmse[i] * a_weight[i];
         weight_sum += a_weight[i];
      }

      if(ws_250[i].scount > 0) {
         go_sum     += ws_250_rmse[i] * a_weight[i];
         weight_sum += a_weight[i];
      }

      if(td_sfc[i].scount > 0) {
         go_sum     += td_sfc_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }

      if(td_850[i].scount > 0) {
         go_sum     += td_850_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }

      if(td_700[i].scount > 0) {
         go_sum     += td_700_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }

      if(td_400[i].scount > 0) {
         go_sum     += td_400_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }

      if(tt_sfc[i].scount > 0) {
         go_sum     += tt_sfc_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }

      if(tt_400[i].scount > 0) {
         go_sum     += tt_400_rmse[i] * a_weight[i];
         weight_sum += a_weight[i];
      }

      if(ht_400[i].scount > 0) {
         go_sum     += ht_400_rmse[i] * a_weight[i];
         weight_sum += a_weight[i];
      }

      if(slp[i].scount > 0) {
         go_sum     += slp_rmse[i] * b_weight[i];
         weight_sum += b_weight[i];
      }
   }
   if(weight_sum == 0.0) go_index = bad_data_double;
   else                  go_index = go_sum/weight_sum;

   //
   // If an output file was created, close it and deallocate memory
   //
   if(dr_out) {
      dr_out->close();
      delete dr_out;
      dr_out = (ofstream *) 0;
   }

   return(go_index);
}

////////////////////////////////////////////////////////////////////////

double compute_sl1l2_rmse(SL1L2Info &s_info, const char *var,
                        const char *lvl, int lead_hr) {
   int n;
   double ff_sum, oo_sum, fo_sum, rmse;

   n = s_info.scount;

   if(n == 0) {
      cout << "WARNING: compute_sl1l2_rmse() -> "
           << "no matching VSDB lines for GO Index variable \"" << var
           << "\" at level \"" << lvl << "\" for lead time of \""
           << lead_hr << "\" hours.\n" << flush;
      return(0.0);
   }

   ff_sum = s_info.ffbar*s_info.scount;
   oo_sum = s_info.oobar*s_info.scount;
   fo_sum = s_info.fobar*s_info.scount;

   rmse = sqrt( (ff_sum + oo_sum - 2.0*fo_sum)/(double) n);

   return(rmse);
}

////////////////////////////////////////////////////////////////////////

void clean_up() {

   //
   // Delete the temp file
   //
   if(remove(tmp_path) != 0) {
      cerr << "\n\nERROR: clean_up() -> "
           << "can't remove temporary file \"" << tmp_path
           << "\"\n\n" << flush;

      exit(1);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_jobstring(const char *jobstring) {
   char out_line[max_line_len];

   sprintf(out_line, "%-15s %s", "JOB_COMMAND:", jobstring);

   write_line(out_line);

   return;
}

////////////////////////////////////////////////////////////////////////

void write_line(const char *str) {

   if(va_out) *(va_out) << str << "\n" << flush;
   else       cout      << str << "\n" << flush;

   return;
}

////////////////////////////////////////////////////////////////////////

void usage() {

   cout << "\nUsage: "
        << program_name << "\n"
        << "\t-lookin path\n"
        << "\t[-out filename]\n"
        << "\t[-v level]\n"
        << "\t[-config config_file] | [JOB COMMAND LINE]\n\n"

        << "\twhere\t\"-lookin path\" specifies a VSDB file or "
        << "top-level directory containing VSDB files.  It allows the "
        << "use of wildcards (at least one required).\n"

        << "\t\t\"-out filename\" specifies a file name to which "
        << "output should written rather than the screen (optional).\n"

        << "\t\t\"-v level\" overrides the default level of logging ("
        << verbosity << ") (optional).\n"

        << "\t\t\"-config config_file\" specifies a VSDBAnalysis "
        << "config file containing VSDBAnalysis jobs to be run.\n"

        << "\t\t\"JOB COMMAND LINE\" specifies all the arguments "
        << "necessary to perform a single VSDBAnalysis job.\n\n"

        << "\tNOTE: Refer to a VSDBAnalysis config file for details "
        << "on how to specify the JOB COMMAND LINE.\n\n"

        << flush;

   return;
}

////////////////////////////////////////////////////////////////////////
